/*
 * Java
 *
 * Copyright 2018-2019 MicroEJ Corp. All rights reserved.
 * For demonstration purpose only.
 * MicroEJ Corp. PROPRIETARY. Use is subject to license terms.
 */
package com.microej.demo.thermorest.endpoint;

import java.io.IOException;
import java.util.Map;

import org.json.me.JSONException;
import org.json.me.JSONObject;

import com.microej.demo.thermorest.HeadlessThermostat;
import com.microej.demo.thermorest.Messages;
import com.microej.demo.thermorest.cloud.CloudConnector;
import com.microej.demo.thermorest.model.ThermostatModel;
import com.microej.demo.thermorest.util.JSONStringBuilder;

import ej.hoka.http.HTTPConstants;
import ej.hoka.http.HTTPRequest;
import ej.hoka.http.HTTPResponse;
import ej.hoka.http.body.StringBodyParser;
import ej.restserver.RestEndpoint;
import ej.util.message.Level;

/**
 * The endpoint to get the temperature
 *
 * /thermostat: get: description: Return the current and target temperature produces: - application/json responses: 200:
 * description: Request was successful headers: schema: type: object properties: current: type: float description: the
 * current temperature target: type: float description: the target temperature put: description: Set the target
 * temperature produces: - application/json urlencoded parameters: - target: the target temperature responses: 200:
 * description: Request was successful headers: schema: type: object properties: current: type: float description: the
 * current temperature target: type: float description: the target temperature 400: description: target parameter was
 * not found.
 *
 */
public class ThermostatEndPoint extends RestEndpoint {

	private static final String APPLICATION_JSON = "application/json"; //$NON-NLS-1$
	private static final String CLOUDS = "clouds"; //$NON-NLS-1$
	private static final String ERROR = "Error"; //$NON-NLS-1$
	private static final String URI = "/temperature"; //$NON-NLS-1$
	private final HeadlessThermostat thermostat;

	/**
	 * Instantiates a {@link ThermostatEndPoint}.
	 *
	 * @param thermostat
	 *            the thermostat to use.
	 * @throws IllegalArgumentException
	 *             if URI is empty
	 */
	public ThermostatEndPoint(HeadlessThermostat thermostat) throws IllegalArgumentException {
		super(URI);
		this.thermostat = thermostat;
	}

	@Override
	public HTTPResponse get(HTTPRequest request, Map<String, String> headers, Map<String, String> parameters) {
		return getResponse();
	}

	@Override
	public HTTPResponse post(HTTPRequest request, Map<String, String> headers, Map<String, String> parameters) {
		HTTPResponse response = null;
		JSONObject body = null;
		try {
			StringBodyParser bodyParser = new StringBodyParser();
			bodyParser.parseBody(request);
			body = new JSONObject(bodyParser.getBody());
		} catch (JSONException | IOException e) {
			Messages.LOGGER.log(Level.INFO, Messages.CATEGORY, Messages.INVALID_PARAMETER, e);
			// Do nothing, target will be null.
			response = badRequest(e.getMessage());
		}
		double target = (body == null) ? Double.NaN : body.optDouble(ThermostatModel.TARGET);
		if (target != Double.NaN) {
			this.thermostat.getModel().setTargetTemperature((float) target);
			response = getResponse();
		} else {
			response = badRequest(ThermostatModel.TARGET + " param not found."); //$NON-NLS-1$
		}
		return response;
	}

	private HTTPResponse badRequest(String error) {
		StringBuilder builder = new StringBuilder();
		builder.append('{');
		JSONStringBuilder.append(builder, ERROR, error);
		builder.append('}');
		HTTPResponse response = new HTTPResponse(builder.toString());
		response.setStatus(HTTPConstants.HTTP_STATUS_BADREQUEST);
		return response;
	}

	private HTTPResponse getResponse() {

		StringBuilder builder = new StringBuilder();
		builder.append('{');
		JSONStringBuilder.appendJSON(builder, ThermostatModel.TEMPERATURE, this.thermostat.getModel().serialize())
				.append(',');
		JSONStringBuilder.appendString(builder, CLOUDS).append(':').append('[');
		boolean isFirst = true;
		for (CloudConnector cloud : this.thermostat.getCloudConnectors()) {
			String name = cloud.getName();
			if (name != null && cloud.isConnected()) {
				if (!isFirst) {
					builder.append(',');
				}
				isFirst = false;
				JSONStringBuilder.appendString(builder, name);
			}
		}
		builder.append(']').append('}');
		HTTPResponse response = new HTTPResponse(builder.toString());
		response.addHeaderField(HTTPConstants.FIELD_CONTENT_TYPE, APPLICATION_JSON);
		response.setStatus(HTTPConstants.HTTP_STATUS_OK);
		return response;
	}
}
