/*
 * Java
 *
 * Copyright 2018-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package com.microej.profiling.internal.snapshot;

import com.microej.profiling.internal.Strings;
import com.microej.profiling.internal.snapshot.Snapshot.Type;
import com.microej.profiling.profiler.ComparableProfiler;

/**
 * A wrapper to print a snapshot.
 */
public class PrintableSnapshot {

	private final Snapshot start;
	private Snapshot end;
	private Type type;
	private final int level;

	/**
	 * Instantiates a {@link PrintableSnapshot}.
	 *
	 * @param start
	 *            the snapshot taken as a reference.
	 * @param level
	 *            the level of the stack.
	 */
	public PrintableSnapshot(Snapshot start, int level) {
		this.start = start;
		this.level = level;
	}

	/**
	 * Gets the base snapshot.
	 *
	 * @return the snapshot.
	 */
	public Snapshot getStart() {
		return this.start;
	}

	/**
	 * Sets the end.
	 *
	 * @param end
	 *            the {@link Snapshot}.
	 * @param type
	 *            the type of snapshot.
	 */
	public void setEnd(Snapshot end, Type type) {
		this.end = end;
		this.type = type;
	}

	/**
	 * Sets the end value.
	 *
	 * @param end
	 *            the {@link Snapshot}.
	 */
	public void setEnd(Snapshot end) {
		this.end = end;
		this.type = end.getType();
	}

	/**
	 * Gets the end snapshot.
	 *
	 * @return the endSnapshot.
	 *
	 */
	public Snapshot getEnd() {
		return this.end;
	}

	@Override
	public String toString() {
		StringBuilder builder = new StringBuilder();
		String endTag = (this.end != null) ? this.end.getTag() : null;
		String startTag = this.start.getTag();
		String tag = (this.getType() == Type.LAP) ? endTag : startTag;
		if (tag == null) {
			tag = Strings.NO_TAG;
		}
		builder.append(tag + Strings.TAB);
		ComparableProfiler profiler = this.start.getProfiler();
		if (this.end != null) {
			if (this.getType() == Type.LAP || endTag == startTag || (endTag != null && endTag.equals(startTag))) {
				long diff = profiler.compare(this.start.getSample(), this.end.getSample());
				builder.append(profiler.printValue(diff));
			} else {
				builder.append(Strings.DIFFERENT_TAG + startTag + Strings.NOT_EQUAL + endTag + Strings.CLOSE_BRACKET);
			}
		} else {
			builder.append(Strings.NOT_EXIT);
		}

		return builder.toString();
	}

	/**
	 * Gets the level.
	 *
	 * @return the level.
	 */
	public int getLevel() {
		return this.level;
	}

	/**
	 * Gets the type.
	 *
	 * @return the type.
	 */
	public Type getType() {
		return this.type;
	}
}
