/*
 * Java
 *
 * Copyright 2018-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package com.microej.wadapps.launcher.util;

import org.eclipse.core.runtime.IProgressMonitor;

import com.microej.wadapps.kernel.TrustService;

import ej.components.dependencyinjection.ServiceLoaderFactory;
import ej.wadapps.admin.Application;
import ej.wadapps.admin.Application.State;
import ej.wadapps.admin.ApplicationOperationException;
import ej.wadapps.admin.ApplicationsManager;
import ej.wadapps.app.Activity;
import ej.wadapps.app.BackgroundService;
import ej.wadapps.management.ActivitiesList;
import ej.wadapps.management.ActivitiesScheduler;
import ej.wadapps.management.BackgroundServicesList;
import ej.wadapps.management.LaunchersList;

/**
 * Class gathering util functions for Launcher.
 */
public final class LauncherHelper {

	private static final int ACTIVITY_SHOW_TIMEOUT = 5000;

	private LauncherHelper() {
		// Prevents instantiation
	}

	/**
	 * Checks whether the main launcher is a given application.
	 *
	 * @param application
	 *            the application.
	 * @throws NullPointerException
	 *             if application is <code>null</code>.
	 * @return <code>true</code> if the given application is the current launcher.
	 */
	public static boolean isCurrentLauncher(Application application) throws NullPointerException {
		if (application == null) {
			throw new NullPointerException();
		}
		ActivitiesScheduler activitiesScheduler = ServiceLoaderFactory.getServiceLoader()
				.getService(ActivitiesScheduler.class);
		Activity launcher = (activitiesScheduler == null) ? null : activitiesScheduler.getLauncher();
		ApplicationsManager am = ServiceLoaderFactory.getServiceLoader().getService(ApplicationsManager.class);
		return am == null || launcher == null || application.equals(am.getApplication(launcher));
	}

	/**
	 * Starts an application if necessary and shows its activity
	 *
	 * @param application
	 *            the application.
	 * @throws IllegalStateException
	 *             if the application state is not {@link State#INSTALLED} or {@link State#STARTED} state.
	 * @throws SecurityException
	 *             if a security manager exits and if the caller does not have the appropriate permissions.
	 * @throws ApplicationOperationException
	 *             If any unexpected errors occur while starting this application.
	 * @throws NullPointerException
	 *             if application is <code>null</code>.
	 */
	public static void startAndShow(Application application)
			throws IllegalStateException, SecurityException, ApplicationOperationException, NullPointerException {
		start(application);
		show(application);
	}

	/**
	 * Starts an application if necessary.
	 *
	 * @param application
	 *            the application.
	 * @throws IllegalStateException
	 *             if the application state is not {@link State#INSTALLED} or {@link State#STARTED} state.
	 * @throws SecurityException
	 *             if a security manager exits and if the caller does not have the appropriate permissions.
	 * @throws ApplicationOperationException
	 *             If any unexpected errors occur while starting this application.
	 * @throws NullPointerException
	 *             if application is <code>null</code>.
	 */
	public static void start(Application application) throws ApplicationOperationException {
		if (application == null) {
			throw new NullPointerException();
		}
		if (application.getState() == State.INSTALLED) {
			application.start();
		} else if (application.getState() != State.STARTED) {
			throw new IllegalStateException();
		}
	}

	/**
	 * Starts an application if necessary and shows its activity
	 *
	 * @param application
	 *            the application.
	 * @param monitor
	 *            the monitor to use to follow the start.
	 * @throws IllegalStateException
	 *             if the application state is not {@link State#INSTALLED} or {@link State#STARTED} state.
	 * @throws SecurityException
	 *             if a security manager exits and if the caller does not have the appropriate permissions.
	 * @throws ApplicationOperationException
	 *             If any unexpected errors occur while starting this application.
	 * @throws NullPointerException
	 *             if application is <code>null</code>.
	 */
	public static void startAndShow(Application application, IProgressMonitor monitor)
			throws IllegalStateException, SecurityException, ApplicationOperationException, NullPointerException {
		start(application, monitor);
		show(application);
	}

	/**
	 * Starts an application if necessary.
	 *
	 * @param application
	 *            the application.
	 * @param monitor
	 *            the monitor to use to follow the start.
	 * @throws IllegalStateException
	 *             if the application state is not {@link State#INSTALLED} or {@link State#STARTED} state.
	 * @throws SecurityException
	 *             if a security manager exits and if the caller does not have the appropriate permissions.
	 * @throws ApplicationOperationException
	 *             If any unexpected errors occur while starting this application.
	 * @throws NullPointerException
	 *             if application is <code>null</code>.
	 */
	public static void start(Application application, IProgressMonitor monitor) throws ApplicationOperationException {
		if (application == null) {
			throw new NullPointerException();
		}
		try {
			if (application.getState() == State.INSTALLED) {
				application.start(monitor);
			} else if (application.getState() != State.STARTED) {
				throw new IllegalStateException();
			}
		} finally {
			if (monitor != null) {
				monitor.done();
			}
		}
	}

	/**
	 * Shows an application. If the application is not yet started, the method returns after starting an asynchronous
	 * process that waits for an activity of this application to be registered, for at least 5000 milliseconds.
	 *
	 * @param application
	 *            the application.
	 *
	 * @throws NullPointerException
	 *             if application is <code>null</code>.
	 */
	public static void show(final Application application) throws NullPointerException {
		show(application, ACTIVITY_SHOW_TIMEOUT);
	}

	/**
	 * Shows an application. If the application is not yet started, the method returns after starting an asynchronous
	 * process that waits for an activity of this application to be registered, for at least the specified amount of
	 * milliseconds.
	 *
	 * @param application
	 *            the application.
	 * @param timeout
	 *            the maximum amount of time to wait for an activity of this application to be registered before
	 *            cancelling this operation
	 *
	 * @throws NullPointerException
	 *             if application is <code>null</code>.
	 */
	public static void show(final Application application, int timeout) throws NullPointerException {
		if (application == null) {
			throw new NullPointerException();
		}

		new ApplicationShower(application).show(timeout);
	}

	/**
	 * Checks whether an application is resident.
	 *
	 * @param application
	 *            the application.
	 * @throws NullPointerException
	 *             if application is <code>null</code>.
	 * @return <code>true</code> if the application is resident.
	 */
	public static boolean isResident(Application application) throws NullPointerException {
		if (application == null) {
			throw new NullPointerException();
		}
		boolean isResidentApp = false;
		TrustService trustService = ServiceLoaderFactory.getServiceLoader().getService(TrustService.class);
		if (trustService != null) {
			isResidentApp = trustService.isResident(application);
		} else {
			String identifier = application.getIdentifier();
			isResidentApp = identifier != null && identifier.startsWith("WADAPPS"); //$NON-NLS-1$
		}
		return isResidentApp;
	}

	/**
	 * Checks whether an application is a system app.
	 *
	 * @param application
	 *            the application.
	 * @throws NullPointerException
	 *             if application is <code>null</code>.
	 * @return <code>true</code> if the application is a system app.
	 */
	public static boolean isSystemApp(Application application) throws NullPointerException {
		if (application == null) {
			throw new NullPointerException();
		}
		boolean isSystemApp = false;
		TrustService trustService = ServiceLoaderFactory.getServiceLoader().getService(TrustService.class);
		if (trustService != null) {
			isSystemApp = trustService.isSystemApp(application);
		} else {
			String identifier = application.getIdentifier();
			isSystemApp = identifier != null && identifier.startsWith("WADAPPS"); //$NON-NLS-1$
		}
		return isSystemApp;
	}

	/**
	 * Checks whether an application has some background services.
	 *
	 * @param application
	 *            the application.
	 * @throws NullPointerException
	 *             if application is <code>null</code>.
	 * @return <code>true</code> if the application has at least one background service.
	 */
	public static boolean hasService(final Application application) throws NullPointerException {
		if (application == null) {
			throw new NullPointerException();
		}
		boolean hasService = false;
		BackgroundServicesList backgroundServicesList = ServiceLoaderFactory.getServiceLoader()
				.getService(BackgroundServicesList.class);
		ApplicationsManager am = ServiceLoaderFactory.getServiceLoader().getService(ApplicationsManager.class);
		// Do not stop applications with BG services
		if (am != null && backgroundServicesList != null) {
			for (BackgroundService backgroundService : backgroundServicesList.getBackgroundServices()) {
				if (application.equals(am.getApplication(backgroundService))) {
					hasService = true;
					break;
				}
			}
		}
		return hasService;
	}

	/**
	 * Checks whether an application has some an activity.
	 *
	 * @param application
	 *            the application.
	 * @throws NullPointerException
	 *             if application is <code>null</code>.
	 * @return <code>true</code> if the application has at least one activity.
	 */
	public static boolean hasActivity(Application application) throws NullPointerException {
		return getFirstActivity(application) != null;
	}

	/**
	 * Gets the first activity of an application.
	 *
	 * @param application
	 *            the application.
	 * @throws NullPointerException
	 *             if application is <code>null</code>.
	 * @return the first activity if found, <code>null</code> otherwise.
	 */
	public static Activity getFirstActivity(Application application) throws NullPointerException {
		if (application == null) {
			throw new NullPointerException();
		}
		Activity firstActivity = null;
		ActivitiesList activitiesList = ServiceLoaderFactory.getServiceLoader().getService(ActivitiesList.class);
		ApplicationsManager am = ServiceLoaderFactory.getServiceLoader().getService(ApplicationsManager.class);
		if (am != null && activitiesList != null) {
			for (Activity activity : activitiesList.getActivities()) {
				if (application.equals(am.getApplication(activity))) {
					firstActivity = activity;
					break;
				}
			}
		}
		return firstActivity;
	}

	/**
	 * Checks whether an application is a launcher.
	 *
	 * @param application
	 *            the application.
	 * @throws NullPointerException
	 *             if application is <code>null</code>.
	 * @return <code>true</code> if the application is the launcher.
	 */
	public static boolean isLauncher(final Application application) throws NullPointerException {
		if (application == null) {
			throw new NullPointerException();
		}
		boolean isLauncher = false;
		LaunchersList launchersList = ServiceLoaderFactory.getServiceLoader().getService(LaunchersList.class);
		ApplicationsManager am = ServiceLoaderFactory.getServiceLoader().getService(ApplicationsManager.class);
		if (am != null && launchersList != null) {
			for (Activity activity : launchersList.getLaunchers()) {
				if (application.equals(am.getApplication(activity))) {
					isLauncher = true;
					break;
				}
			}
		}
		return isLauncher;
	}

	/**
	 * Counts the number of launchers available.
	 *
	 * @return the number of launchers available.
	 */
	public static int countLaunchers() {
		int launcherCount = 0;
		LaunchersList launchersList = ServiceLoaderFactory.getServiceLoader().getService(LaunchersList.class);
		if (launchersList != null) {
			launcherCount = launchersList.getLaunchers().length;
		}
		return launcherCount;
	}

	/**
	 * Gets the application that owns the given background service.
	 *
	 * @param bg
	 *            the background service.
	 * @return the application that owns this object.
	 * @throws IllegalArgumentException
	 *             when the background service is null.
	 */
	public static Application getApplication(BackgroundService bg) throws IllegalArgumentException {
		return getApp(bg);
	}

	/**
	 * Gets the application that owns an activity.
	 *
	 * @param activity
	 *            the activity.
	 * @return the application that owns this object.
	 * @throws IllegalArgumentException
	 *             when the activity is null.
	 */
	public static Application getApplication(Activity activity) throws IllegalArgumentException {
		return getApp(activity);
	}

	/**
	 * Gets the application that owns the given object.
	 *
	 * @param o
	 *            the object.
	 * @return the application that owns this object or <code>null</code> if the application has been stopped.
	 * @throws IllegalArgumentException
	 *             when the object is null.
	 */
	private static Application getApp(Object o) throws IllegalArgumentException {
		Application app = null;
		ApplicationsManager am = ServiceLoaderFactory.getServiceLoader().getService(ApplicationsManager.class);
		if (am != null) {
			app = am.getApplication(o);
		}
		return app;
	}

	/**
	 * Set the next launcher as default launcher.
	 */
	public static void setNextLauncher() {
		LaunchersList launchersList = ServiceLoaderFactory.getServiceLoader().getService(LaunchersList.class);
		ApplicationsManager am = ServiceLoaderFactory.getServiceLoader().getService(ApplicationsManager.class);
		ActivitiesScheduler scheduler = ServiceLoaderFactory.getServiceLoader().getService(ActivitiesScheduler.class);
		if (am != null && scheduler != null && launchersList != null) {
			Application myApp = am.getCurrentApplication();
			for (Activity activity : launchersList.getLaunchers()) {
				if (am.getApplication(activity) != myApp) {
					scheduler.setLauncher(activity);
					break;
				}
			}
		}
	}

	/**
	 * Gets whether the application is the current application.
	 *
	 * @param app
	 *            the Application to check
	 *
	 * @return <code>true</code> if the application is the current application.
	 */
	public static boolean isCurrentApplication(Application app) {
		boolean isCurrent = false;
		ApplicationsManager am = ServiceLoaderFactory.getServiceLoader().getService(ApplicationsManager.class);
		if (am != null && app != null) {
			isCurrent = app.equals(am.getCurrentApplication());
		}
		return isCurrent;
	}
}
