/*
 * Java
 *
 * Copyright 2015-2018 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 *
 */
package com.microej.wadapps.metadata;

import java.io.IOException;

import ej.wadapps.management.ApplicationMetadata;
import ej.wadapps.management.ExtraMetadata;
import ej.wadapps.management.ImageInfo;

/**
 * Gives access to an application meta-data.
 * <p>
 * Some meta-data like the name, version and description are directly available,
 * but some other like the icon and screenshot lists are lazy retrieved.
 * <p>
 * Two meta-data are equals if they are the same name and the same version.
 */
public class DefaultApplicationMetadata implements ApplicationMetadata {

	/** The name. */
	private String name;

	/** The version. */
	private String version;

	/** The description. */
	private String description;

	/** The icons. */
	private final ImageInfo[] icons;

	/** The screenshots. */
	private final ImageInfo[] screenshots;

	/** The extra metadatas. */
	private final ExtraMetadata[] extraMetadatas;

	/** The application ID. */
	private final String applicationID;

	/**
	 * Instantiates a new application metadata.
	 *
	 * @param applicationID
	 *            the application id.
	 *
	 * @param icons
	 *            the application icons.
	 *
	 * @param screenshots
	 *            the application screenshots.
	 *
	 * @param extraMetadatas
	 *            the application extra metadatas.
	 */
	public DefaultApplicationMetadata(String applicationID, ImageInfo[] icons, ImageInfo[] screenshots,
			ExtraMetadata[] extraMetadatas) {
		this.applicationID = applicationID;
		this.icons = icons;
		this.screenshots = screenshots;
		this.extraMetadatas = extraMetadatas;
	}

	/**
	 * Sets the name.
	 *
	 * @param name
	 *            the name to set
	 */
	public void setName(String name) {
		this.name = name;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see ej.wadapps.management.ApplicationMetadata#getName()
	 */
	@Override
	public String getName() {
		return this.name;
	}

	/**
	 * Sets the version.
	 *
	 * @param version
	 *            the version to set
	 */
	public void setVersion(String version) {
		this.version = version;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see ej.wadapps.management.ApplicationMetadata#getVersion()
	 */
	@Override
	public String getVersion() {
		return this.version;
	}

	/**
	 * Sets the description.
	 *
	 * @param description
	 *            the description to set
	 */
	public void setDescription(String description) {
		this.description = description;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see ej.wadapps.management.ApplicationMetadata#getDescription()
	 */
	@Override
	public String getDescription() {
		return this.description;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}

		if (!(obj instanceof ApplicationMetadata)) {
			return false;
		}

		ApplicationMetadata metadata = (ApplicationMetadata) obj;
		return this.applicationID.equals(metadata.getApplicationIdentifier());
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		return Math.abs(this.applicationID.hashCode());
	}

	/**
	 * Gets the application identifier.
	 *
	 * @return the application identifier
	 */
	@Override
	public String getApplicationIdentifier() {
		return this.applicationID;
	}

	@Override
	public ImageInfo[] getIconsList() throws IOException {
		return this.icons;
	}

	@Override
	public ImageInfo[] getScreenshotsList() throws IOException {
		return this.screenshots;
	}

	@Override
	public ExtraMetadata[] getExtraMetadatasList() throws IOException {
		return this.extraMetadatas;
	}

}
