/*
 * Java
 *
 * Copyright 2015-2021 MicroEJ Corp. All rights reserved.
 * MicroEJ Corp. PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.microej.wadapps.rcommand;

import java.io.InputStream;

import ej.components.dependencyinjection.ServiceLoaderFactory;
import ej.rcommand.CommandReader;
import ej.rcommand.CommandSender;
import ej.wadapps.admin.Application;
import ej.wadapps.admin.ApplicationOperationException;
import ej.wadapps.admin.ApplicationsManager;
import ej.wadapps.management.ApplicationMetadata;
import ej.wadapps.management.ApplicationMetadataProvider;
import ej.wadapps.management.ImageInfo;

/**
 * A command allowing to retrieve the metadata of an application.
 */
@SuppressWarnings("nls")
public class GetApplicationMetadataCommand extends AbstractCommand {

	private final ApplicationsManager applicationsManager;

	public GetApplicationMetadataCommand() {
		this.applicationsManager = ServiceLoaderFactory.getServiceLoader().getService(ApplicationsManager.class);
	}

	@Override
	public String getManagedCommand() {
		return CommandConstants.GET_APPLICATION_METADATA;
	}

	@Override
	protected void commandReceived(CommandReader commandReader, CommandSender commandSender) {
		try {
			String applicationId = commandReader.readString();
			int preferredIconSize = commandReader.readInt();

			ApplicationMetadata metadata = getApplicationMetadata(applicationId);
			if (metadata != null) {
				ImageInfo preferredIcon = getPreferredIcon(metadata.getIconsList(), preferredIconSize);
				commandSender.startCommand(CommandConstants.GET_APPLICATION_METADATA_SUCCESS);
				commandSender.sendString(metadata.getApplicationIdentifier());
				commandSender.sendString(metadata.getVersion());
				commandSender.sendString(metadata.getName());
				commandSender.sendString(metadata.getDescription());

				if (preferredIcon != null) {
					try (InputStream icon = preferredIcon.getImage()) {
						if (icon != null) {
							commandSender.sendBoolean(true);
							commandSender.sendString(preferredIcon.getIdentifier());
							commandSender.sendInt(preferredIcon.getWidth());
							commandSender.sendInt(preferredIcon.getHeight());
							commandSender.sendInputStream(icon);
						} else {
							commandSender.sendBoolean(false);
						}
					}
				} else {
					commandSender.sendBoolean(false);
				}
			} else {
				commandSender.startCommand(CommandConstants.GET_APPLICATION_METADATA_ERROR);
				commandSender.sendString("No metadata available for the application");
			}

			commandSender.flushCommand();
		} catch (Exception e) {
			commandSender.startCommand(CommandConstants.GET_APPLICATION_METADATA_ERROR);
			commandSender.sendString(e.getMessage());
			commandSender.flushCommand();
		}
	}

	private ApplicationMetadata getApplicationMetadata(String applicationId)
			throws IllegalStateException, SecurityException, ApplicationOperationException {
		ApplicationMetadataProvider applicationMetadataProvider = ServiceLoaderFactory.getServiceLoader()
				.getService(ApplicationMetadataProvider.class);

		if (applicationMetadataProvider == null) {
			return null;
		}

		for (Application application : this.applicationsManager.getInstalledApplications()) {
			if (application.getIdentifier().equals(applicationId)) {
				return applicationMetadataProvider.getApplicationMetadata(application);
			}
		}
		return null;
	}

	// We try to get the biggest icon smaller than the preferred size.
	private static ImageInfo getPreferredIcon(ImageInfo[] icons, int preferredSize) {
		if (icons.length == 0) {
			return null;
		}
		if (preferredSize < 0) {
			return null;
		}

		// return the first one, by default
		int index = 0;
		int currentScore = preferredSize - icons[index].getWidth();

		for (int i = 1; i < icons.length; i++) {
			int score = preferredSize - icons[i].getWidth();
			if ((currentScore > 0) && (score > 0)) {
				currentScore = Math.min(currentScore, score);
			} else if (currentScore <= 0 && score <= 0) {
				currentScore = Math.max(currentScore, score);
			}
			index = currentScore == score ? i : index;
		}
		return icons[index];
	}
}
