/*
 * Java
 *
 * Copyright 2015-2021 MicroEJ Corp. All rights reserved.
 * MicroEJ Corp. PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.microej.wadapps.rcommand;

import java.io.IOException;
import java.io.InputStream;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.microej.wadapps.metadata.DefaultApplicationMetadata;

import ej.components.dependencyinjection.ServiceLoader;
import ej.components.dependencyinjection.ServiceLoaderFactory;
import ej.rcommand.CommandReader;
import ej.rcommand.CommandSender;
import ej.wadapps.admin.Application;
import ej.wadapps.admin.Application.State;
import ej.wadapps.admin.ApplicationOperationException;
import ej.wadapps.admin.ApplicationsManager;
import ej.wadapps.management.ApplicationMetadataProvider;
import ej.wadapps.management.ExtraMetadata;
import ej.wadapps.management.ImageInfo;

/**
 *
 */
public class InstallCommand extends AbstractCommand {
	private final ApplicationsManager applicationsManager;

	/** The logger. */
	private final Logger logger;

	/**
	 *
	 */
	public InstallCommand() {
		this.applicationsManager = ServiceLoaderFactory.getServiceLoader().getService(ApplicationsManager.class);
		this.logger = Logger.getLogger(InstallCommand.class.getSimpleName());
	}

	@Override
	public String getManagedCommand() {
		return CommandConstants.PUSH_APPLICATION;
	}

	@Override
	protected void commandReceived(CommandReader commandReader, CommandSender commandSender) {

		// read data
		try {
			final String applicationName = commandReader.readString();
			String applicationVersion = "1.0.0";

			this.logger.info("Receive an application " + applicationName); //$NON-NLS-1$

			deleteOldApplication(applicationName, applicationVersion);

			ImageInfo icon = new ImageInfo() {
				@Override
				public int getWidth() {
					return 32;
				}

				@Override
				public InputStream getImage() throws IOException {
					return InstallCommand.class.getResourceAsStream("/images/application-icon.png"); //$NON-NLS-1$
				}

				@Override
				public String getIdentifier() {
					return applicationName;
				}

				@Override
				public int getHeight() {
					return 32;
				}
			};

			DefaultApplicationMetadata metadata = new DefaultApplicationMetadata(applicationName,
					new ImageInfo[] { icon }, new ImageInfo[0], new ExtraMetadata[0]);
			metadata.setName(applicationName);
			metadata.setVersion(applicationVersion);
			metadata.setDescription(applicationName);
			try (InputStream stream = commandReader.readByteArrayAsInputStream()) {
				this.logger.info("Install the application " + applicationName); //$NON-NLS-1$
				final Application application = this.applicationsManager.install(stream, applicationName,
						applicationVersion);
				this.logger.info("Install the application Done " + applicationName); //$NON-NLS-1$

				ServiceLoader serviceLoader = ServiceLoaderFactory.getServiceLoader();
				ApplicationMetadataProvider applicationMetadataProvider = serviceLoader
						.getService(ApplicationMetadataProvider.class);
				if (applicationMetadataProvider != null) {
					this.logger.info("Set the application metadata"); //$NON-NLS-1$
					applicationMetadataProvider.setApplicationMetadata(application, metadata);
					this.logger.info("Set the application metadata done"); //$NON-NLS-1$
				}

				// Launch the application
				this.logger.info("Start the application " + applicationName); //$NON-NLS-1$
				application.start();
				commandSender.startCommand(CommandConstants.PUSH_APPLICATION_SUCCESS);
				commandSender.flushCommand();
			}
		} catch (UnsupportedOperationException | SecurityException | ApplicationOperationException | IOException e) {
			this.logger.log(Level.SEVERE, e.getMessage(), e);
			commandSender.startCommand(CommandConstants.PUSH_APPLICATION_ERROR);
			if (e.getMessage() != null) {
				commandSender.sendString(e.getMessage());
			} else {
				commandSender.sendString("board internal error"); //$NON-NLS-1$
			}
			commandSender.flushCommand();
		}
	}

	/**
	 * @param applicationName
	 * @param applicationVersion
	 * @throws ApplicationOperationException
	 * @throws SecurityException
	 * @throws IllegalStateException
	 */
	private void deleteOldApplication(String applicationName, String applicationVersion)
			throws IllegalStateException, SecurityException, ApplicationOperationException {
		for (Application application : this.applicationsManager.getInstalledApplications()) {
			if (application.getIdentifier().equals(applicationName)) {
				State state = application.getState();
				if (state.equals(State.STARTED)) {
					this.logger.info("Stop an old application " + applicationName); //$NON-NLS-1$
					application.stop();
				}

				this.logger.info("Uninstall an old application " + applicationName); //$NON-NLS-1$
				application.uninstall();
			}
		}
	}
}
