/*
 * Java
 *
 * Copyright 2015-2021 MicroEJ Corp. All rights reserved.
 * MicroEJ Corp. PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.microej.wadapps.rcommand;

import java.io.IOException;
import java.io.InputStream;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.microej.wadapps.metadata.DefaultApplicationMetadata;

import ej.components.dependencyinjection.ServiceLoader;
import ej.components.dependencyinjection.ServiceLoaderFactory;
import ej.rcommand.CommandReader;
import ej.rcommand.CommandSender;
import ej.wadapps.admin.Application;
import ej.wadapps.admin.Application.State;
import ej.wadapps.admin.ApplicationOperationException;
import ej.wadapps.admin.ApplicationsManager;
import ej.wadapps.management.ApplicationMetadataProvider;
import ej.wadapps.management.ExtraMetadata;
import ej.wadapps.management.ImageInfo;
import ej.wadapps.storage.Storage;

/**
 * A command allowing to install an application with more parameters.
 */
public class InstallCommandWithMetadata extends AbstractCommand {

	private static final int DEFAULT_ICON_WIDTH = 32;
	private static final int DEFAULT_ICON_HEIGHT = 32;
	private static final String DEFAULT_ICON_PATH = "/images/application-icon.png";
	private static final String TEMPORARY_ICON_ID = "tmp-icon";

	private final ApplicationsManager applicationsManager;
	private final Logger logger;

	public InstallCommandWithMetadata() {
		this.applicationsManager = ServiceLoaderFactory.getServiceLoader().getService(ApplicationsManager.class);
		this.logger = Logger.getLogger(InstallCommandWithMetadata.class.getSimpleName());
	}

	@Override
	public String getManagedCommand() {
		return CommandConstants.INSTALL_APPLICATION;
	}

	@Override
	protected void commandReceived(CommandReader commandReader, CommandSender commandSender) {
		try {
			String applicationId = commandReader.readString();
			String applicationName = commandReader.readString();
			String applicationVersion = commandReader.readString();
			String applicationDescription = commandReader.readString();
			boolean withIcon = commandReader.readBoolean();
			this.logger.info("Receive an application " + applicationName); //$NON-NLS-1$

			deleteOldApplication(applicationId, applicationName, applicationVersion);

			ServiceLoader serviceLoader = ServiceLoaderFactory.getServiceLoader();
			ApplicationMetadataProvider applicationMetadataProvider = serviceLoader
					.getService(ApplicationMetadataProvider.class);
			DefaultApplicationMetadata metadata = null;
			Storage storage = ServiceLoaderFactory.getServiceLoader().getService(Storage.class);

			// Needs a icon.
			if (applicationMetadataProvider != null) {
				int iconWidth;
				int iconHeight;
				String iconIdentifier;
				InputStream iconImage;
				if (withIcon && storage != null) {
					iconWidth = commandReader.readInt();
					iconHeight = commandReader.readInt();
					iconIdentifier = commandReader.readString();

					storage.store(TEMPORARY_ICON_ID, commandReader.readInputStream());
					iconImage = new IconOneShotInputStream(storage.load(TEMPORARY_ICON_ID));
				} else {
					// IconInfoProvider iconInfoProvider =
					// ServiceLoaderFactory.getServiceLoader()
					// .getService(IconInfoProvider.class);
					// iconWidth = iconInfoProvider.getAcceptedIconWidth();
					// iconHeight = iconInfoProvider.getAcceptedIconHeight();
					iconWidth = DEFAULT_ICON_WIDTH;
					iconHeight = DEFAULT_ICON_HEIGHT;
					iconIdentifier = applicationId;
					iconImage = InstallCommandWithMetadata.class.getResourceAsStream(DEFAULT_ICON_PATH);
				}
				BasicImageInfo imageInfo = new BasicImageInfo();
				imageInfo.setIdentifier(iconIdentifier);
				imageInfo.setWidth(iconWidth);
				imageInfo.setHeight(iconHeight);
				imageInfo.setImage(iconImage);

				metadata = new DefaultApplicationMetadata(applicationId, new ImageInfo[] { imageInfo },
						new ImageInfo[0], new ExtraMetadata[0]);
				metadata.setName(applicationName);
				metadata.setVersion(applicationVersion);
				metadata.setDescription(applicationDescription);
			}

			if (withIcon) {
				// Skips unused parameters.
				commandReader.readInt();
				commandReader.readInt();
				commandReader.readString();
				// Consumes the icon image stream.
				InputStream image = commandReader.readInputStream();
				int read = image.read();
				while (read != -1) {
					read = image.read();
				}
			}

			try (InputStream applicationBinary = commandReader.readInputStream()) {
				this.logger.info("Install the application " + applicationName); //$NON-NLS-1$
				final Application application = this.applicationsManager.install(applicationBinary, applicationId,
						applicationVersion);
				this.logger.info("Install the application Done " + applicationName); //$NON-NLS-1$

				if (applicationMetadataProvider != null) {
					this.logger.info("Set the application metadata"); //$NON-NLS-1$
					applicationMetadataProvider.setApplicationMetadata(application, metadata);
					this.logger.info("Set the application metadata done"); //$NON-NLS-1$
				}

				commandSender.startCommand(CommandConstants.INSTALL_APPLICATION_SUCCESS);
				commandSender.flushCommand();
			}
		} catch (UnsupportedOperationException | SecurityException | ApplicationOperationException | IOException e) {
			this.logger.log(Level.SEVERE, e.getMessage(), e);
			commandSender.startCommand(CommandConstants.INSTALL_APPLICATION_ERROR);
			if (e.getMessage() != null) {
				commandSender.sendString(e.getMessage());
			} else {
				commandSender.sendString("board internal error"); //$NON-NLS-1$
			}
			commandSender.flushCommand();
		}
	}

	private void deleteOldApplication(String applicationId, String applicationName, String applicationVersion)
			throws IllegalStateException, SecurityException, ApplicationOperationException {
		for (Application application : this.applicationsManager.getInstalledApplications()) {
			if (application.getIdentifier().equals(applicationId)) {
				State state = application.getState();
				if (state.equals(State.STARTED)) {
					this.logger.info("Stop an old application " + applicationName); //$NON-NLS-1$
					application.stop();
				}

				this.logger.info("Uninstall an old application " + applicationName); //$NON-NLS-1$
				application.uninstall();
			}
		}
	}

	private class BasicImageInfo implements ImageInfo {

		private String identifier;
		private int width;
		private int height;
		private InputStream image;

		@Override
		public String getIdentifier() {
			return this.identifier;
		}

		/**
		 * Sets the identifier.
		 *
		 * @param identifier the identifier to set.
		 */
		public void setIdentifier(String identifier) {
			this.identifier = identifier;
		}

		@Override
		public int getWidth() {
			return this.width;
		}

		/**
		 * Sets the width.
		 *
		 * @param width the width to set.
		 */
		public void setWidth(int width) {
			this.width = width;
		}

		@Override
		public int getHeight() {
			return this.height;
		}

		/**
		 * Sets the height.
		 *
		 * @param height the height to set.
		 */
		public void setHeight(int height) {
			this.height = height;
		}

		@Override
		public InputStream getImage() throws IOException {
			return this.image;
		}

		/**
		 * Sets the image.
		 *
		 * @param image the image to set.
		 */
		public void setImage(InputStream image) {
			this.image = image;
		}
	}

	private class IconOneShotInputStream extends InputStream {

		private final InputStream in;

		public IconOneShotInputStream(InputStream in) {
			this.in = in;
		}

		@Override
		public int available() throws IOException {
			return this.in.available();
		}

		@Override
		public int read() throws IOException {
			return this.in.read();
		}

		@Override
		public int read(byte[] b, int off, int len) throws IOException {
			return this.in.read(b, off, len);
		}

		@Override
		public long skip(long n) throws IOException {
			return this.in.skip(n);
		}

		@Override
		public boolean markSupported() {
			return this.in.markSupported();
		}

		@Override
		public synchronized void reset() throws IOException {
			this.in.reset();
		}

		@Override
		public synchronized void mark(int readlimit) {
			this.in.mark(readlimit);
		}

		@Override
		public void close() throws IOException {
			this.in.close();
			ServiceLoaderFactory.getServiceLoader().getService(Storage.class).remove(TEMPORARY_ICON_ID);
		}
	}
}
