/*
 * Java
 *
 * Copyright 2018-2021 MicroEJ Corp. All rights reserved.
 * MicroEJ Corp. PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.microej.wadapps.rcommand;

import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;

import com.microej.wadapps.metadata.DefaultApplicationMetadata;
import com.microej.wadapps.rcommand.util.BasicImageInfo;

import ej.components.dependencyinjection.ServiceLoaderFactory;
import ej.rcommand.CommandReader;
import ej.rcommand.CommandSender;
import ej.wadapps.admin.Application;
import ej.wadapps.admin.ApplicationsManager;
import ej.wadapps.management.ApplicationMetadataProvider;
import ej.wadapps.management.ExtraMetadata;
import ej.wadapps.management.ImageInfo;
import ej.wadapps.storage.Storage;

/**
 *
 */
@SuppressWarnings("nls")
public class SetApplicationMetadataCommand extends AbstractCommand {
	private final ApplicationsManager applicationsManager;
	private final Logger logger;

	/**
	 * Constructs an UpdateMetadataCommand.
	 */
	public SetApplicationMetadataCommand() {
		this.applicationsManager = ServiceLoaderFactory.getServiceLoader().getService(ApplicationsManager.class);
		this.logger = Logger.getLogger(InstallApplicationCommand.class.getSimpleName());
	}

	@Override
	protected String getManagedCommand() {
		return CommandConstants.SET_METADATA;
	}

	@Override
	protected void commandReceived(CommandReader commandReader, CommandSender commandSender) {
		try {
			/* int commandVersion = */ commandReader.readInt();
			String applicationId = commandReader.readString();
			String applicationName = commandReader.readString();

			Application application = searchForApplication(applicationId);

			if (application == null) {
				this.logger.warning("Application " + applicationId + " (" + applicationName
						+ ") is not installed; metadata ignored");
				commandSender.startCommand(CommandConstants.SET_METADATA_ERROR);
				commandSender
				.sendString("No application found with id " + applicationId + " (" + applicationName + ")");
				commandSender.flushCommand();
				return;
			}
			ApplicationMetadataProvider applicationMetadataProvider = ServiceLoaderFactory.getServiceLoader()
					.getService(ApplicationMetadataProvider.class);
			if (applicationMetadataProvider == null) {
				// If there is no applicationMetadataProvider, this command is useless.
				this.logger.warning("No implementation of ApplicationMetadataProvider found.");
				commandSender.startCommand(CommandConstants.SET_METADATA_SUCCESS);
				// non-empty message indicates command "success" with a warning
				commandSender.sendString("No ApplicationMetadataProvider found, metadata ignored.");
				commandSender.flushCommand();
				return;
			}
			String applicationDescription = commandReader.readString();
			this.logger.info("Set the application metadata for " + applicationName);

			int iconCount = commandReader.readInt();
			ImageInfo[] icons = new ImageInfo[iconCount];
			Storage storage = ServiceLoaderFactory.getServiceLoader().getService(Storage.class);

			for (int i = 0; i < iconCount; i++) {
				BasicImageInfo icon = new BasicImageInfo();
				icon.setIdentifier("tmp_icon_" + commandReader.readString());
				icon.setWidth(commandReader.readInt());
				icon.setHeight(commandReader.readInt());
				storage.store(icon.getIdentifier(), commandReader.readInputStream());
				icons[i] = icon;
			}

			DefaultApplicationMetadata metadata = new DefaultApplicationMetadata(applicationId, icons,
					new ImageInfo[0], new ExtraMetadata[0]);
			metadata.setName(applicationName);
			metadata.setDescription(applicationDescription);
			metadata.setVersion(application.getVersion());
			applicationMetadataProvider.setApplicationMetadata(application, metadata);

			this.logger.info("Set the application metadata Done");
			commandSender.startCommand(CommandConstants.SET_METADATA_SUCCESS);
			// empty message indicates full command success
			commandSender.sendString("");
			commandSender.flushCommand();
		} catch (SecurityException | IOException e) {
			this.logger.log(Level.SEVERE, e.getMessage(), e);
			commandSender.startCommand(CommandConstants.SET_METADATA_ERROR);
			if (e.getMessage() != null) {
				commandSender.sendString(e.getMessage());
			} else {
				commandSender.sendString("board internal error");
			}
			commandSender.flushCommand();
		}
	}

	/**
	 * @param applicationId
	 * @param application
	 * @return
	 */
	private Application searchForApplication(String applicationId) {
		final Application[] applications = this.applicationsManager.getInstalledApplications();
		for (Application app : applications) {
			if (applicationId.equals(app.getIdentifier())) {
				return app;
			}
		}
		return null;
	}

}
