/*
 * Java
 *
 * Copyright 2014-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.basedriver;
import ej.ecom.Device;


/**
 * Describe the capabilities of a Controller.
 */
public interface Controller extends Device {

	/**
	 * A controller state. A controller can be in one of the following states:
	 * <ul>
	 * <li>RUNNING<br>
	 * A controller that has started is in this state.<br>
	 * After apply {@link Controller#inclusionStop()} the controller is in this state.<br>
	 * After apply {@link Controller#exclusionStop()} the controller is in this state.</li>
	 * <li>INCLUSION<br>
	 * After apply {@link Controller#inclusionStart()} the controller is in this state.</li>
	 * <li>EXCLUSION<br>
	 * After apply {@link Controller#exclusionStart()} the controller is in this state.</li>
	 * </ul>
	 * 
	 * @see Controller#inclusionStart()
	 * @see Controller#inclusionStop()
	 * @see Controller#exclusionStart()
	 * @see Controller#exclusionStop()
	 */
	int RUNNING = 0;
	/**
	 * A controller in that mode can include a new device.
	 */
	int INCLUSION = 1;
	/**
	 *  A controller in that mode can exclude a new device.
	 */
	int EXCLUSION = 2;

	/**
	 * Type of exclusion mode.
	 */
	/**
	 * No physical action is required on the Node to exclude it from the network.
	 */
	int	AUTOMATIC = 0;
	/**
	 * A physical action is on the Node to exclude it from the network.
	 */
	int MANUAL = 1;

	/**
	 * The state has not been determined yet.
	 */
	int UNKNOWN = -1;

	/**
	 * Start the inclusion mode. The controller is now in INCLUSION state.
	 * During inclusion, a new Node is allowed to be included. When this Node has been included, the inclusion mode is stopped, and the controller is in RUNNING state.
	 *	
	 * @throws IllegalStateException
	 * 								 if the controller is not in  RUNNING state before this call.
	 */
	void inclusionStart();

	/**
	 * Stop the inclusion mode. Does nothing if the inclusion mode is already stopped.
	 * The controller is now in RUNNING state.
	 * 
	 * @throws IllegalStateException
	 * 								 if the controller is not in  INCLUSION state before this call.
	 * 
	 */
	void inclusionStop();

	/**
	 * Return The supported exclusion mode.
	 * 
	 * @return the supported exclusion mode
	 */
	int getExclusionMode();

	/**
	 * Starts the exclusion mode. The controller is now in EXCLUSION state.
	 * 
	 * @throws IllegalStateException
	 * 								 if the controller is not in  RUNNING state before this call.
	 */
	void exclusionStart();

	/**
	 * Stops the exclusion mode. The controller is now in RUNNING state.
	 * 
	 * @throws IllegalStateException
	 * 								 if the controller is not in  EXCLUSION state before this call.
	 */
	void exclusionStop();

	/**
	 * Start this controller.
	 * 
	 * @param listener
	 *            the listener which receive all events.
	 * @return true if the controller has been successfully started, false otherwise.
	 * @throws NullPointerException
	 *             if the given listener is null.
	 * @throws IllegalStateException
	 *             if the controller is already started.
	 */
	boolean start(ControllerListener listener);

	/**
	 * Returns true if the controller is started, false otherwise.
	 * @return true if the controller is started, false otherwise.
	 */
	boolean isStarted ();

	/**
	 * Stop this controller.
	 * 
	 * @throws IllegalStateException
	 *             if the controller is not started.
	 */
	void stop();

	/**
	 * Get the {@link ControllerListener} attached to this controller.
	 * 
	 * @return the associated controller, or 'null' if the controller is not started
	 */
	ControllerListener getListener();

	/**
	 * Returns the actual state on the controller.
	 * @return the actual state on the controller.
	 * 
	 * @see #inclusionStart()
	 * @see #inclusionStop()
	 * @see #exclusionStart()
	 * @see #exclusionStop()
	 */
	int getState ();

}
