/*
 * Java
 *
 * Copyright 2016-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.basedriver.event;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Executor;

import ej.ecom.Device;

public abstract class DeviceEventGenerator <D extends Device, E extends Event<D>>{
	private final Executor executor;
	private final Map<D, List<EventHandler<D,E>>> specificEventHandlers;
	private final List<EventHandler<D,E>> genericEventHandlers;
	/*
	 * Maybe use a ServiceLoarder to find the executor to use?
	 */
	public DeviceEventGenerator(Executor executor) {
		this.executor = executor;
		specificEventHandlers = new HashMap<>();
		genericEventHandlers = new ArrayList<>();
	}

	public void addEventHandler (EventHandler<D,E> eventHandler) {
		add(genericEventHandlers, eventHandler);
	}

	public void addEventHandler (EventHandler<D,E> eventHandler, D device) {
		List<EventHandler<D, E>> eventHandlers = specificEventHandlers.get(device);
		if (eventHandlers == null) {
			eventHandlers = new ArrayList<>();
			specificEventHandlers.put(device, eventHandlers);
		}

		add(eventHandlers, eventHandler);
	}

	public void removeEventHandler (final EventHandler<?,?> eventHandler) {
		removeEventHandlers(new EventHandlerFilter() {

			@Override
			public boolean accept(EventHandler<?, ?> next) {
				return next == eventHandler;
			}
		});
	}

	public void removeEventHandlers(EventHandlerFilter filter){
		Iterator<D> it = specificEventHandlers.keySet().iterator();
		while(it.hasNext()){
			D device = it.next();
			List<EventHandler<D, E>> list = specificEventHandlers.get(device);
			synchronized (list) {
				removeEventHandlers(list, filter);
			}
			if(list.isEmpty()){
				it.remove();// free the record
			}
		}

		synchronized (genericEventHandlers) {
			removeEventHandlers(genericEventHandlers, filter);
		}
	}

	private boolean removeEventHandlers(List<EventHandler<D, E>> list, EventHandlerFilter filter) {
		Iterator<EventHandler<D,E>> it = list.iterator();
		while(it.hasNext()){
			EventHandler<?, ?> next = it.next();
			if(filter.accept(next)){
				it.remove();
				return true;
			}
		}

		return false;
	}

	private boolean add (List<EventHandler<D, E>> eventHandlers, EventHandler<D, E> eventHandler) {
		if (! eventHandlers.contains(eventHandler)) {
			return eventHandlers.add(eventHandler);
		}

		return false;
	}


	public void postEvent (D device) {
		executor.execute(new DeviceEventGeneratorRunnable(createEvent(device), false));
	}

	public void postError (D device) {
		executor.execute(new DeviceEventGeneratorRunnable(createEvent(device), true));
	}

	protected abstract E createEvent (D device);


	class DeviceEventGeneratorRunnable implements Runnable {
		private final boolean error;
		private final E event;
		private final D device;

		public DeviceEventGeneratorRunnable (E event, boolean error) {
			this.event = event;
			this.device = event.getDevice();
			this.error = error;
		}

		@Override
		public void run() {
			List<EventHandler<D, E>> eventHandlers = specificEventHandlers.get(device);
			if (eventHandlers != null) {
				notifyEventHandlers(eventHandlers);
			}

			notifyEventHandlers(genericEventHandlers);
		}

		private void notifyEventHandlers (List<EventHandler<D, E>> eventHandlers) {
			synchronized (eventHandlers) {
				for (EventHandler<D, E> eventHandler : eventHandlers) {
					if (error) {
						eventHandler.handleError(event);
					} else {
						eventHandler.handleEvent(event);
					}
				}
			}
		}
	}
}
