/*
 * Java
 *
 * Copyright 2016-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */

package ej.basedriver.util;

import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.Executor;

import ej.basedriver.Controller;
import ej.basedriver.DryContact;
import ej.basedriver.EventControllerListener;
import ej.basedriver.MultilevelSensor;
import ej.basedriver.Switch;
import ej.basedriver.SwitchWithReturnState;
import ej.basedriver.Thermostat;
import ej.basedriver.event.DeviceEventGenerator;
import ej.basedriver.event.Event;
import ej.basedriver.event.EventHandler;
import ej.basedriver.event.generator.impl.ControllerEventGenerator;
import ej.basedriver.event.generator.impl.DryContactEventGenerator;
import ej.basedriver.event.generator.impl.MultilevelSensorEventGenerator;
import ej.basedriver.event.generator.impl.SwitchEventGenerator;
import ej.basedriver.event.generator.impl.SwitchStateEventGenerator;
import ej.basedriver.event.generator.impl.ThermostatEventGenerator;
import ej.ecom.Device;
import ej.ecom.DeviceManager;
import ej.util.concurrent.SingleThreadExecutor;

public class EcomEventControllerListener implements EventControllerListener {
	private final Map<String, DeviceEventGenerator<?,?>> deviceEventGenerators;
	/**
	 *
	 */
	public EcomEventControllerListener() {
		Executor executor = new SingleThreadExecutor();
		deviceEventGenerators = new HashMap<>();
		deviceEventGenerators.put(Controller.class.getName(), new ControllerEventGenerator(executor));
		deviceEventGenerators.put(Switch.class.getName(), new SwitchEventGenerator(executor));
		deviceEventGenerators.put(SwitchWithReturnState.class.getName(), new SwitchStateEventGenerator(executor));
		deviceEventGenerators.put(DryContact.class.getName(), new DryContactEventGenerator(executor));
		deviceEventGenerators.put(Thermostat.class.getName(), new ThermostatEventGenerator(executor));
		deviceEventGenerators.put(MultilevelSensor.class.getName(), new MultilevelSensorEventGenerator(executor));
	}


	@Override
	public <D extends Device> void onError(Class<D> deviceType, D device) {
		DeviceEventGenerator<D,?> deviceEventGenerator = getDeviceEventGenerator(deviceType);
		if (deviceEventGenerator != null) {
			deviceEventGenerator.postError(device);
		}
	}

	@Override
	public <D extends Device> void onEvent(Class<D> deviceType, D device) {
		DeviceEventGenerator<D,?> deviceEventGenerator = getDeviceEventGenerator(deviceType);
		if (deviceEventGenerator != null) {
			deviceEventGenerator.postEvent(device);
		}
	}

	@Override
	public <D extends Device> void onRegister(Class<D> deviceType, final D device) {
		DeviceManager.register(deviceType, device);
	}

	@Override
	public <D extends Device> void onUnregister(D device) {
		DeviceManager.unregister(device);
	}

	@Override
	public <D extends Device, E extends Event<D>> void addEventHandler(
			String deviceType, EventHandler<D, E> eventHandler, D device) {
		DeviceEventGenerator<D, E> deviceEventGenerator = getDeviceEventGenerator(deviceType);
		if (deviceEventGenerator != null) {
			deviceEventGenerator.addEventHandler(eventHandler, device);
		}
	}

	@Override
	public <D extends Device, E extends Event<D>> void addEventHandler(
			String deviceType, EventHandler<D, E> eventHandler) {
		DeviceEventGenerator<D, E> deviceEventGenerator = getDeviceEventGenerator(deviceType);
		if (deviceEventGenerator != null) {
			deviceEventGenerator.addEventHandler(eventHandler);
		}
	}

	@Override
	public <D extends Device, E extends Event<D>> void removeEventHandler(EventHandler<D, E> eventHandler) {
		for (DeviceEventGenerator<?, ?> deviceEventGenerator : deviceEventGenerators.values()) {
			deviceEventGenerator.removeEventHandler(eventHandler);
		}
	}

	private <D extends Device, E extends Event<D>> DeviceEventGenerator<D, E> getDeviceEventGenerator (Class<D> deviceType) {
		return getDeviceEventGenerator(deviceType.getName());
	}

	@SuppressWarnings("unchecked")
	private <D extends Device, E extends Event<D>> DeviceEventGenerator<D, E> getDeviceEventGenerator (String deviceType) {
		return (DeviceEventGenerator<D, E>) deviceEventGenerators.get(deviceType);
	}
}

