/*
 * Java
 *
 * Copyright 2014-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.basedriver.zwave;

import java.io.IOException;

import ej.basedriver.DryContact;
import ej.basedriver.MultilevelSensor;
import ej.basedriver.SwitchWithReturnState;
import ej.basedriver.impl.AbstractNode;
import ej.basedriver.zwave.device.ZwaveDryContact;
import ej.basedriver.zwave.device.ZwaveMultilevelSensor;
import ej.basedriver.zwave.device.ZwaveSwitch;
import ej.basedriver.zwave.device.ZwaveThermostat;
import ej.basedriver.zwave.frame.SendDataListener;
import ej.ecom.Device;
import ej.ecom.HardwareDescriptor;

/**
 * A zwave node.
 */
public class ZwaveNode extends AbstractNode<ZwaveController> implements SendDataListener {
	/**
	 * Id of the node.
	 */
	protected byte nodeid;
	private boolean listening;
	private boolean awake;
	private HardwareDescriptor descriptor;

	/**
	 * Instantiate a zwave node for a specific controller.
	 *
	 * @param controller
	 *            the controller.
	 * @param nodeid
	 *            the id of the node.
	 */
	public ZwaveNode(ZwaveController controller, byte nodeid) {
		super(controller);
		this.nodeid = nodeid;
		controller.addNode(nodeid, this);
		awake = true;
	}


	@Override
	public String getName() {
		return Byte.toString(nodeid);
	}

	/**
	 * Gets the node unique id.
	 *
	 * @return the nodeid
	 */
	public byte getNodeUid() {
		return nodeid;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}

		if (!(obj instanceof ZwaveNode)) {
			return false;
		}

		ZwaveNode node = (ZwaveNode) obj;
		return nodeid == node.nodeid;
	}

	@Override
	public <D extends Device> HardwareDescriptor<D> getDescriptor() {
		if (descriptor == null) {
			descriptor = new HardwareDescriptor<D>() {
				@Override
				public String getName() {
					return ZwaveNode.this.getName();
				}

				@Override
				public String[] getPropertyNames() {
					return new String[0];
				}

				@Override
				public String[] getPropertyValues() {
					return new String[0];
				}

				@Override
				public String getProperty(String propertyName) {
					return null;
				}
			};
		}
		return descriptor;
	}

	/**
	 * Creates and add a profile.
	 *
	 *
	 * @param basic
	 *            The basic type of profile.
	 * @param generic
	 *            The generic type.
	 * @param specific
	 *            The specific type.
	 * @see ZwaveProtocol
	 */
	public void createProfiles(byte basic, byte generic, byte specific) {
		switch (generic) {
		case ZwaveProtocol.GENERIC_TYPE_SWITCH_BINARY :
			try {
				controller.getListener().onRegister(SwitchWithReturnState.class, new ZwaveSwitch(this));
			} catch (Exception e) {
			}
			break;
		case ZwaveProtocol.GENERIC_TYPE_SENSOR_BINARY :
			try {
				controller.getListener().onRegister(DryContact.class, new ZwaveDryContact(this));
			} catch (Exception e) {
			}
			break;
		case ZwaveProtocol.GENERIC_TYPE_THERMOSTAT:
			controller.getListener().onRegister(MultilevelSensor.class, new ZwaveMultilevelSensor(this));
			controller.getListener().onRegister(DryContact.class, new ZwaveDryContact(this));
			// Do not register the thermostat because it waits for its modes to be set.
			try {
				new ZwaveThermostat(this);
			} catch (IOException e) {
				e.printStackTrace();
			}
			break;
		default:
			break;
		}

	}

	/**
	 * Set the option.
	 *
	 * @param capabilities
	 * @param security
	 */
	public void options(byte capabilities, byte security) {
		listening = (capabilities & 0x80) != 0;
		//		awake = listening;
		boolean optional = (security & 0x80) != 0;
	}

	public boolean isAwake() {
		return awake;
	}

	/**
	 * @param awake
	 */
	public void setAwake(boolean awake) {
		this.awake = awake;
	}

	public boolean isListening() {
		return listening;
	}

	/**
	 * @param listening
	 */
	public void setListening(boolean listening) {
		this.listening = listening;
	}

	public boolean isReady() {
		return listening || awake;
	}

	@Override
	public void notifyComplete() {
	}

	@Override
	public void notifyError(RequestFrameError error) {
	}

	@Override
	public ZwaveNode getNode() {
		return this;
	}

	@Override
	public boolean exclude() {
		return false;
	}

	@Override
	public void callbackComplete(byte[] bytes, int commandClassIndex) {
	}
}
