/*
 * Java
 *
 * Copyright 2013-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.components.dependencyinjection;

import ej.basictool.map.PackedMap;
import ej.components.dependencyinjection.util.ServiceLoaderHelper;

/**
 * Abstract service loader that provides a default skeleton of service loader.
 * <p>
 * Subclasses need to provide the implementation class name matching the requested services. The implementation is then
 * instantiated using {@link Class#forName(String)} and {@link Class#newInstance()}. The singletons are stored in a map.
 */
public abstract class AbstractServiceLoader implements ServiceLoader {

	private final PackedMap<Class<?>, Object> services;

	/**
	 * Creates a new abstract service loader.
	 */
	public AbstractServiceLoader() {
		this.services = new PackedMap<>();
	}

	@Override
	public <T> T getService(Class<T> service) {
		return getService(service, null, true);
	}

	@Override
	public <T> T getService(Class<T> service, Class<? extends T> defaultImplementation) {
		return getService(service, defaultImplementation, true);
	}

	@Override
	public <T> T getService(Class<T> service, boolean singleton) {
		return getService(service, null, singleton);
	}

	@Override
	public <T> T getService(Class<T> service, Class<? extends T> defaultImplementation, boolean singleton) {
		PackedMap<Class<?>, Object> services = this.services;
		if (singleton) {
			@SuppressWarnings("unchecked")
			T implementation = (T) services.get(service); // cast ensured by construction (see put() call)
			if (implementation != null) {
				return implementation;
			}
		}

		String serviceName = service.getName();
		String implementationName = this.getImplementationName(serviceName);
		if (implementationName == null) {
			if (defaultImplementation == null) {
				return null;
			}
			implementationName = defaultImplementation.getName();
		}

		T serviceImplementation = ServiceLoaderHelper.createClassInstance(service, implementationName);
		if (singleton) {
			services.put(service, serviceImplementation);
		}
		return serviceImplementation;
	}

	/**
	 * Gets the implementation class name for the given key.
	 *
	 * @param key
	 *            the service name.
	 * @return the implementation class name, or {@code null} if none.
	 */
	protected abstract String getImplementationName(String key);

}