/*
 * Java
 *
 * Copyright 2015-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.components.dependencyinjection;

import ej.basictool.map.PackedMap;
import ej.components.dependencyinjection.util.ServiceLoaderHelper;

/**
 * The default service loader allows to get service instance based on default service implementation class.
 */
public class DefaultServiceLoader implements ServiceLoader {

	private final PackedMap<Class<?>, Object> services;

	/**
	 * Instantiates a new default service loader.
	 */
	public DefaultServiceLoader() {
		this.services = new PackedMap<>();
	}

	@Override
	public <T> T getService(Class<T> service) {
		return getService(service, null, true);
	}

	@Override
	public <T> T getService(Class<T> service, Class<? extends T> defaultImplementation) {
		return getService(service, defaultImplementation, true);
	}

	@Override
	public <T> T getService(Class<T> service, boolean singleton) throws ServiceLoadingException {
		return getService(service, null, singleton);
	}

	@Override
	public <T> T getService(Class<T> service, Class<? extends T> defaultImplementation, boolean singleton) {
		if (singleton) {
			T cachedService = getCachedService(service);
			if (cachedService != null) {
				return cachedService;
			}
		}
		if (defaultImplementation != null) {
			T instantiatedService = ServiceLoaderHelper.createClassInstance(service, defaultImplementation.getName());
			if (singleton) {
				this.services.put(service, instantiatedService);
			}
			return instantiatedService;
		}
		return null;
	}

	@SuppressWarnings("unchecked")
	private <T> T getCachedService(Class<T> service) {
		return (T) this.services.get(service); // cast ensured by cache policy
	}

}
