/*
 * Java
 *
 * Copyright 2013-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.components.dependencyinjection;

/**
 * A simple dependency injection facility.
 * <p>
 * Gets a service loader that is able to retrieve services implementation.
 * <p>
 * The service loader implementation is retrieved from system properties (see {@link AbstractServiceLoader}), if none is
 * found, a {@link AbstractServiceLoader} is used.
 * <p>
 * Usage:
 *
 * <pre>
 * MyInterface obj = ServiceLoaderFactory.getServiceLoader().getService(MyInterface.class);
 * </pre>
 */
public class ServiceLoaderFactory {

	private static final ServiceLoader SERVICE_LOADER = clinit();

	private ServiceLoaderFactory() {
	}

	private static ServiceLoader clinit() {
		ServiceLoader propertiesServiceLoader = new SystemPropertiesServiceLoader();
		ServiceLoader serviceLoaderCandidate = propertiesServiceLoader.getService(ServiceLoader.class);
		if (serviceLoaderCandidate != null) {
			return serviceLoaderCandidate;
		} else {
			CompositeServiceLoader compositeServiceLoader = new CompositeServiceLoader();
			compositeServiceLoader.addServiceLoader(propertiesServiceLoader);
			compositeServiceLoader.addServiceLoader(new MetaInfServiceLoader());
			compositeServiceLoader.addServiceLoader(new RegistryServiceLoader());
			return compositeServiceLoader;
		}
	}

	/**
	 * Gets a service loader unique instance.
	 *
	 * @return a service loader.
	 */
	public static ServiceLoader getServiceLoader() {
		return SERVICE_LOADER;
	}

}