/*
 * Java
 *
 * Copyright 2016-2017 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package ej.data;

import java.io.IOException;


/**
 * A {@link DataWriter} provides APIs for writing a structured data in streaming.
 * The structured data format is documented in {@link DataReader}.
 *
 * <p>
 * Here are some examples of usage. The data expected to be written is displayed in the JSON format.
 * <p>
 * <ul>
 * <li>
 * Writing a base type value of type boolean.
 * <p>
 * Data:
 * <pre>
 * 	true
 * </pre>
 * Code:
 * <pre>
 * 	dw.writeBooleanValue(true);
 * </pre>
 * </li>
 * <li>
 * Writing an array:
 * <p>
 * Data:
 * <pre>
 * 	[ true, "MicroEJ", 2016 ]
 * </pre>
 * Code:
 * <pre>
 * 	dw.writeArrayValueStart();
 * 	dw.writeBooleanValue(true);
 * 	dw.writeStringValue("MicroEJ");
 * 	dw.writeIntValue(2016);
 * 	dw.writeArrayValueEnd();
 * </pre>
 * </li>
 * <li>
 * Writing a map:
 * <p>
 * Data:
 * <pre>
 * 	{ "aKey1" : "MicroEJ", "aKey2" : 2016 }
 * </pre>
 * Code:
 * <pre>
 * 	dw.writeMapValueStart();
 * 	dw.writeKeyString("aKey1");
 * 	dw.writeStringValue("MicroEJ");
 * 	dw.writeKeyString("aKey2");
 * 	dw.writeIntValue(2016);
 * 	dw.writeMapValueEnd();
 * </pre>
 * </li>
 * </ul>
 *
 * @see DataReader
 */
public interface DataWriter {

	/**
	 * Writes the next boolean value.
	 * @param value the boolean value.
	 * @throws IOException on I/O error.
	 * @throws IllegalStateException if the end of the data is reached or if a map key is expected.
	 */
	void writeBooleanValue(boolean value) throws IOException;

	/**
	 * Writes the next integer value (signed 32bits).
	 * @param value the int value.
	 * @throws IOException on I/O error.
	 * @throws IllegalStateException if the end of the data is reached or if a map key is expected.
	 */
	void writeIntValue(int value) throws IOException;

	/**
	 * Writes the next expected long integer value (signed 64bits).
	 * @param value the long value.
	 * @throws IOException on I/O error.
	 * @throws IllegalStateException if the end of the data is reached or if a map key is expected.
	 */
	void writeLongValue(long value) throws IOException;

	/**
	 * Writes the next double precision floating point value (64bits).
	 * @param value the double value.
	 * @throws IOException on I/O error.
	 * @throws IllegalStateException if the end of the data is reached or if a map key is expected.
	 */
	void writeDoubleValue(double value) throws IOException;

	/**
	 * Writes the next {@link String} value.
	 * @param value the String value.
	 * @throws IOException on I/O error.
	 * @throws IllegalStateException if the end of the data is reached or if a map key is expected.
	 */
	void writeStringValue(String value) throws IOException;

	/**
	 * Writes the expected integer key (signed 32bits) within the current map.
	 * @param key the int key.
	 * @throws IOException on I/O error.
	 * @throws IllegalStateException if the current collection is not a map or if a map value is expected.
	 */
	void writeIntKey(int key) throws IOException;

	/**
	 * Writes the expected {@link String} key within the current map.
	 * @param key the String key.
	 * @throws IOException on I/O error.
	 * @throws IllegalStateException if the current collection is not a map or if a map value is expected.
	 */
	void writeStringKey(String key) throws IOException;

	/**
	 * Writes the beginning of an array value.
	 * @throws IOException on I/O error.
	 * @throws IllegalStateException if the end of the data is reached or if a map key is expected.
	 */
	void writeArrayValueStart() throws IOException;

	/**
	 * Writes the end of an array value.
	 * @throws IOException on I/O error.
	 * @throws IllegalStateException if the current collection is not an array.
	 */
	void writeArrayValueEnd() throws IOException;

	/**
	 * Writes the beginning of a map value.
	 * @throws IOException on I/O error.
	 * @throws IllegalStateException if the end of the data is reached or if a map key is expected.
	 */
	void writeMapValueStart() throws IOException;

	/**
	 * Writes the end of a map value.
	 * @throws IOException on I/O error.
	 * @throws IllegalStateException if the current collection is not a map of if a key has been previously written without a value.
	 */
	void writeMapValueEnd() throws IOException;
}
