/*
 * Java
 *
 * Copyright 2014-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.eventswatcher;

import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import ej.microui.event.EventGenerator;
import ej.microui.util.EventHandler;

/**
 * A events watcher forwards events generated by an event generator to a list of event handlers.
 */
public class EventsWatcher implements EventHandler {

	private final List<EventHandler> watchHandlers;
	private final List<EventHandler> captureHandlers;
	private final EventGenerator eventGenerator;
	private EventHandler initialListener;

	/**
	 * Creates an events watcher.
	 *
	 * @param eventGenerator
	 *            the eventGenerator to manage.
	 */
	public EventsWatcher(EventGenerator eventGenerator) {
		this.eventGenerator = eventGenerator;
		this.watchHandlers = new ArrayList<>();
		this.captureHandlers = new ArrayList<>();
	}

	/**
	 * Add an event handler.
	 *
	 * @param eventHandler
	 *            the event handler to add.
	 * @param watch
	 *            <code>true</code> if the given event handler watches the events, <code>false</code> if it captures the
	 *            events.
	 * @throws NullPointerException
	 *             if eventHandler is <code>null</code>.
	 * @throws SecurityException
	 *             if the {@link EventsWatcher} can not be set as the EventHandler of the event generator.
	 */
	public synchronized void add(EventHandler eventHandler, boolean watch)
			throws SecurityException, NullPointerException {
		if (eventHandler == null) {
			throw new NullPointerException();
		}
		if (isEmpty()) {
			EventHandler handler = this.eventGenerator.getEventHandler();
			this.eventGenerator.setEventHandler(this);
			this.initialListener = handler;
		}
		if (watch) {
			this.watchHandlers.add(eventHandler);
		} else {
			this.captureHandlers.add(eventHandler);
		}
	}

	/**
	 * Removes an event handler.
	 *
	 * @param eventHandler
	 *            the event handler to remove.
	 * @param watch
	 *            <code>true</code> if the given event handler watches the events, <code>false</code> if it captures the
	 *            events.
	 */
	public synchronized void remove(EventHandler eventHandler, boolean watch) {
		if (watch) {
			removeHandler(eventHandler, this.watchHandlers);
		} else {
			removeHandler(eventHandler, this.captureHandlers);
		}
		if (isEmpty() && this.initialListener != null) {
			this.eventGenerator.setEventHandler(this.initialListener);
			this.initialListener = null;
		}
	}

	private boolean isEmpty() {
		return this.watchHandlers.isEmpty() && this.captureHandlers.isEmpty();
	}

	@Override
	public boolean handleEvent(int event) {
		boolean handle = false;
		List<EventHandler> watchersList = this.watchHandlers;
		// browse list by the end
		for (int i = watchersList.size(); --i >= 0;) {
			EventHandler handler = watchersList.get(i);
			handleEvent(event, handler);
		}

		List<EventHandler> captureList = this.captureHandlers;
		// browse list by the end until a robber rob the event!
		for (int i = captureList.size(); --i >= 0;) {
			EventHandler handler = captureList.get(i);
			if (handleEvent(event, handler)) {
				handle = true;
				break;
			}
		}

		if (!handle) {
			EventHandler initialListener = this.initialListener;
			if (initialListener == null) {
				initialListener = this.eventGenerator.getEventHandler();
			}
			if (initialListener != null) {
				handle = initialListener.handleEvent(event);
			}
		}

		return handle;
	}

	private boolean handleEvent(int value, EventHandler eventHandler) {
		try {
			return eventHandler.handleEvent(value);
		} catch (Throwable t) {
			// protect against user errors
			uncaughtException(t);
		}
		return false;
	}

	private void removeHandler(EventHandler eventHandler, List<EventHandler> list) {
		int index = list.size() - 1;
		EventHandler currentHandler;

		while (index >= 0) {
			currentHandler = list.get(index);

			if (currentHandler.equals(eventHandler)) {
				list.remove(index);
				break;
			}

			index--;
		}
	}

	private void uncaughtException(Throwable t) {
		Logger.getAnonymousLogger().log(Level.WARNING, t.getMessage(), t);
	}
}