/*
 * Java
 *
 * Copyright 2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.fp.widget;

import java.util.concurrent.atomic.AtomicInteger;

import ej.fp.FrontPanel;
import ej.fp.Image;
import ej.fp.Widget;

/**
 * This widget holds all methods required by MicroUI implementation to be compatible with the MicroUI graphical engine.
 * It is also compatible with the MicroEJ platform scripts.
 * <p>
 * A subclass of this widget is required to run a MicroUI application which uses the class MicroUI Display. Otherwise
 * some errors will be thrown at runtime and the application will not able to run correctly.
 * <p>
 * The display size may be different than the widget size. This allows to show on computer a bigger or smaller display
 * in pixels than the embedded display. The display size is the size given to the MicroEJ application. It has to be used
 * to store the drawings (see {@link #getDrawingBuffer()}. If not set by the front panel parser using
 * {@link #setDisplayWidth(int)} and {@link #setDisplayHeight(int)}, the display size is equals to the widget size (see
 * {@link #finalizeConfiguration()}).
 */
@SuppressWarnings("nls")
public abstract class MicroUIDisplay extends Widget {

	/**
	 * The abstract base class for display extensions. A {@code DisplayExtension} specifies the configuration of the
	 * simulated display.
	 */
	public abstract static class AbstractDisplayExtension {

		/**
		 * Must have a no-arg constructor.
		 */
		public AbstractDisplayExtension() {
		}

		/**
		 * Gets the number of colors that can be represented on the device.
		 * <p>
		 * Note that the number of colors for a black and white display is 2.
		 * <p>
		 * A null or negative value allows the widget finds the number of colors according the available number of bits
		 * per pixels set during the MicroEJ platform build: <code>1 << BPP</code>. Override this function when the
		 * number of colors does not follow this rule.
		 * <p>
		 * The default implementation returns <code>0</code>.
		 *
		 * @return the number of colors of the display.
		 */
		public int numberOfColors() {
			return 0;
		}

		/**
		 * Returns whether the display provides color.
		 * <p>
		 * The default implementation returns <code>true</code>.
		 *
		 * @return true if display has color, false otherwise.
		 */
		public boolean isColor() {
			return true;
		}

		/**
		 * Returns whether the display uses an underlying double buffer (either hardware or software).
		 * <p>
		 * The default implementation returns <code>true</code>.
		 *
		 * @return true if the display is double buffered, false otherwise.
		 */
		public boolean isDoubleBuffered() {
			return true;
		}

		/**
		 * Converts a display color into an ARGB color format.
		 * <p>
		 * ARGB value is interpreted as a 32-bit ARGB color, where the eight less significant bits matches the blue
		 * component, the next eight bits matches the green component, the next eight bits matches the red component and
		 * the next eight bits matches the alpha level.
		 *
		 * @param displayColor
		 *            the display color to convert.
		 * @return the converted color.
		 */
		public abstract int convertDisplayColorToARGBColor(int displayColor);

		/**
		 * Converts a ARGB color into a display color format.
		 * <p>
		 * ARGB value is interpreted as a 32-bit ARGB color, where the eight less significant bits matches the blue
		 * component, the next eight bits matches the green component, the next eight bits matches the red component and
		 * the next eight bits matches the alpha level.
		 *
		 * @param argbColor
		 *            the ARGB color to convert.
		 * @return the converted color.
		 */
		public abstract int convertARGBColorToDisplayColor(int argbColor);

		/**
		 * Prepare the blending of two ARGB colors. This is useful only when the LCD is a palletized LCD. This function
		 * is called by framework when the MicroEJ application draws something which requires a blending between the
		 * current foreground color and a specific background color (draw a string, draw an anti-aliased line etc.). The
		 * implementation has to replace the ARGB colors by the indexes of these colors in the LCD CLUT. The framework
		 * will use the intermediate values between these two indexes instead of blending in software the ARGB colors.
		 * No post conversion will be performed later.
		 * <p>
		 * When the ARGB colors are not available in the CLUT or when the range specified by the two ARGB colors is not
		 * expected by the CLUT, the implementation should return <code>false</code>. In this way the blending will be
		 * done in software and the result color will be converted later thanks a call to
		 * {@link #convertARGBColorToDisplayColor(int)}.
		 * <p>
		 * The default implementation returns <code>false</code>.
		 *
		 * @param foreground
		 *            the foreground ARGB color to convert.
		 * @param background
		 *            the background ARGB color to convert.
		 * @return <code>true</code> when the indexes have been found.
		 */
		public boolean prepareBlendingOfIndexedColors(AtomicInteger foreground, AtomicInteger background) {
			return false;
		}
	}

	/**
	 * Minimal value of backlight according MicroUI specification.
	 */
	public static final int BACKLIGHT_MIN = 0;

	/**
	 * Maximal value of backlight according MicroUI specification.
	 */
	public static final int BACKLIGHT_MAX = 100;

	/**
	 * Minimal value of contrast according MicroUI specification.
	 */
	public static final int CONTRAST_MIN = 0;

	/**
	 * Maximal value of contrast according MicroUI specification.
	 */
	public static final int CONTRAST_MAX = 100;

	/**
	 * Property set during the platform build which defines the bits organization in pixels. The expected value is among
	 * this list: {@link #GRAPHICAL_ENGINE_STUB_ID}, {@link #GRAPHICAL_ENGINE_CUSTOM_ID},
	 * {@link #GRAPHICAL_ENGINE_ARGB8888_ID}, {@link #GRAPHICAL_ENGINE_RGB888_ID}, {@link #GRAPHICAL_ENGINE_RGB565_ID},
	 * {@link #GRAPHICAL_ENGINE_ARGB1555_ID}, {@link #GRAPHICAL_ENGINE_ARGB4444_ID}, {@link #GRAPHICAL_ENGINE_C4_ID},
	 * {@link #GRAPHICAL_ENGINE_C2_ID} and {@link #GRAPHICAL_ENGINE_C1_ID}.
	 * <p>
	 * A missing value and all others values will throw an {@link IllegalArgumentException} at runtime.
	 */
	private static final String GRAPHICAL_ENGINE_PIXEL_ORGANIZATION_PROPERTY = "com.is2t.microbsp.microui.io.display.pixel";

	/**
	 * Property set during the platform build which defines the number of bits per pixel.
	 */
	private static final String GRAPHICAL_ENGINE_BPP_PROPERTY = "com.is2t.microbsp.microui.io.display.bpp";

	/**
	 * Represents the organization of bits in a pixel according the BPP (bits per pixels). The value is known by MicroUI
	 * graphical engine and must not been changed.
	 * <p>
	 * This constant is used when the graphical engine is not defined by the platform.
	 */
	private static final int GRAPHICAL_ENGINE_NOTDEFINED_ID = -2;

	/**
	 * Represents the organization of bits in a pixel according the BPP (bits per pixels). The value is known by MicroUI
	 * graphical engine and must not been changed.
	 * <p>
	 * This constant is used when there is no display on the platform.
	 */
	private static final int GRAPHICAL_ENGINE_STUB_ID = -1;

	/**
	 * Represents the organization of bits in a pixel according the BPP (bits per pixels). The value is known by MicroUI
	 * graphical engine and must not been changed.
	 * <p>
	 * This constant is used when the graphical engine is not generic: the display widget requires a user extension
	 * class to describe this organization.
	 */
	private static final int GRAPHICAL_ENGINE_CUSTOM_ID = 0;

	/**
	 * Represents the organization of bits in a pixel according the BPP (bits per pixels). The value is known by MicroUI
	 * graphical engine and must not been changed.
	 * <p>
	 * The bits organization is 8 bits for alpha, 8 bits for red, 8 bits for green and 8 bits for blue.
	 */
	private static final int GRAPHICAL_ENGINE_ARGB8888_ID = 2;

	/**
	 * Represents the organization of bits in a pixel according the BPP (bits per pixels). The value is known by MicroUI
	 * graphical engine and must not been changed.
	 * <p>
	 * The bits organization is 0 bit for alpha, 8 bits for red, 8 bits for green and 8 bits for blue.
	 */
	private static final int GRAPHICAL_ENGINE_RGB888_ID = 3;

	/**
	 * Represents the organization of bits in a pixel according the BPP (bits per pixels). The value is known by MicroUI
	 * graphical engine and must not been changed.
	 * <p>
	 * The bits organization is 0 bit for alpha, 5 bits for red, 6 bits for green and 5 bits for blue.
	 */
	private static final int GRAPHICAL_ENGINE_RGB565_ID = 4;

	/**
	 * Represents the organization of bits in a pixel according the BPP (bits per pixels). The value is known by MicroUI
	 * graphical engine and must not been changed.
	 * <p>
	 * The bits organization is 1 bit for alpha, 5 bits for red, 5 bits for green and 5 bits for blue.
	 */
	private static final int GRAPHICAL_ENGINE_ARGB1555_ID = 5;

	/**
	 * Represents the organization of bits in a pixel according the BPP (bits per pixels). The value is known by MicroUI
	 * graphical engine and must not been changed.
	 * <p>
	 * The bits organization is 4 bits for alpha, 4 bits for red, 4 bits for green and 4 bits for blue.
	 */
	private static final int GRAPHICAL_ENGINE_ARGB4444_ID = 6;

	/**
	 * Represents the organization of bits in a pixel according the BPP (bits per pixels). The value is known by MicroUI
	 * graphical engine and must not been changed.
	 * <p>
	 * The bits organization is 0 bit for alpha and 4 bits for gray.
	 */
	private static final int GRAPHICAL_ENGINE_C4_ID = 0xd;

	/**
	 * Represents the organization of bits in a pixel according the BPP (bits per pixels). The value is known by MicroUI
	 * graphical engine and must not been changed.
	 * <p>
	 * The bits organization is 0 bit for alpha and 2 bits for gray.
	 */
	private static final int GRAPHICAL_ENGINE_C2_ID = 0xe;

	/**
	 * Represents the organization of bits in a pixel according the BPP (bits per pixels). The value is known by MicroUI
	 * graphical engine and must not been changed.
	 * <p>
	 * The bits organization is 0 bit for alpha and 1 bit for gray.
	 */
	private static final int GRAPHICAL_ENGINE_C1_ID = 0xf;

	// widget extension which describes the display.
	private AbstractDisplayExtension extension;

	// widget user extension class name. Set in fp file.
	private String extensionClassName;

	// current backlight
	private int backlight;

	// display size in pixels, may be different than widget size (see
	// setDisplaylWidth())
	private int displayWidth; // 0 init
	private int displayHeight; // 0 init

	/**
	 * Creates a widget display compatible with MicroUI implementation.
	 */
	public MicroUIDisplay() {
	}

	/**
	 * Sets the width of the display in pixels (see class comment).
	 * <p>
	 * This method should only be called by front panel parser.
	 *
	 * @param width
	 *            the width to set.
	 */
	public void setDisplayWidth(int width) {
		if (width < 0) {
			throw new IllegalArgumentException("Display width cannot be negative.");
		}
		this.displayWidth = width;
	}

	/**
	 * Sets the height of the display in pixels (see class comment).
	 * <p>
	 * This method should only be called by front panel parser.
	 *
	 * @param height
	 *            the height to set.
	 */
	public void setDisplayHeight(int height) {
		if (height < 0) {
			throw new IllegalArgumentException("Display height cannot be negative.");
		}
		this.displayHeight = height;
	}

	/**
	 * Defines the class name of the user display extension when the graphical engine set in platform is a CUSTOM.
	 * <p>
	 * For all others graphical engine, this extension class is useless. However, when set, a check is performed in
	 * {@link #finalizeConfiguration()} to verify whether this class is valid, even if it is not used.
	 * <p>
	 * This method should only be called by front panel parser.
	 *
	 * @param extensionClassName
	 *            the user display extension class name.
	 */
	public void setExtensionClass(String extensionClassName) {
		this.extensionClassName = extensionClassName;
	}

	@Override
	public synchronized void finalizeConfiguration() {
		super.finalizeConfiguration();

		// display size may be different than widget size
		if (this.displayWidth == 0) {
			this.displayWidth = getWidth();
		}
		if (this.displayHeight == 0) {
			this.displayHeight = getHeight();
		}

		// no need to repaint behind display
		setOverlay(true);

		if (this.extensionClassName != null) {
			// check user class
			FrontPanel.getFrontPanel().verifyUserClass(getClass(), AbstractDisplayExtension.class,
					this.extensionClassName);
		}
	}

	@Override
	public void start() {

		super.start();

		// retrieve display pixel rendering: may load an user extension class which can
		// call Device.getDevice(); so this extension cannot be created in
		// finalizeConfiguration() method
		this.extension = newDisplayExtension();
	}

	/**
	 * Returns true if the display holds a backlight. Sub-classes should overwrite the default behavior which returns
	 * false.
	 *
	 * @return true when widget manages display backlight.
	 */
	public boolean hasBacklight() {
		// default implementation has no backlight
		return false;
	}

	/**
	 * Sets the backlight of the screen.
	 *
	 * @param on
	 *            the new state of the backlight.
	 */
	public void switchBacklight(boolean on) {
		setBacklight(on ? BACKLIGHT_MAX : BACKLIGHT_MIN);
	}

	/**
	 * Returns the backlight of the display. <code>backlight</code> value range is
	 * {@link #BACKLIGHT_MIN}-{@link #BACKLIGHT_MAX}. If the display do not manage backlight (
	 * <code>hasBacklight() == false</code>), returns {@link #BACKLIGHT_MIN}.
	 *
	 * @return the backlight of the display.
	 */
	public int getBacklight() {
		return hasBacklight() ? this.backlight : BACKLIGHT_MIN;
	}

	/**
	 * Sets the backlight of the display.
	 *
	 * @param backlight
	 *            the new value of the backlight.
	 */
	public void setBacklight(int backlight) {
		if (hasBacklight()) {
			if (this.backlight < BACKLIGHT_MIN) {
				this.backlight = BACKLIGHT_MIN;
			} else if (this.backlight > BACKLIGHT_MAX) {
				this.backlight = BACKLIGHT_MAX;
			} else {
				this.backlight = backlight;
			}
			repaint();
		}
	}

	/**
	 * Returns the contrast of the display. Sub-classes should overwrite the default behavior which return 0.
	 *
	 * @return the current contrast of the display (range {@link #CONTRAST_MIN}-{@link #CONTRAST_MAX}).
	 */
	public int getContrast() {
		// default implementation does not manage contrast: nothing to do
		return CONTRAST_MIN;
	}

	/**
	 * Sets the contrast of the display. Sub-classes should overwrite the default behavior which does nothing.
	 * <code>contrast</code> value range is {@link #CONTRAST_MIN}-{@link #CONTRAST_MAX}.
	 *
	 * @param contrast
	 *            the new value of the contrast.
	 */
	public void setContrast(int contrast) {
		// default implementation does not manage contrast: nothing to do
	}

	/**
	 * Gets the number of bits per pixel the display use. The number of bits per pixel depends on the graphical engine
	 * bits organization. This value is set in a system property by the platform.
	 *
	 * @return the number of bits par pixel of the display.
	 */
	public int getNbBitsPerPixel() {
		// in application mode the property exists; in designer mode, returns max BPP
		// value
		return getInteger(GRAPHICAL_ENGINE_BPP_PROPERTY, 32);
	}

	/**
	 * Gets the number of colors that can be represented on the device.
	 * <p>
	 * Note that the number of colors for a black and white display is 2. The extension has to return a valid number of
	 * colors or a null or negative value. In second case, the returned number of colors is calculated with the
	 * available number of bits per pixels set during the MicroEJ platform build: <code>1 << BPP</code>.
	 *
	 * @return the number of colors of the display.
	 */
	public int getNumberOfColors() {
		int n = this.extension.numberOfColors();
		return n <= 0 ? 1 << getNbBitsPerPixel() : n;
	}

	/**
	 * Tells whether the display offers color.
	 *
	 * @return true if display has color, else otherwise.
	 */
	public boolean isColor() {
		// Let the extension returns whether the display provides color.
		return this.extension.isColor();
	}

	/**
	 * Returns if the display uses an underlying double buffer (either hardware or software).
	 *
	 * @return true if the display is double buffered, else otherwise.
	 */
	public boolean isDoubleBuffered() {
		return this.extension.isDoubleBuffered();
	}

	/**
	 * Gets the display width.
	 *
	 * @return the display width.
	 */
	public int getDisplayWidth() {
		return this.displayWidth;
	}

	/**
	 * Gets the display height.
	 *
	 * @return the display height.
	 */
	public int getDisplayHeight() {
		return this.displayHeight;
	}

	/**
	 * Converts an ARGB color (0xAARRGGBB) to the equivalent LCD color.
	 *
	 * @param argbColor
	 *            the color to convert.
	 * @return the display color equivalent to <code>microUIColor</code>.
	 */
	public int convertARGBColorToDisplayColor(int argbColor) {
		return this.extension.convertARGBColorToDisplayColor(argbColor);
	}

	/**
	 * Converts a LCD color to the equivalent color in ARGB format (0xAARRGGBB).
	 *
	 * @param displayColor
	 *            the color to convert.
	 * @return the MicroUI color equivalent to <code>displayColor</code>.
	 */
	public int convertDisplayColorToARGBColor(int displayColor) {
		return this.extension.convertDisplayColorToARGBColor(displayColor);
	}

	/**
	 * Prepare the blending of two ARGB colors. This is useful only when the LCD is a palletized LCD. Default
	 * implementation returns false.
	 *
	 * @param foreground
	 *            the foreground ARGB color to convert.
	 * @param background
	 *            the background ARGB color to convert.
	 * @return <code>true</code> when the indexes have been found.
	 */
	public boolean prepareBlendingOfIndexedColors(AtomicInteger foreground, AtomicInteger background) {
		return this.extension.prepareBlendingOfIndexedColors(foreground, background);
	}

	/**
	 * Gets the buffer where performing the drawings.
	 *
	 * @return the drawing buffer (often called back buffer).
	 */
	public abstract Image getDrawingBuffer();

	/**
	 * Flushes the entire display. If the display do not have a <code>backBuffer</code> (not double buffered), nothing
	 * is done.
	 */
	public void flush() {
		flush(0, 0, getDisplayWidth(), getDisplayHeight());
	}

	/**
	 * Flushes a part of the display limited by specified bounds. If the display do not have a <code>backBuffer</code>
	 * (not double buffered), nothing is done.
	 *
	 * @param x
	 *            the x coordinate of the zone to flush.
	 * @param y
	 *            the x coordinate of the zone to flush.
	 * @param width
	 *            the width of the zone to flush.
	 * @param height
	 *            the height of the zone to flush.
	 */
	public abstract void flush(int x, int y, int width, int height);

	/**
	 * Creates the widget display extension according the platform graphical engine pixel organization fixed during
	 * platform build (this organization should be stored in a system property).
	 */
	private AbstractDisplayExtension newDisplayExtension() {

		AbstractDisplayExtension extension = null;

		// load property value from platform
		int type = getInteger(GRAPHICAL_ENGINE_PIXEL_ORGANIZATION_PROPERTY, GRAPHICAL_ENGINE_NOTDEFINED_ID);

		switch (type) {

		default:
		case GRAPHICAL_ENGINE_NOTDEFINED_ID:
			// the property has not been set: we are in designer mode
			if (this.extensionClassName != null) {
				// try to create the user extension
				extension = newUserDisplayExtension();
			} else {
				// no specific user extension: create a generic extension (be coherent with
				// #getNbBitsPerPixel())
				extension = new DisplayExtensionARGB8888();
			}

			break;

		case GRAPHICAL_ENGINE_STUB_ID:
			// platform mode (not viewer): there is no display on platform: fatal error
			throw new IllegalArgumentException(
					"Front panel is not coherent with the chosen graphical engine in the platform: there is no display in the platform whereas there is a display on the front panel\\! Please remove the front panel display or choose another graphical engine.");

		case GRAPHICAL_ENGINE_CUSTOM_ID:
			// platform mode (not viewer): an user extension is required to describe custom
			// bits organization in pixels
			if (this.extensionClassName == null) {
				// no defined extension -> fatal error
				throw new IllegalArgumentException(
						"An user extension is required to describe the custom graphical engine.");
			}
			extension = newUserDisplayExtension();
			break;

		case GRAPHICAL_ENGINE_ARGB8888_ID:
			extension = new DisplayExtensionARGB8888();
			break;

		case GRAPHICAL_ENGINE_RGB888_ID:
			extension = new DisplayExtensionRGB888();
			break;

		case GRAPHICAL_ENGINE_RGB565_ID:
			extension = new DisplayExtensionRGB565();
			break;

		case GRAPHICAL_ENGINE_ARGB1555_ID:
			extension = new DisplayExtensionARGB1555();
			break;

		case GRAPHICAL_ENGINE_ARGB4444_ID:
			extension = new DisplayExtensionARGB4444();
			break;

		case GRAPHICAL_ENGINE_C4_ID:
			extension = new DisplayExtensionC4();
			break;

		case GRAPHICAL_ENGINE_C2_ID:
			extension = new DisplayExtensionC2();
			break;

		case GRAPHICAL_ENGINE_C1_ID:
			extension = new DisplayExtensionC1();
			break;

		}

		return extension;
	}

	/**
	 * Create display user extension using the extension class defined in fp file.
	 *
	 * @return the widget display extension.
	 */
	private AbstractDisplayExtension newUserDisplayExtension() {
		return FrontPanel.getFrontPanel().newUserInstance(getClass(), AbstractDisplayExtension.class,
				this.extensionClassName);
	}

	private int getInteger(String prop, int defaultValue) {
		try {
			return Integer.parseInt(System.getProperty(prop).trim());
		} catch (Exception e) {
			return defaultValue;
		}
	}

	private abstract static class DisplayExtensionGeneric extends AbstractDisplayExtension {

		/**
		 * Transparency position on ARGB8888 format.
		 */
		protected static final int ALPHA_SHIFT = 24;

		/**
		 * Transparency mask on ARGB8888 format.
		 */
		protected static final int ALPHA_MASK = 0xff << ALPHA_SHIFT;

		/**
		 * Maximum value for transparency on ARGB8888 format.
		 */
		protected static final int ALPHA_FULLY_OPAQUE = 0xff << ALPHA_SHIFT;

		/**
		 * Minimum value for transparency on ARGB8888 format.
		 */
		protected static final int ALPHA_FULLY_TRANSPARENT = 0x00 << ALPHA_SHIFT;

	}

	private static class DisplayExtensionARGB8888 extends DisplayExtensionGeneric {

		@Override
		public int numberOfColors() {
			return 1 << 24;
		}

		@Override
		public int convertDisplayColorToARGBColor(int displayColor) {
			return displayColor;
		}

		@Override
		public int convertARGBColorToDisplayColor(int rgbColor) {
			return rgbColor;
		}
	}

	private static class DisplayExtensionRGB888 extends DisplayExtensionGeneric {

		@Override
		public int convertDisplayColorToARGBColor(int displayColor) {
			return ALPHA_FULLY_OPAQUE | displayColor;
		}

		@Override
		public int convertARGBColorToDisplayColor(int rgbColor) {
			return rgbColor & ~ALPHA_MASK;
		}
	}

	private static class DisplayExtensionRGB565 extends DisplayExtensionGeneric {

		@Override
		public int convertDisplayColorToARGBColor(int displayColor) {
			return ALPHA_FULLY_OPAQUE | ((displayColor & 0xf800) << 8) | ((displayColor & 0x07e0) << 5)
					| ((displayColor & 0x001f) << 3);
		}

		@Override
		public int convertARGBColorToDisplayColor(int rgbColor) {
			return ((rgbColor & 0xf80000) >> 8) | ((rgbColor & 0x00fc00) >> 5) | ((rgbColor & 0x0000f8) >> 3);
		}
	}

	private static class DisplayExtensionARGB1555 extends DisplayExtensionGeneric {

		@Override
		public int numberOfColors() {
			return 1 << 15;
		}

		@Override
		public int convertDisplayColorToARGBColor(int displayColor) {
			return ((displayColor & 0x8000) == 0x8000 ? ALPHA_FULLY_OPAQUE : ALPHA_FULLY_TRANSPARENT)
					| ((displayColor & 0x7c00) << 9) | ((displayColor & 0x03e0) << 6) | ((displayColor & 0x001f) << 3);
		}

		@Override
		public int convertARGBColorToDisplayColor(int rgbColor) {
			return (((rgbColor & ALPHA_FULLY_OPAQUE) == ALPHA_FULLY_OPAQUE) ? 0x8000 : 0) | ((rgbColor & 0xf80000) >> 9)
					| ((rgbColor & 0x00f800) >> 6) | ((rgbColor & 0x0000f8) >> 3);
		}
	}

	private static class DisplayExtensionARGB4444 extends DisplayExtensionGeneric {

		@Override
		public int numberOfColors() {
			return 1 << 12;
		}

		@Override
		public int convertDisplayColorToARGBColor(int displayColor) {
			return ((displayColor & 0xf000) << 16) | ((displayColor & 0xf000) << 12) | ((displayColor & 0x0f00) << 12)
					| ((displayColor & 0x0f00) << 8) | ((displayColor & 0x00f0) << 8) | ((displayColor & 0x00f0) << 4)
					| ((displayColor & 0x000f) << 4) | (displayColor & 0x000f);
		}

		@Override
		public int convertARGBColorToDisplayColor(int rgbColor) {
			return ((rgbColor & 0xf0000000) >> 16) | ((rgbColor & 0x00f00000) >> 12) | ((rgbColor & 0x0000f000) >> 8)
					| ((rgbColor & 0x000000f0) >> 4);
		}
	}

	private abstract static class DisplayExtensionCx extends DisplayExtensionGeneric {

		@Override
		public boolean isColor() {
			return false;
		}

		protected int getGray(int argb) {
			return getGray((argb >> 16) & 0xff, (argb >> 8) & 0xff, argb & 0xff);
		}

		protected int getGray(int red, int green, int blue) {
			return (red * 3483 /* 0.2126 * 16384 */
					+ green * 11718 /* 0.7152 * 16384 */
					+ blue * 1183 /* 0.0722 * 16384 */
			) >> 14; /* 16384 */
		}
	}

	private static class DisplayExtensionC4 extends DisplayExtensionCx {

		@Override
		public int convertDisplayColorToARGBColor(int displayColor) {
			return ALPHA_FULLY_OPAQUE | (displayColor * 0x111111);
		}

		@Override
		public int convertARGBColorToDisplayColor(int rgbColor) {
			return getGray(rgbColor) / 0x11;
		}
	}

	private static class DisplayExtensionC2 extends DisplayExtensionCx {

		@Override
		public int convertDisplayColorToARGBColor(int displayColor) {
			return ALPHA_FULLY_OPAQUE | (displayColor * 0x555555);
		}

		@Override
		public int convertARGBColorToDisplayColor(int rgbColor) {
			return getGray(rgbColor) / 0x55;
		}
	}

	private static class DisplayExtensionC1 extends DisplayExtensionCx {

		@Override
		public int convertDisplayColorToARGBColor(int displayColor) {
			return ALPHA_FULLY_OPAQUE | (displayColor * 0xffffff);
		}

		@Override
		public int convertARGBColorToDisplayColor(int rgbColor) {
			return getGray(rgbColor) == 0xff ? 1 : 0;
		}
	}
}
