/*
 * Copyright 2015-2020 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.mwt.style.outline.border;

import ej.annotation.Nullable;
import ej.bon.XMath;
import ej.drawing.ShapePainter;
import ej.drawing.ShapePainter.Cap;
import ej.microui.display.GraphicsContext;
import ej.mwt.style.outline.Outline;
import ej.mwt.util.Outlineable;
import ej.mwt.util.Size;

/**
 * Draws a uniform border with round corners.
 * <p>
 * May be associated with {@link ej.mwt.style.background.RoundedBackground} or used on its own.
 * <p>
 * The thickness and corner radius are stored as a <code>char</code> for heap optimization and therefore cannot exceed
 * <code>65535</code>.
 */
public class RoundedBorder implements Outline {

	private static final int EMPTY_CIRCLE = 0;
	private static final int QUARTER_CIRCLE = 90;
	private static final int HALF_CIRCLE = 180;
	private static final int THREE_QUARTER_CIRCLE = 270;

	private static final int FADE = 1;
	private static final Cap CAP = Cap.ROUNDED;

	private final int color;
	private final char thickness;
	private final char cornerRadius;

	/**
	 * Creates a rounded border specifying its attributes.
	 * <p>
	 * The given corner radius and thickness are clamped between <code>0</code> and <code>Character.MAX_VALUE</code>.
	 *
	 * @param color
	 *            the color.
	 * @param cornerRadius
	 *            the corner radius.
	 * @param thickness
	 *            the border thickness.
	 */
	public RoundedBorder(int color, int cornerRadius, int thickness) {
		this.color = color;
		this.thickness = (char) XMath.limit(thickness, 0, Character.MAX_VALUE);
		this.cornerRadius = (char) XMath.limit(cornerRadius, 0, Character.MAX_VALUE);
	}

	/**
	 * Gets the color.
	 *
	 * @return the color.
	 */
	public int getColor() {
		return this.color;
	}

	/**
	 * Gets the thickness.
	 *
	 * @return the thickness.
	 */
	public int getThickness() {
		return this.thickness;
	}

	/**
	 * Gets the corner radius.
	 *
	 * @return the corner radius.
	 */
	public int getCornerRadius() {
		return this.cornerRadius;
	}

	@Override
	public void apply(Outlineable outlineable) {
		int thickness = this.thickness;
		outlineable.removeOutline(thickness, thickness, thickness, thickness);
	}

	@Override
	public void apply(GraphicsContext g, Size size) {
		boolean hasBackgroundColor = g.hasBackgroundColor();
		int backgroundColor = g.getBackgroundColor();
		g.removeBackgroundColor();

		int width = size.getWidth();
		int height = size.getHeight();
		int thickness = this.thickness;

		boolean emptyBorder = thickness == 0;

		// Since draw ellipse arc is not available, both radius have the same size.
		int cornerRadius = Math.min(this.cornerRadius, (Math.min(width, height) >> 1) - thickness);

		int borderSize = ((cornerRadius + thickness) << 1) + 1; // +1 for previous /2 precision
		boolean halfCircleHorizontal = borderSize >= width;
		boolean halfCircleVertical = borderSize >= height;
		boolean fullCircle = halfCircleHorizontal && halfCircleVertical;

		if (!emptyBorder) {
			g.setColor(this.color);
		}

		int borderDiameter = (cornerRadius << 1);
		int halfRectangle = thickness >> 1;
		int leftX = halfRectangle + 1;
		int topY = halfRectangle + 1;
		if (fullCircle) {
			ShapePainter.drawThickFadedCircle(g, leftX, topY, borderDiameter, thickness, FADE);
		} else {
			int rightX = leftX + width - thickness - 2 - 1;
			int bottomY = topY + height - thickness - 2 - 1;

			if (halfCircleVertical) {
				ShapePainter.drawThickFadedCircleArc(g, leftX, topY, borderDiameter, QUARTER_CIRCLE, HALF_CIRCLE,
						thickness, FADE, CAP, CAP);
				ShapePainter.drawThickFadedCircleArc(g, rightX - borderDiameter, topY, borderDiameter, -QUARTER_CIRCLE,
						HALF_CIRCLE, thickness, FADE, CAP, CAP);
			} else if (halfCircleHorizontal) {
				ShapePainter.drawThickFadedCircleArc(g, leftX, topY, borderDiameter, EMPTY_CIRCLE, HALF_CIRCLE,
						thickness, FADE, CAP, CAP);
				ShapePainter.drawThickFadedCircleArc(g, leftX, bottomY - borderDiameter, borderDiameter, HALF_CIRCLE,
						HALF_CIRCLE, thickness, FADE, CAP, CAP);
			} else {
				ShapePainter.drawThickFadedCircleArc(g, leftX, topY, borderDiameter, QUARTER_CIRCLE, QUARTER_CIRCLE,
						thickness, FADE, CAP, CAP);
				ShapePainter.drawThickFadedCircleArc(g, leftX, bottomY - borderDiameter, borderDiameter, HALF_CIRCLE,
						QUARTER_CIRCLE, thickness, FADE, CAP, CAP);
				ShapePainter.drawThickFadedCircleArc(g, rightX - borderDiameter, bottomY - borderDiameter,
						borderDiameter, THREE_QUARTER_CIRCLE, QUARTER_CIRCLE, thickness, FADE, CAP, CAP);
				ShapePainter.drawThickFadedCircleArc(g, rightX - borderDiameter, topY, borderDiameter, EMPTY_CIRCLE,
						QUARTER_CIRCLE, thickness, FADE, CAP, CAP);
			}

			// TODO: Avoid overlap of arc and line.
			if (!halfCircleHorizontal) {
				int horizontalLeftX = leftX + cornerRadius + 1;
				int horizontalRightX = rightX - cornerRadius;
				int bottomLineY = bottomY + 1;
				ShapePainter.drawThickFadedLine(g, horizontalLeftX, topY, horizontalRightX, topY, thickness, FADE, CAP,
						CAP);
				ShapePainter.drawThickFadedLine(g, horizontalLeftX, bottomLineY, horizontalRightX, bottomLineY,
						thickness, FADE, CAP, CAP);
			}
			if (!halfCircleVertical) {
				int verticalTopY = topY + cornerRadius + 1;
				int verticalBottomY = bottomY - cornerRadius;
				int rightLineX = rightX + 1;
				ShapePainter.drawThickFadedLine(g, leftX, verticalTopY, leftX, verticalBottomY, thickness, FADE, CAP,
						CAP);
				ShapePainter.drawThickFadedLine(g, rightLineX, verticalTopY, rightLineX, verticalBottomY, thickness,
						FADE, CAP, CAP);
			}
		}

		size.removeOutline(thickness, thickness, thickness, thickness);

		g.translate(thickness, thickness);
		g.intersectClip(0, 0, size.getWidth(), size.getHeight());

		if (hasBackgroundColor) {
			g.setBackgroundColor(backgroundColor);
		}
	}

	@Override
	public boolean equals(@Nullable Object obj) {
		if (obj instanceof RoundedBorder) {
			RoundedBorder border = (RoundedBorder) obj;
			return this.color == border.color && this.thickness == border.thickness
					&& this.cornerRadius == border.cornerRadius;
		}
		return false;
	}

	@Override
	public int hashCode() {
		return 17 * this.color + this.thickness * this.cornerRadius;
	}

}
