/*
 * Java
 *
 * Copyright 2015-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.style.background;

import ej.annotation.Nullable;
import ej.microui.display.GraphicsContext;
import ej.microui.display.Image;
import ej.style.container.AlignmentHelper;
import ej.style.container.Rectangle;

/**
 * Image-based background. If the image is not large enough, a plain color fills the remaining area.
 */
public abstract class AbstractImageBackground implements Background {

	@Nullable
	private Image image;
	private int alignment;
	private boolean fillColor;

	/**
	 * Creates an abstract image background.
	 */
	public AbstractImageBackground() {
	}

	/**
	 * Creates an abstract image background specifying its image and alignment.
	 *
	 * @param image
	 *            the background image to set.
	 * @param alignment
	 *            the alignment to set.
	 * @param fillColor
	 *            <code>true</code> to fill remaining area with background color, <code>false</code> otherwise.
	 */
	public AbstractImageBackground(Image image, int alignment, boolean fillColor) {
		this.image = image;
		this.alignment = alignment;
		this.fillColor = fillColor;
	}

	/**
	 * Sets the image.
	 *
	 * @param image
	 *            the image to set.
	 */
	public void setImage(@Nullable Image image) {
		this.image = image;
	}

	/**
	 * Gets the image.
	 *
	 * @return the image.
	 */
	@Nullable
	public Image getImage() {
		return this.image;
	}

	/**
	 * Sets the alignment.
	 *
	 * @param alignment
	 *            the alignment to set.
	 * @throws IllegalArgumentException
	 *             if the alignment is not valid.
	 * @see AlignmentHelper#validateAlignment(int)
	 */
	public void setAlignment(int alignment) {
		AlignmentHelper.validateAlignment(alignment);
		this.alignment = alignment;
	}

	/**
	 * Gets the alignment.
	 *
	 * @return the alignment.
	 */
	public int getAlignment() {
		return this.alignment;
	}

	/**
	 * Sets whether or not the remaining area is filled with the background color.
	 *
	 * @param fillColor
	 *            <code>true</code> to fill remaining area with background color, <code>false</code> otherwise.
	 */
	public void setFillColor(boolean fillColor) {
		this.fillColor = fillColor;
	}

	/**
	 * Gets whether or not the remaining area is filled with the background color.
	 *
	 * @return <code>true</code> if the remaining area is filled with the background color, <code>false</code>
	 *         otherwise.
	 */
	public boolean isFillColor() {
		return this.fillColor;
	}

	@Override
	public boolean isTransparent() {
		return false;
		// TODO WI 19217
		// if (this.fillColor) {
		// return false;
		// } else {
		// Image image = getImage();
		// if (image != null) {
		// return image.isTransparent();
		// } else {
		// // Do not fill background and no image.
		// return true;
		// }
		// }
	}

	@Override
	public void apply(GraphicsContext g, Rectangle bounds, int color) {
		// Always set color in case the image to draw is a alpha one (A1, A2, A4, A8 formats).
		g.setColor(color);
		if (this.fillColor) {
			g.fillRect(0, 0, bounds.getWidth(), bounds.getHeight());
			// TODO WI 19217 if (image.isTransparent()) {
			// If the background is filled and the image is transparent (with possible antialiasing), setting the
			// background color will optimize a bit the drawing.
			g.setBackgroundColor(color);
			// }
		}
		drawImage(g, bounds);
		g.removeBackgroundColor();
	}

	/**
	 * Draws the background image.
	 *
	 * @param g
	 *            the graphics context.
	 * @param bounds
	 *            the available rectangle.
	 */
	protected abstract void drawImage(GraphicsContext g, Rectangle bounds);

}