/*
 * Java
 *
 * Copyright 2015-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.style.border;

import ej.annotation.Nullable;
import ej.microui.display.GraphicsContext;
import ej.microui.display.shape.AntiAliasedShapes;
import ej.style.background.SimpleRoundedPlainBackground;
import ej.style.container.Rectangle;

/**
 * Draws a simple border with round corners.
 * <p>
 * May be associated with {@link SimpleRoundedPlainBackground} or used on its own.
 */
public class SimpleRoundedBorder implements Border {

	private static final int EMPTY_CIRCLE = 0;
	private static final int QUARTER_CIRCLE = 90;
	private static final int HALF_CIRCLE = 180;
	private static final int THREE_QUARTER_CIRCLE = 270;

	private int thickness;
	private int radius;

	/**
	 * Creates a simple rounded border.
	 */
	public SimpleRoundedBorder() {
		// thickness = 0; VM DONE
	}

	/**
	 * Creates a simple rounded border specifying its attributes.
	 *
	 * @param radius
	 *            the border radius to set.
	 * @param thickness
	 *            the border thickness to set.
	 */
	public SimpleRoundedBorder(int radius, int thickness) {
		this.thickness = thickness;
		this.radius = radius;
	}

	/**
	 * Sets the thickness.
	 *
	 * @param thickness
	 *            the thickness to set.
	 */
	public void setThickness(int thickness) {
		this.thickness = thickness;
	}

	/**
	 * Gets the thickness.
	 *
	 * @return the thickness.
	 */
	public int getThickness() {
		return this.thickness;
	}

	/**
	 * Sets the border radius.
	 *
	 * @param radius
	 *            the border radius to set.
	 */
	public void setRadius(int radius) {
		this.radius = radius;
	}

	/**
	 * Gets the border radius.
	 *
	 * @return the border radius.
	 */
	public int getRadius() {
		return this.radius;
	}

	@Override
	public void wrap(Rectangle rectangle) {
		int thickness = this.thickness;
		rectangle.update(-thickness, -thickness, thickness << 1, thickness << 1);
	}

	@Override
	public void unwrap(Rectangle rectangle) {
		int thickness = this.thickness;
		rectangle.update(thickness, thickness, -thickness << 1, -thickness << 1);
	}

	@Override
	public void apply(GraphicsContext g, Rectangle bounds, int color) {
		boolean hasBackgroundColor = g.hasBackgroundColor();
		int backgroundColor = g.getBackgroundColor();
		g.removeBackgroundColor();

		int width = bounds.getWidth();
		int height = bounds.getHeight();
		int thickness = this.thickness;

		boolean emptyBorder = thickness == 0;

		// Since draw ellipse arc is not available, both radius have the same size.
		int minSize = Math.min(width, height);
		if ((minSize & 0x1) == 0) {
			// Draw an even circle.
			minSize--;
		}
		int borderRadius = Math.min(this.radius, (minSize >> 1) - thickness);

		int borderSize = ((borderRadius + thickness) << 1) + 1; // +1 for previous /2 precision
		boolean halfCircleHorizontal = borderSize >= width;
		boolean halfCircleVertical = borderSize >= height;
		boolean fullCircle = halfCircleHorizontal && halfCircleVertical;

		if (!emptyBorder) {
			g.setColor(color);

			AntiAliasedShapes antiAliasedShapes = AntiAliasedShapes.Singleton;
			antiAliasedShapes.setThickness(thickness);
			antiAliasedShapes.setFade(1);

			int borderDiameter = (borderRadius << 1);
			int halfRectangle = thickness >> 1;
			int leftX = halfRectangle + 1;
			int topY = halfRectangle + 1;
			if (fullCircle) {
				antiAliasedShapes.drawCircle(g, leftX, topY, borderDiameter);
			} else {
				int rightX = leftX + width - thickness - 2 - 1;
				int bottomY = topY + height - thickness - 2 - 1;

				int rightCircleX = rightX - borderDiameter;
				int bottomCircleX = bottomY - borderDiameter;
				if (halfCircleVertical) {
					antiAliasedShapes.drawCircleArc(g, leftX, topY, borderDiameter, QUARTER_CIRCLE, HALF_CIRCLE);
					antiAliasedShapes.drawCircleArc(g, rightCircleX, topY, borderDiameter, -QUARTER_CIRCLE,
							HALF_CIRCLE);
				} else if (halfCircleHorizontal) {
					antiAliasedShapes.drawCircleArc(g, leftX, topY, borderDiameter, EMPTY_CIRCLE, HALF_CIRCLE);
					antiAliasedShapes.drawCircleArc(g, leftX, bottomCircleX, borderDiameter, HALF_CIRCLE, HALF_CIRCLE);
				} else {
					antiAliasedShapes.drawCircleArc(g, leftX, topY, borderDiameter, QUARTER_CIRCLE, QUARTER_CIRCLE);
					antiAliasedShapes.drawCircleArc(g, leftX, bottomCircleX, borderDiameter, HALF_CIRCLE,
							QUARTER_CIRCLE);
					antiAliasedShapes.drawCircleArc(g, rightCircleX, bottomCircleX, borderDiameter,
							THREE_QUARTER_CIRCLE, QUARTER_CIRCLE);
					antiAliasedShapes.drawCircleArc(g, rightCircleX, topY, borderDiameter, EMPTY_CIRCLE,
							QUARTER_CIRCLE);
				}

				// TODO: Avoid overlap of arc and line.
				if (!halfCircleHorizontal) {
					int horizontalLeftX = leftX + borderRadius + 1;
					int horizontalRightX = rightX - borderRadius;
					int bottomLineY = bottomY + 1;
					antiAliasedShapes.drawLine(g, horizontalLeftX, topY, horizontalRightX, topY);
					antiAliasedShapes.drawLine(g, horizontalLeftX, bottomLineY, horizontalRightX, bottomLineY);
				}
				if (!halfCircleVertical) {
					int verticalTopY = topY + borderRadius + 1;
					int verticalBottomY = bottomY - borderRadius;
					int rightLineX = rightX + 1;
					antiAliasedShapes.drawLine(g, leftX, verticalTopY, leftX, verticalBottomY);
					antiAliasedShapes.drawLine(g, rightLineX, verticalTopY, rightLineX, verticalBottomY);
				}
			}
		}

		g.translate(thickness, thickness);
		unwrap(bounds);
		g.clipRect(0, 0, bounds.getWidth(), bounds.getHeight());

		if (hasBackgroundColor) {
			g.setBackgroundColor(backgroundColor);
		}
	}

	@Override
	public boolean equals(@Nullable Object obj) {
		if (obj instanceof SimpleRoundedBorder) {
			SimpleRoundedBorder border = (SimpleRoundedBorder) obj;
			return this.thickness == border.thickness && this.radius == border.radius;
		}
		return false;
	}

	@Override
	public int hashCode() {
		// Since the fields are not final, they can't be used to compute a constant hashcode.
		return 0;
	}

}
