/*
 * Java
 *
 * Copyright 2015-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.style.cascading;

import java.lang.ref.Reference;
import java.lang.ref.WeakReference;
import java.util.Map;
import java.util.WeakHashMap;

import ej.annotation.Nullable;
import ej.basictool.ArrayTools;
import ej.microui.display.Colors;
import ej.microui.display.GraphicsContext;
import ej.style.Element;
import ej.style.Selector;
import ej.style.Style;
import ej.style.Stylesheet;
import ej.style.background.PlainBackground;
import ej.style.border.NoBorder;
import ej.style.dimension.NoDimension;
import ej.style.font.FontProfile;
import ej.style.outline.EmptyOutline;
import ej.style.text.SimpleTextManager;
import ej.style.util.EditableStyle;

/**
 * Cascading stylesheet implementation.
 * <p>
 * The style of an element is determined following these steps:
 * <ol>
 * <li>create an empty result style (no attributes set),</li>
 * <li>merge element with matching selectors rules (set by {@link #addRule(Selector, Style)}),</li>
 * <li>recursively merge with inherited attributes of parents (repeat previous step for parent recursively) (see
 * {@link Element#getParentElement()}),</li>
 * <li>merge global style (set by {@link #setDefaultStyle(Style)}).</li>
 * </ol>
 * The merge consists in completing the result style with all the set attributes of another style (see {@link Style}).
 * The result style is complete at the end of the resolution as long as the global style is complete (see
 * {@link Style#isComplete()}).
 */
public class CascadingStylesheet implements Stylesheet {

	/**
	 * Global style.
	 */
	private CascadingStyle style;
	/**
	 * Rules.
	 */
	// The list ensures the definition order.
	private SelectorStyle[] selectorsStyles;
	/**
	 * Cached styles.
	 */
	private Reference<CascadingStyle>[] stylesCache;
	/**
	 * Cached styles by elements.
	 */
	private final Map<Element, CascadingStyle> elementsStyles;

	/**
	 * Creates a new cascading stylesheet.
	 */
	public CascadingStylesheet() {
		this.style = createDefaultStyle();
		this.selectorsStyles = new SelectorStyle[0];
		@SuppressWarnings("unchecked")
		Reference<CascadingStyle>[] stylesCache = new Reference[0];
		this.stylesCache = stylesCache;
		this.elementsStyles = new WeakHashMap<>(1);
	}

	private CascadingStyle createDefaultStyle() {
		EditableStyle defaultStyle = new EditableStyle();
		defaultStyle.setForegroundColor(Colors.BLACK);
		defaultStyle.setBackgroundColor(Colors.WHITE);
		defaultStyle.setBorderColor(Colors.BLACK);
		defaultStyle.setBorder(NoBorder.NO_BORDER);
		defaultStyle.setFontProfile(new FontProfile());
		defaultStyle.setAlignment(GraphicsContext.LEFT | GraphicsContext.TOP);
		defaultStyle.setTextManager(new SimpleTextManager());
		defaultStyle.setDimension(NoDimension.NO_DIMENSION);
		defaultStyle.setPadding(EmptyOutline.EMPTY_OUTLINE);
		defaultStyle.setMargin(EmptyOutline.EMPTY_OUTLINE);
		defaultStyle.setBackground(new PlainBackground());
		return new CascadingStyle(defaultStyle);
	}

	@Override
	public Style getStyle(Element element) {
		CascadingStyle resultingStyle = computeStyle(element);

		CascadingStyle cachedStyle = getCachedStyle(resultingStyle);
		synchronized (this.elementsStyles) {
			this.elementsStyles.put(element, cachedStyle);
		}
		return cachedStyle;
	}

	private CascadingStyle computeStyle(Element element) {
		CascadingStyle resultingStyle = new CascadingStyle();

		// Global style selectors only.
		if (mergeSelectors(element, resultingStyle)) {
			return resultingStyle;
		}

		// Merge with parents inherited attributes.
		Element parentElement = element.getParentElement();
		if (parentElement != null) {
			Style parentStyle;
			synchronized (this.elementsStyles) {
				parentStyle = this.elementsStyles.get(parentElement);
			}
			if (parentStyle != null) {
				// Parent style has already been resolved.
				resultingStyle.inheritMerge((CascadingStyle) parentStyle);
			}
		}
		// Global.
		mergeGlobal(resultingStyle);

		return resultingStyle;
	}

	/**
	 * Search a style in the cache to avoid having several instances of the same style.
	 * <p>
	 * If an equal style is available, it is returned, otherwise the given style is cached.
	 *
	 * @param style
	 *            the style to search.
	 * @return a cached style.
	 */
	private CascadingStyle getCachedStyle(CascadingStyle style) {
		Reference<CascadingStyle>[] stylesCache = this.stylesCache;
		Reference<CascadingStyle>[] updatedCache = stylesCache;
		try {
			for (Reference<CascadingStyle> reference : stylesCache) {
				CascadingStyle cachedStyle = reference.get();
				if (cachedStyle == null) {
					// Garbage collected, remove it from the list.
					updatedCache = ArrayTools.remove(updatedCache, reference);
				} else {
					if (cachedStyle.equals(style)) {
						return cachedStyle;
					}
				}
			}
			// Not found, add it in the cache.
			updatedCache = ArrayTools.add(updatedCache, new WeakReference<CascadingStyle>(style));
		} finally {
			if (updatedCache != stylesCache) {
				this.stylesCache = updatedCache;
			}
		}

		return style;
	}

	@Override
	public void setDefaultStyle(Style style) {
		overrideIn(style);
	}

	@Override
	public void resetDefaultStyle() {
		this.style = createDefaultStyle();
	}

	@Override
	public void addRule(Selector selector, Style style) {
		// Get the array in local to avoid synchronization (with add and remove).
		SelectorStyle[] selectorsStyles = this.selectorsStyles;
		SelectorStyle selectorStyle = getSelectorStyle(selector, selectorsStyles);
		CascadingStyle registeredStyle;
		if (selectorStyle == null) {
			registeredStyle = new CascadingStyle(style);
			selectorStyle = new SelectorStyle(selector, registeredStyle);
			int index = getIndex(selector, selectorsStyles);
			// Add the selector at the first place to ensure order (last added is resolved first).
			this.selectorsStyles = ArrayTools.insert(selectorsStyles, index, selectorStyle);
		} else {
			registeredStyle = selectorStyle.style;
			registeredStyle.override(style);
		}
	}

	@Override
	public Style getRuleStyle(Selector selector) {
		SelectorStyle[] selectorsStyles = this.selectorsStyles;
		SelectorStyle selectorStyle = getSelectorStyle(selector, selectorsStyles);
		if (selectorStyle != null) {
			return new Style(selectorStyle.style);
		}
		return new Style();
	}

	@Override
	public void removeRule(Selector selector) {
		// Get the array in local to avoid synchronization (with add and remove).
		SelectorStyle[] selectorsStyles = this.selectorsStyles;
		SelectorStyle selectorStyle = getSelectorStyle(selector, selectorsStyles);
		if (selectorStyle != null) {
			this.selectorsStyles = ArrayTools.remove(selectorsStyles, selectorStyle);
		}
	}

	@Override
	public void reset() {
		this.style = createDefaultStyle();
		this.selectorsStyles = new SelectorStyle[0];
	}

	private void overrideIn(Style style) {
		this.style.override(style);
	}

	private int getIndex(Selector selector, SelectorStyle[] selectorsStyles) {
		int specificity = selector.getSpecificity();
		int minIndex = 0;
		int maxIndex = selectorsStyles.length;
		while (maxIndex - minIndex > 1) {
			int currentIndex = (maxIndex - minIndex) / 2 + minIndex;
			if (isMoreSpecific(selectorsStyles, currentIndex, specificity)) {
				minIndex = currentIndex;
			} else {
				maxIndex = currentIndex;
			}
		}
		if (maxIndex > minIndex) {
			if (isMoreSpecific(selectorsStyles, minIndex, specificity)) {
				return maxIndex;
			} else {
				return minIndex;
			}
		}
		return minIndex;
	}

	private boolean isMoreSpecific(SelectorStyle[] selectorsStyles, int index, int specificity) {
		int specificityCandidate = selectorsStyles[index].selector.getSpecificity();
		// If equals, the first added remains first.
		return specificityCandidate > specificity;
	}

	private boolean mergeSelectors(Element element, CascadingStyle resultingStyle) {
		for (SelectorStyle selectorStyle : this.selectorsStyles) {
			if (selectorStyle.selector.fit(element)) {
				CascadingStyle stateStyle = selectorStyle.style;
				if (resultingStyle.merge(stateStyle)) {
					return true;
				}
			}
		}
		return false;
	}

	private boolean mergeGlobal(CascadingStyle resultingStyle) {
		return resultingStyle.merge(this.style);
	}

	@Nullable
	private SelectorStyle getSelectorStyle(Selector selector, SelectorStyle[] selectorsStyles) {
		for (SelectorStyle selectorStyle : selectorsStyles) {
			if (selector.equals(selectorStyle.selector)) {
				return selectorStyle;
			}
		}
		return null;
	}

}
