/*
 * Java
 *
 * Copyright 2015-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.style.container;

import ej.microui.display.GraphicsContext;

/**
 * Provides some alignment facilities.
 */
public class AlignmentHelper {

	// Prevents initialization.
	private AlignmentHelper() {
	}

	/**
	 * Verify whether an alignment bit-field is valid or not. Equivalent to calling
	 * {@link #checkHorizontalAlignment(int)} and {@link #checkVerticalAlignment(int)} sequentially.
	 *
	 * @param anchor
	 *            the anchor to check.
	 * @throws IllegalArgumentException
	 *             if the alignment is not valid.
	 */
	public static void validateAlignment(int anchor) {
		if (!checkAlignment(anchor)) {
			throw new IllegalArgumentException();
		}
	}

	/**
	 * Checks whether an alignment bit-field is valid or not. Equivalent to calling
	 * {@link #checkHorizontalAlignment(int)} and {@link #checkVerticalAlignment(int)} sequentially.
	 *
	 * @param anchor
	 *            the anchor to check.
	 * @return <code>true</code> if the given alignment is valid, <code>false</code> otherwise.
	 */
	public static boolean checkAlignment(int anchor) {
		return checkHorizontalAlignment(anchor) && checkVerticalAlignment(anchor);
	}

	/**
	 * Checks whether an alignment bit-field is valid horizontally or not. That means it contains one and only one of
	 * {@link GraphicsContext#LEFT}, {@link GraphicsContext#RIGHT} or {@link GraphicsContext#HCENTER}.
	 *
	 * @param anchor
	 *            the anchor to check.
	 * @return <code>true</code> if the given alignment is valid horizontally, <code>false</code> otherwise.
	 */
	public static boolean checkHorizontalAlignment(int anchor) {
		int isSet = 0;
		if ((anchor & GraphicsContext.HCENTER) != 0) {
			++isSet;
		}
		if ((anchor & GraphicsContext.RIGHT) != 0) {
			++isSet;
		}
		if ((anchor & GraphicsContext.LEFT) != 0) {
			++isSet;
		}
		return isSet == 1;
	}

	/**
	 * Checks whether an alignment bit-field is valid vertically or not. That means it contains one and only one of
	 * {@link GraphicsContext#TOP}, {@link GraphicsContext#BOTTOM} or {@link GraphicsContext#VCENTER}.
	 *
	 * @param anchor
	 *            the anchor to check.
	 * @return <code>true</code> if the given alignment is valid vertically, <code>false</code> otherwise.
	 */
	public static boolean checkVerticalAlignment(int anchor) {
		int isSet = 0;
		if ((anchor & GraphicsContext.VCENTER) != 0) {
			++isSet;
		}
		if ((anchor & GraphicsContext.TOP) != 0) {
			++isSet;
		}
		if ((anchor & GraphicsContext.BOTTOM) != 0) {
			++isSet;
		}
		return isSet == 1;
	}

	/**
	 * Computes the x coordinate of the left edge of an element in an area.
	 * <p>
	 * The result is not defined if the anchor is invalid.
	 *
	 * @param width
	 *            the element width.
	 * @param areaX
	 *            the area x coordinate.
	 * @param areaWidth
	 *            the area width.
	 * @param anchor
	 *            the anchor.
	 * @return the x coordinate of the left of the element.
	 * @see #checkHorizontalAlignment(int)
	 */
	public static int computeXLeftCorner(int width, int areaX, int areaWidth, int anchor) {
		if ((anchor & GraphicsContext.HCENTER) != 0) {
			return areaX + ((areaWidth - width) >> 1);
		} else if ((anchor & GraphicsContext.RIGHT) != 0) {
			return areaX + (areaWidth - width);
		} else { // left or undefined
			return areaX;
		}
	}

	/**
	 * Computes the y coordinate of the top edge of an element in an area.
	 * <p>
	 * The result is not defined if the anchor is invalid.
	 *
	 * @param height
	 *            the element height.
	 * @param areaY
	 *            the area y coordinate.
	 * @param areaHeight
	 *            the area height.
	 * @param anchor
	 *            the anchor.
	 * @return the y coordinate of the top of the element.
	 * @throws IllegalArgumentException
	 *             if the anchor is invalid.
	 * @see #checkVerticalAlignment(int)
	 */
	public static int computeYTopCorner(int height, int areaY, int areaHeight, int anchor) {
		if ((anchor & GraphicsContext.VCENTER) != 0) {
			return areaY + ((areaHeight - height) >> 1);
		} else if ((anchor & GraphicsContext.BOTTOM) != 0) {
			return areaY + (areaHeight - height);
		} else { // top or undefined
			return areaY;
		}
	}

}
