/*
 * Java
 *
 * Copyright 2015-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.style.selector;

import ej.annotation.Nullable;
import ej.style.Element;
import ej.style.Selector;

/**
 * An attribute value prefix selector selects by checking that the value of an attribute equals a prefix or starts with
 * this prefix.
 * <p>
 * Checks that the attribute value is exactly the given value or starts with the given value immediately followed by "-"
 * (U+002D).
 * <p>
 * Equivalent to <code>[attribute|=value]</code> selector in CSS. Its specificity is (0,0,1,0).
 *
 * @see Element#getAttribute(String)
 * @see SpecificityHelper
 */
public class AttributeValuePrefixSelector implements Selector {

	private static final char SEPARATOR = '-';
	private final String attribute;
	private final String value;

	/**
	 * Creates an attribute value prefix selector.
	 *
	 * @param attribute
	 *            the attribute to check.
	 * @param value
	 *            the value to check.
	 */
	public AttributeValuePrefixSelector(String attribute, String value) {
		this.attribute = attribute;
		this.value = value;
	}

	@Override
	public boolean fit(Element element) {
		String value = element.getAttribute(this.attribute);
		return this.value.equals(value) || (value != null && value.startsWith(this.value + SEPARATOR));
	}

	@Override
	public int getSpecificity() {
		return SpecificityHelper.getSpecificity(0, 0, 1, 0);
	}

	@Override
	public boolean equals(@Nullable Object obj) {
		if (obj instanceof AttributeValuePrefixSelector) {
			AttributeValuePrefixSelector other = (AttributeValuePrefixSelector) obj;
			return this.attribute.equals(other.attribute) && this.value.equals(other.value);
		}
		return false;
	}

	@Override
	public int hashCode() {
		return this.attribute.hashCode() * 31 + this.value.hashCode();
	}

}
