/*
 * Java
 *
 * Copyright 2015-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.style.selector;

import ej.annotation.Nullable;
import ej.style.Element;
import ej.style.Selector;
import ej.style.State;

/**
 * A state selector selects from a state. That means that the state of this selector is available in the element states.
 * <p>
 * Equivalent to pseudo-class (eg. <code>:active</code>) selector in CSS. Its specificity is (0,0,1,0).
 *
 * @see State
 * @see Element#isInState(State)
 * @see SpecificityHelper
 */
public class StateSelector implements Selector {

	private final State state;

	/**
	 * Creates a state selector with one state.
	 *
	 * @param state
	 *            the state.
	 */
	public StateSelector(State state) {
		this.state = state;
	}

	@Override
	public boolean fit(Element element) {
		// Check that the state of this selector is available in the given element.
		return element.isInState(this.state);
	}

	@Override
	public int getSpecificity() {
		return SpecificityHelper.getSpecificity(0, 0, 1, 0);
	}

	@Override
	public boolean equals(@Nullable Object obj) {
		if (obj instanceof StateSelector) {
			StateSelector other = (StateSelector) obj;
			return this.state == other.state;
		}
		return false;
	}

	@Override
	public int hashCode() {
		return this.state.hashCode();
	}
}
