/*
 * Java
 *
 * Copyright 2015-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.style.selector.combinator;

import ej.style.Element;
import ej.style.Selector;

/**
 * A child combinator selects by checking the element direct hierarchy.
 * <p>
 * Equivalent to <code>element&gt;element</code> selector in CSS.
 *
 * @see Element#getParentElement()
 */
public class ChildCombinator extends Combinator {

	/**
	 * Creates a child combinator.
	 *
	 * @param selector1
	 *            the first child.
	 * @param selector2
	 *            the second child.
	 */
	public ChildCombinator(Selector selector1, Selector selector2) {
		super(selector1, selector2);
	}

	/**
	 * Creates a child combinator.
	 *
	 * @param selectors
	 *            the list of selectors.
	 * @throws IllegalArgumentException
	 *             if there is less than two selectors.
	 */
	public ChildCombinator(Selector... selectors) {
		super(selectors);
	}

	@Override
	public boolean fit(Element element) {
		Selector[] selectors = this.selectors;
		int size = selectors.length;
		int selectorIndex = size - 1;
		while (true) {
			Selector ancestorSelector = selectors[selectorIndex--];
			if (!ancestorSelector.fit(element)) {
				return false;
			}

			if (selectorIndex != -1) {
				Element parent = element.getParentElement();
				if (parent != null) {
					element = parent;
				} else {
					// It remains some parents to check, but we cannot go upward in the hierarchy.
					return false;
				}
			} else {
				return true;
			}
		}
	}

}
