/*
 * Java
 *
 * Copyright 2015-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.style.util;

import ej.annotation.NonNull;
import ej.microui.display.GraphicsContext;
import ej.style.Style;
import ej.style.background.Background;
import ej.style.border.Border;
import ej.style.container.AlignmentHelper;
import ej.style.dimension.Dimension;
import ej.style.font.FontProfile;
import ej.style.outline.Outline;
import ej.style.text.TextManager;

/**
 * The attributes of an editable style can be set.
 * <p>
 * An editable style is used to create rules in a stylesheet.
 * <p>
 * A stylesheet does not keep a reference to the style of a rule. It means that:
 * <ul>
 * <li>any modification on a style after being set in a rule does not modify the rule,</li>
 * <li>the same instance of editable style can be used for several rules (by using the {@link #clear()} method between
 * each one for example).</li>
 * </ul>
 */
public class EditableStyle extends Style {

	/**
	 * Creates an editable style.
	 *
	 * @see Style#Style()
	 */
	public EditableStyle() {
		super();
	}

	/**
	 * Creates an editable style from another one.
	 *
	 * @param style
	 *            the style to copy.
	 */
	public EditableStyle(@NonNull Style style) {
		super(style);
	}

	/**
	 * Clears all the attributes of the style.
	 */
	public void clear() {
		this.foregroundColor = 0;
		this.backgroundColor = 0;
		this.border = null;
		this.fontProfile = null;
		this.alignment = 0;
		this.textManager = null;
		this.dimension = null;
		this.padding = null;
		this.margin = null;
		this.background = null;
		this.map = 0;
	}

	/**
	 * Sets the foreground color.
	 *
	 * @param foregroundColor
	 *            the foreground color to set.
	 */
	public void setForegroundColor(int foregroundColor) {
		this.map |= 0x1 << FOREGROUND_COLOR_SHIFT;
		this.foregroundColor = foregroundColor;
	}

	/**
	 * Sets the background color.
	 *
	 * @param backgroundColor
	 *            the background color to set.
	 */
	public void setBackgroundColor(int backgroundColor) {
		this.map |= 0x1 << BACKGROUND_COLOR_SHIFT;
		this.backgroundColor = backgroundColor;
	}

	/**
	 * Sets the border color.
	 *
	 * @param borderColor
	 *            the border color to set.
	 */
	public void setBorderColor(int borderColor) {
		this.map |= 0x1 << BORDER_COLOR_SHIFT;
		this.borderColor = borderColor;
	}

	/**
	 * Sets the border.
	 *
	 * @param border
	 *            the border to set.
	 */
	public void setBorder(@NonNull Border border) {
		this.map |= 0x1 << BORDER_SHIFT;
		this.border = border;
	}

	/**
	 * Sets the font profile.
	 *
	 * @param fontProfile
	 *            the font profile to set.
	 */
	public void setFontProfile(@NonNull FontProfile fontProfile) {
		this.map |= 0x1 << FONT_PROFILE_SHIFT;
		this.fontProfile = fontProfile;
	}

	/**
	 * Sets the content alignment.
	 *
	 * Alignment must be a combination of exactly one horizontal constant (taken from &#123;
	 * {@link GraphicsContext#LEFT}, {@link GraphicsContext#HCENTER}, {@link GraphicsContext#RIGHT} &#125; ) and exactly
	 * one vertical constant (taken from &#123; {@link GraphicsContext#TOP}, {@link GraphicsContext#VCENTER},
	 * {@link GraphicsContext#BOTTOM} &#125; ). Default alignment matches the ({@link GraphicsContext#TOP} |
	 * {@link GraphicsContext#LEFT}) constant combination.
	 *
	 * @param alignment
	 *            the alignment to set.
	 * @throws IllegalArgumentException
	 *             if the alignment is not valid.
	 * @see AlignmentHelper#validateAlignment(int)
	 */
	public void setAlignment(int alignment) {
		AlignmentHelper.validateAlignment(alignment);
		this.map |= 0x1 << ALIGNMENT_SHIFT;
		this.alignment = alignment;
	}

	/**
	 * Sets the text manager.
	 *
	 * @param textManager
	 *            the text manager to set.
	 */
	public void setTextManager(@NonNull TextManager textManager) {
		this.map |= 0x1 << TEXT_MANAGER_SHIFT;
		this.textManager = textManager;
	}

	/**
	 * Sets the dimension.
	 *
	 * @param dimension
	 *            the dimension to set.
	 */
	public void setDimension(@NonNull Dimension dimension) {
		this.map |= 0x1 << DIMENSION_SHIFT;
		this.dimension = dimension;
	}

	/**
	 * Sets the padding.
	 *
	 * @param padding
	 *            the padding to set.
	 */
	public void setPadding(@NonNull Outline padding) {
		this.map |= 0x1 << PADDING_SHIFT;
		this.padding = padding;
	}

	/**
	 * Sets the margin.
	 *
	 * @param margin
	 *            the margin to set.
	 */
	public void setMargin(@NonNull Outline margin) {
		this.map |= 0x1 << MARGIN_SHIFT;
		this.margin = margin;
	}

	/**
	 * Sets the background.
	 *
	 * @param background
	 *            the background to set.
	 */
	public void setBackground(@NonNull Background background) {
		this.map |= 0x1 << BACKGROUND_SHIFT;
		this.background = background;
	}

}
