/*
 * Java
 *
 * Copyright 2015-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.style.util;

import ej.annotation.NonNull;
import ej.annotation.Nullable;
import ej.components.dependencyinjection.ServiceLoaderFactory;
import ej.microui.display.Font;
import ej.microui.display.GraphicsContext;
import ej.microui.display.Image;
import ej.style.Style;
import ej.style.Stylesheet;
import ej.style.background.Background;
import ej.style.border.Border;
import ej.style.cascading.CascadingStylesheet;
import ej.style.container.AlignmentHelper;
import ej.style.container.Rectangle;
import ej.style.dimension.Dimension;
import ej.style.font.FontProfile;
import ej.style.font.loader.DefaultFontLoader;
import ej.style.font.loader.FontLoader;
import ej.style.image.DefaultImageLoader;
import ej.style.image.ImageLoader;
import ej.style.outline.Outline;

/**
 * Convenient class to use style.
 *
 * @see Style
 */
public class StyleHelper {

	@Nullable
	private static Stylesheet Stylesheet;
	@Nullable
	private static FontLoader FontLoader;
	@Nullable
	private static ImageLoader ImageLoader;

	// Prevents initialization.
	private StyleHelper() {
	}

	/**
	 * Gets the stylesheet instance.
	 *
	 * @return the stylesheet instance.
	 */
	public static Stylesheet getStylesheet() {
		Stylesheet stylesheet = Stylesheet;
		if (stylesheet == null) {
			stylesheet = ServiceLoaderFactory.getServiceLoader().getService(Stylesheet.class,
					CascadingStylesheet.class);
			Stylesheet = stylesheet;
		}
		return stylesheet;
	}

	/**
	 * Utility method to get the font matching a front profile from a style using the font loader.
	 *
	 * @param style
	 *            the style that contains the font profile.
	 * @return the retrieved font.
	 */
	public static Font getFont(@NonNull Style style) {
		FontProfile fontProfile = style.getFontProfile();
		return getFont(fontProfile);
	}

	/**
	 * Utility method to get the font matching a font profile using the font loader.
	 *
	 * @param fontProfile
	 *            the font profile.
	 * @return the retrieved font.
	 */
	public static Font getFont(@NonNull FontProfile fontProfile) {
		FontLoader fontLoader = getFontLoader();
		if (fontLoader != null) {
			Font font = fontLoader.getFont(fontProfile);
			return font;
		}
		return getDefaultFont();
	}

	private static Font getDefaultFont() {
		@SuppressWarnings("null")
		@NonNull
		Font defaultFont = Font.getDefaultFont();
		return defaultFont;
	}

	@Nullable
	private static FontLoader getFontLoader() {
		if (FontLoader == null) {
			FontLoader = ServiceLoaderFactory.getServiceLoader().getService(FontLoader.class, DefaultFontLoader.class);
		}
		return FontLoader;
	}

	/**
	 * Utility method to get the image from a style using the image loader.
	 *
	 * @param path
	 *            the path of the image to load.
	 * @return the retrieved image.
	 */
	@Nullable
	public static Image getImage(@NonNull String path) {
		ImageLoader imageLoader = getImageLoader();
		if (imageLoader != null) {
			Image image = imageLoader.getImage(path);
			return image;
		}
		return null;
	}

	@Nullable
	private static ImageLoader getImageLoader() {
		if (ImageLoader == null) {
			ImageLoader = ServiceLoaderFactory.getServiceLoader().getService(ImageLoader.class,
					DefaultImageLoader.class);
		}
		return ImageLoader;
	}

	/**
	 * Unwraps the different style outlines then apply the style dimension to get the content effective bounds from the
	 * global bounds.
	 * <p>
	 * The bounds hint is modified.
	 * <p>
	 * Precondition: the given style is complete.
	 *
	 * @param boundsHint
	 *            the bounds hint.
	 * @param style
	 *            the style to apply.
	 * @return the unwrapped bounds.
	 * @see Outline#unwrap(Rectangle)
	 */
	@NonNull
	public static Rectangle computeContentBounds(@NonNull Rectangle boundsHint, Style style) {
		// … minus margin size…
		Outline margin = style.getMargin();
		margin.unwrap(boundsHint);

		// … minus border size…
		Border border = style.getBorder();
		border.unwrap(boundsHint);

		// … minus padding size…
		Outline padding = style.getPadding();
		padding.unwrap(boundsHint);

		// … then apply the dimension…
		Dimension dimension = style.getDimension();
		Rectangle contentBounds = new Rectangle(boundsHint);
		dimension.apply(contentBounds, boundsHint);

		// … finally align the content if the dimension has changed the bounds.
		int alignment = style.getAlignment();
		int x = AlignmentHelper.computeXLeftCorner(contentBounds.getWidth(), boundsHint.getX(), boundsHint.getWidth(),
				alignment);
		int y = AlignmentHelper.computeYTopCorner(contentBounds.getHeight(), boundsHint.getY(), boundsHint.getHeight(),
				alignment);
		contentBounds.setX(x);
		contentBounds.setY(y);

		return contentBounds;
	}

	/**
	 * Applies the style dimension then wraps the different style outlines to get the global bounds from the content
	 * effective bounds.
	 * <p>
	 * The content bounds is modified.
	 * <p>
	 * Precondition: the given style is complete.
	 *
	 * @param boundsHint
	 *            the bounds hint.
	 * @param contentBounds
	 *            the content bounds.
	 * @param style
	 *            the style to apply.
	 * @see Outline#wrap(Rectangle)
	 */
	public static void computePreferredSize(@NonNull Rectangle boundsHint, @NonNull Rectangle contentBounds,
			Style style) {
		// Apply the dimension…
		Dimension dimension = style.getDimension();
		dimension.apply(contentBounds, boundsHint);

		// … plus padding size…
		Outline padding = style.getPadding();
		padding.wrap(contentBounds);

		// … plus border size…
		Border border = style.getBorder();
		border.wrap(contentBounds);

		// … plus margin size…
		Outline margin = style.getMargin();
		margin.wrap(contentBounds);
	}

	/**
	 * Applies the different outlines and the background to a graphics context.
	 * <p>
	 * The order is: margin, background, border, padding.
	 * <p>
	 * Precondition: the given style is complete.
	 *
	 * @param g
	 *            the graphics context to modify.
	 * @param bounds
	 *            the available bounds.
	 * @param style
	 *            the style to apply.
	 * @see Outline#apply(GraphicsContext, Rectangle)
	 * @see Background#apply(GraphicsContext, Rectangle, int)
	 */
	public static void renderWithoutContent(@NonNull GraphicsContext g, @NonNull Rectangle bounds,
			@NonNull Style style) {
		int boundsX = bounds.getX();
		int boundsY = bounds.getY();
		int boundsWidth = bounds.getWidth();
		int boundsHeight = bounds.getHeight();

		Outline margin = style.getMargin();
		Border border = style.getBorder();
		Outline padding = style.getPadding();
		Dimension dimension = style.getDimension();

		// Unwrap then re-wrap to apply correctly the given dimension.

		// … minus margin size…
		margin.unwrap(bounds);

		// … minus border size…
		border.unwrap(bounds);

		// … minus padding size…
		padding.unwrap(bounds);

		// … then apply the dimension…
		dimension.apply(bounds, bounds);

		// … plus padding size…
		padding.wrap(bounds);

		// … plus border size…
		border.wrap(bounds);

		// … plus margin size…
		margin.wrap(bounds);

		int alignment = style.getAlignment();
		int x = AlignmentHelper.computeXLeftCorner(bounds.getWidth(), boundsX, boundsWidth, alignment);
		int y = AlignmentHelper.computeYTopCorner(bounds.getHeight(), boundsY, boundsHeight, alignment);
		bounds.setX(x);
		bounds.setY(y);
		g.translate(x - boundsX, y - boundsY);

		// Apply margin.
		margin.apply(g, bounds);

		// Apply background.
		Background background = style.getBackground();
		background.apply(g, bounds, style.getBackgroundColor());

		// Apply border.
		border.apply(g, bounds, style.getBorderColor());

		// Apply padding.
		padding.apply(g, bounds);
	}
}
