/*
 * Java
 *
 * Copyright 2015-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.wadapps.admin;

import java.io.InputStream;

import org.eclipse.core.runtime.IProgressMonitor;

/**
 * An application represents a deployment unit.
 * <p>
 * Two applications are equals if and only if they have the same identifier.
 */
public interface Application {

	/**
	 * Represents an application state.
	 */
	public static enum State {
		/**
		 * An application in the {@link #INSTALLED} state has been successfully
		 * installed and is not running.
		 */
		INSTALLED,

		/**
		 * An application in the {@link #STARTED} state has been started and is running.
		 */
		STARTED,

		/**
		 * An application in the {@link #UNINSTALLED} state has been uninstalled. All
		 * application methods except {@link Application#getState()} throw an
		 * {@link IllegalStateException}
		 */
		UNINSTALLED,
	}

	/**
	 * Gets the application identifier.
	 * <p>
	 * The identifier remains the same even between different application version.
	 *
	 * @return the application identifier.
	 */
	String getIdentifier();

	/**
	 * Gets the application version.
	 *
	 * @return the application version.
	 */
	String getVersion();

	/**
	 * Gets an input stream for reading the specified resource from the application.
	 *
	 * @param name
	 *            the resource name.
	 *
	 * @return an input stream for reading the resource, or <code>null</code> if the
	 *         resource could not be found.
	 * @throws IllegalStateException
	 *             if the application state is not {@link State#STARTED} state.
	 */
	InputStream getResourceAsStream(String name) throws IllegalStateException;

	/**
	 * Gets an application property.
	 *
	 * @param key
	 *            the property key to search.
	 * @return the requested property or <code>null</code> if not existing.
	 * @throws IllegalStateException
	 *             if the application state is not {@link State#STARTED} state.
	 */
	String getProperty(String key) throws IllegalStateException;

	/**
	 * Returns the current application state.
	 *
	 * @return the application state.
	 * @see State
	 */
	State getState();

	/**
	 * Causes this application to be installed.
	 *
	 * @throws IllegalStateException
	 *             if the application state is not {@link State#UNINSTALLED} state.
	 * @throws SecurityException
	 *             if a security manager exits and if the caller does not have the
	 *             appropriate permissions.
	 * @throws ApplicationOperationException
	 *             If any unexpected errors occur while installing this application.
	 */
	@Deprecated
	void install() throws IllegalStateException, SecurityException, ApplicationOperationException;

	/**
	 * Causes this application to be installed.
	 *
	 * @param monitor
	 *            the progress monitor to use for reporting progress to the user. It
	 *            is the caller's responsibility to call done() on the given
	 *            monitor. Accepts <code>null</code>, indicating that no progress
	 *            should be reported and that the operation cannot be cancelled.
	 *
	 * @throws IllegalStateException
	 *             if the application state is not {@link State#UNINSTALLED} state.
	 * @throws SecurityException
	 *             if a security manager exits and if the caller does not have the
	 *             appropriate permissions.
	 * @throws ApplicationOperationException
	 *             If any unexpected errors occur while installing this application.
	 */
	@Deprecated
	void install(IProgressMonitor monitor)
			throws IllegalStateException, SecurityException, ApplicationOperationException;

	/**
	 * Causes this application to start. A new thread owned by the application is
	 * created and started. The application is switched in the {@link State#STARTED}
	 * state and this method returns.
	 *
	 * @throws IllegalStateException
	 *             if the application state is not {@link State#INSTALLED} state.
	 * @throws SecurityException
	 *             if a security manager exits and if the caller does not have the
	 *             appropriate permissions.
	 * @throws ApplicationOperationException
	 *             If any unexpected errors occur while starting this application.
	 */
	void start() throws IllegalStateException, SecurityException, ApplicationOperationException;

	/**
	 * Causes this application to start. A new thread owned by the application is
	 * created and started. The application is switched in the {@link State#STARTED}
	 * state and this method returns.
	 *
	 * @param monitor
	 *            the progress monitor to use for reporting progress to the user. It
	 *            is the caller's responsibility to call done() on the given
	 *            monitor. Accepts <code>null</code>, indicating that no progress
	 *            should be reported and that the operation cannot be cancelled.
	 *
	 * @throws IllegalStateException
	 *             if the application state is not {@link State#INSTALLED} state.
	 * @throws SecurityException
	 *             if a security manager exits and if the caller does not have the
	 *             appropriate permissions.
	 * @throws ApplicationOperationException
	 *             If any unexpected errors occur while starting this application.
	 */
	void start(IProgressMonitor monitor) throws IllegalStateException, SecurityException, ApplicationOperationException;

	/**
	 * Causes this application to stop. The application is switched in the
	 * {@link State#INSTALLED} state and this method returns.
	 *
	 * @throws IllegalStateException
	 *             if the application is in the {@link State#INSTALLED} or
	 *             {@link State#UNINSTALLED} state.
	 * @throws SecurityException
	 *             if a security manager exits and if the caller does not have the
	 *             appropriate permissions.
	 * @throws ApplicationOperationException
	 *             If any unexpected errors occur while stopping this application.
	 */
	void stop() throws IllegalStateException, SecurityException, ApplicationOperationException;

	/**
	 * Causes the application to be uninstalled.
	 *
	 * @throws IllegalStateException
	 *             if the application is not in the {@link State#INSTALLED} state.
	 * @throws SecurityException
	 *             if a security manager exits and if the caller does not have the
	 *             appropriate permissions.
	 * @throws ApplicationOperationException
	 *             If any unexpected errors occur while uninstalling this
	 *             application.
	 */
	void uninstall() throws IllegalStateException, SecurityException, ApplicationOperationException;

}
