/*
 * Java
 *
 * Copyright 2019 IS2T. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break IS2T warranties on the whole library.
 */
package ej.wadapps.common.util;

import ej.service.MissingServiceException;
import ej.service.ServiceFactory;
import ej.util.message.Level;
import ej.wadapps.application.Application;
import ej.wadapps.application.ApplicationManager;
import ej.wadapps.application.ApplicationOperationException;

/**
 * Contains tools to start an application.
 */
public class StartupHelper {

	private StartupHelper() {
	}

	/**
	 * Starts all the applications available in the application manager registered in the service loader.
	 *
	 * @throws MissingServiceException
	 *             if no application manager is registered.
	 * @see ServiceFactory#getServiceLoader()
	 */
	public static void startAllApplications() {
		ApplicationManager applicationManager = ServiceFactory.getRequiredService(ApplicationManager.class);
		startAllApplications(applicationManager);
	}

	/**
	 * Starts all the applications available in an application manager.
	 *
	 * @param applicationManager
	 *            the application manager that contains the applications.
	 */
	public static void startAllApplications(ApplicationManager applicationManager) {
		Application[] applications = applicationManager.getApplications();
		for (Application application : applications) {
			try {
				application.start();
			} catch (IllegalStateException | SecurityException | ApplicationOperationException e) {
				WadappsCommonMessage.LOGGER.log(Level.WARNING, WadappsCommonMessage.CATEGORY,
						WadappsCommonMessage.USER_ERROR, e);
			}
		}
	}

	/**
	 * Stops all the applications available in the application manager registered in the service loader.
	 *
	 * @throws MissingServiceException
	 *             if no application manager is registered.
	 * @see ServiceFactory#getServiceLoader()
	 */
	public static void stopAllApplications() {
		ApplicationManager applicationManager = ServiceFactory.getRequiredService(ApplicationManager.class);
		stopAllApplications(applicationManager);
	}

	/**
	 * Stops all the applications available in an application manager.
	 *
	 * @param applicationManager
	 *            the application manager that contains the applications.
	 */
	public static void stopAllApplications(ApplicationManager applicationManager) {
		Application[] applications = applicationManager.getApplications();
		for (Application application : applications) {
			try {
				application.stop();
			} catch (IllegalStateException | SecurityException | ApplicationOperationException e) {
				WadappsCommonMessage.LOGGER.log(Level.WARNING, WadappsCommonMessage.CATEGORY,
						WadappsCommonMessage.USER_ERROR, e);
			}
		}
	}

}
