/*
 * Java
 *
 * Copyright 2019-2020 IS2T. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break IS2T warranties on the whole library.
 */
package ej.wadapps.multisandbox.application;

import java.io.IOException;
import java.io.InputStream;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.NullProgressMonitor;

import ej.annotation.Nullable;
import ej.basictool.ArrayTools;
import ej.kf.AlreadyLoadedFeatureException;
import ej.kf.Feature;
import ej.kf.Feature.State;
import ej.kf.FeatureStateListener;
import ej.kf.IncompatibleFeatureException;
import ej.kf.InvalidFormatException;
import ej.kf.Kernel;
import ej.util.message.Level;
import ej.wadapps.application.Application;
import ej.wadapps.application.ApplicationInstallationListener;
import ej.wadapps.application.ApplicationInstaller;
import ej.wadapps.application.ApplicationOperationException;
import ej.wadapps.application.ApplicationPermission;
import ej.wadapps.multisandbox.util.WadappsMultiSandboxMessage;

/**
 * The application installer allows to install and uninstall applications.
 */
/* package */ abstract class AbstractApplicationInstallerKF implements ApplicationInstaller, FeatureStateListener {

	/**
	 * Install task name
	 */
	protected static final String INSTALL = "Install"; //$NON-NLS-1$
	/**
	 * Uninstall task name
	 */
	protected static final String UNINSTALL = "Uninstall"; //$NON-NLS-1$

	/**
	 * the application manager to update when applications are installed or uninstalled
	 */
	protected final ApplicationManagerKF applicationManager;
	private ApplicationInstallationListener[] listeners;

	/**
	 * Creates an application installer on KF.
	 *
	 * @param applicationManager
	 *            the applications manager it depends on.
	 */
	public AbstractApplicationInstallerKF(ApplicationManagerKF applicationManager) {
		this.applicationManager = applicationManager;
		this.listeners = new ApplicationInstallationListener[0];
		Kernel.addFeatureStateListener(this);
	}

	/**
	 * Gets the application manager.
	 *
	 * @return the application manager.
	 */
	public ApplicationManagerKF getApplicationManager() {
		return this.applicationManager;
	}

	@Override
	public Application install(InputStream stream) throws ApplicationOperationException {
		return install(stream, new NullProgressMonitor());
	}

	/**
	 * Installs an application into KF.
	 *
	 * @param stream
	 *            the stream containing the application.
	 * @param monitor
	 *            the progress monitor to use for reporting progress to the user. It is the caller's responsibility to
	 *            call done() on the given monitor.
	 * @return the application instance.
	 * @throws ApplicationOperationException
	 *             if any unexpected errors occur while installing this application.
	 */
	protected Application installKF(InputStream stream, IProgressMonitor monitor) throws ApplicationOperationException {
		Kernel.enter();
		try {
			Feature feature = Kernel.install(stream);
			monitor.worked(1);
			Application application = this.applicationManager.addApplication(feature);
			monitor.worked(1);
			return application;
		} catch (IOException | InvalidFormatException | IncompatibleFeatureException
				| AlreadyLoadedFeatureException e) {
			throw new ApplicationOperationException(e);
		}
		// Kernel.exit();
	}

	@Override
	public boolean uninstall(Application application) throws ApplicationOperationException {
		return uninstall(application, new NullProgressMonitor());
	}

	/**
	 * Uninstalls an application from KF.
	 *
	 * @param application
	 *            the application to uninstall.
	 * @param monitor
	 *            the progress monitor to use for reporting progress to the user. It is the caller's responsibility to
	 *            call done() on the given monitor.
	 * @return {@code true} if the application has been uninstalled, {@code false} otherwise.
	 * @throws ApplicationOperationException
	 *             if any unexpected errors occur while uninstalling this application.
	 */
	protected boolean uninstallKF(Application application, IProgressMonitor monitor)
			throws ApplicationOperationException {
		Kernel.enter();
		try {
			Feature feature = this.applicationManager.getFeature(application);
			if (feature != null) {
				if (application.isStarted()) {
					application.stop();
				}
				monitor.worked(1);
				Kernel.uninstall(feature);
				monitor.worked(1);
				return feature.getState() == State.UNINSTALLED;
			} else {
				throw new ApplicationOperationException();
			}
		} catch (IllegalStateException e) {
			throw new ApplicationOperationException(e);
		}
		// Kernel.exit();
	}

	@Override
	public void addApplicationInstallationListener(ApplicationInstallationListener applicationInstallationListener) {
		removeApplicationInstallationListener(applicationInstallationListener);
		Kernel.enter();
		synchronized (this) {
			this.listeners = ArrayTools.add(this.listeners, applicationInstallationListener);
		}
		// Kernel.exit();
	}

	@Override
	public void removeApplicationInstallationListener(ApplicationInstallationListener applicationInstallationListener) {
		Kernel.enter();
		// Iterative process and test with == (do not call equals() on 2 objects owned by 2 different features).
		synchronized (this) {
			this.listeners = ArrayTools.remove(this.listeners, applicationInstallationListener);
		}
		// Kernel.exit();
	}

	@Override
	public void stateChanged(@Nullable Feature feature, @Nullable State previousState) {
		if (feature != null) {
			Application application = this.applicationManager.getApplication(feature);
			switch (feature.getState()) {
			case STOPPED:
				cleanListeners(feature);
				break;
			case INSTALLED:
				if (application == null) {
					application = this.applicationManager.addApplication(feature);
					notifyInstall(application);
				}
				break;
			case UNINSTALLED:
				if (application != null) {
					notifyUninstall(application);
					this.applicationManager.removeApplication(feature);
				}
				break;
			default:// Nothing to do.
			}
		}
	}

	private void cleanListeners(Feature feature) {
		synchronized (this) {
			int length = this.listeners.length;
			for (int i = 0; i < length; i++) {
				ApplicationInstallationListener listener = this.listeners[i];
				if (Kernel.getOwner(listener) == feature) {
					this.listeners = ArrayTools.remove(this.listeners, i);
				}
			}
		}
	}

	/**
	 * Notifies the listener that an application has been installed.
	 *
	 * @param application
	 *            the installed application.
	 */
	protected void notifyInstall(Application application) {
		ApplicationInstallationListener[] listeners = this.listeners; // Keep reference to avoid synchronization.
		for (ApplicationInstallationListener applicationInstallationListener : listeners) {
			try {
				applicationInstallationListener.onInstall(application);
			} catch (Throwable e) {
				WadappsMultiSandboxMessage.LOGGER.log(Level.INFO, WadappsMultiSandboxMessage.CATEGORY,
						WadappsMultiSandboxMessage.USER_ERROR, e);
			}
		}
	}

	/**
	 * Notifies the listener that an application has been uninstalled.
	 *
	 * @param application
	 *            the uninstalled application.
	 */
	protected void notifyUninstall(Application application) {
		ApplicationInstallationListener[] listeners = this.listeners; // Keep reference to avoid synchronization.
		for (ApplicationInstallationListener applicationInstallationListener : listeners) {
			try {
				applicationInstallationListener.onUninstall(application);
			} catch (Throwable e) {
				WadappsMultiSandboxMessage.LOGGER.log(Level.INFO, WadappsMultiSandboxMessage.CATEGORY,
						WadappsMultiSandboxMessage.USER_ERROR, e);
			}
		}
	}

	/**
	 * Checks that an action on a application is permitted.
	 *
	 * @param application
	 *            the application.
	 * @param action
	 *            the action.
	 */
	protected void checkPermission(@Nullable Application application, String action) {
		SecurityManager securityManager = System.getSecurityManager();
		if (securityManager != null) {
			securityManager.checkPermission(new ApplicationPermission(application, action));
		}
	}

}
