/*
 * Java
 *
 * Copyright 2020 IS2T. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break IS2T warranties on the whole library.
 */
package ej.wadapps.multisandbox.application;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;

import ej.annotation.Nullable;
import ej.storage.Storage;

/**
 * An Helper class which uses the {@link Storage} for storing/loading applications.
 */
/* package */ class ApplicationStorage {

	private static final String TMP = "_tmp"; //$NON-NLS-1$
	private static final String APP = "app_"; //$NON-NLS-1$
	private final Storage storage;
	private final Random random;

	/**
	 * Creates an application storage using the given storage.
	 *
	 * @param storage
	 *            the storage to use for storing/loading applications.
	 */
	public ApplicationStorage(Storage storage) {
		this.storage = storage;
		this.random = new Random();
	}

	public String store(InputStream stream) throws IOException {
		byte[] buffer = new byte[1024];

		// First copy the input stream in a app storage token tmp
		String applicationTokenTmp = randomStorageId(APP, TMP);
		try (OutputStream os = this.storage.store(applicationTokenTmp)) {
			int r = stream.read(buffer);
			while (r != -1) {
				os.write(buffer, 0, r);
				r = stream.read(buffer);
			}
		} catch (IOException e) {
			// clear the storage if io error occurs.
			this.storage.remove(applicationTokenTmp);
			throw e;
		}

		// second rename the app storage token tmp to the final app storage token
		String applicationToken = randomStorageId(APP, null);
		this.storage.move(applicationTokenTmp, applicationToken);
		return applicationToken;
	}

	/*
	 * Returns a random storage id that is not already use.
	 */
	private String randomStorageId(String prefix, @Nullable String suffix) throws IOException {
		// loop until we found a token that is not already use
		while (true) {
			long n = this.random.nextInt(Integer.MAX_VALUE);
			String token = prefix + n + (suffix == null ? "" : suffix); //$NON-NLS-1$
			if (!this.storage.exists(token)) {
				return token;
			}
		}
	}

	public InputStream load(String applicationToken) throws IOException {
		InputStream stream = this.storage.load(applicationToken);
		if (stream == null) {
			throw new IOException();
		}
		return stream;
	}

	public void remove(String applicationToken) throws IOException {
		if (!this.storage.exists(applicationToken)) {
			throw new IOException();
		}
		this.storage.remove(applicationToken);
	}

	@SuppressWarnings("null")
	public String[] list() throws IOException {
		// gets all the known storage ids
		String[] ids = this.storage.getIds();
		// filter the list by only keeping app tokens.
		List<String> appStorages = new ArrayList<>();
		for (String id : ids) {
			if (id.startsWith(APP)) {
				appStorages.add(id);
			}
		}
		return appStorages.toArray(new String[appStorages.size()]);
	}
}
