/*
 * Java
 *
 * Copyright 2020 IS2T. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break IS2T warranties on the whole library.
 */
package ej.wadapps.multisandbox.application;

import java.io.IOException;
import java.io.InputStream;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.NullProgressMonitor;

import ej.basictool.map.PackedMap;
import ej.kf.IncompatibleFeatureException;
import ej.kf.Kernel;
import ej.storage.Storage;
import ej.wadapps.application.Application;
import ej.wadapps.application.ApplicationOperationException;
import ej.wadapps.application.ApplicationPermission;

/**
 * The application installer allows to install and uninstall applications.
 * <p>
 * The application binary is persisted before the installation, therefore the application binary is available after a
 * reboot.
 */
public class PersistentApplicationInstallerKF extends AbstractApplicationInstallerKF {

	private final ApplicationStorage applicationStorage;
	private final PackedMap<Application, String> mappingsApplicationsStorageToken;

	/**
	 * Creates a persistent application installer on KF.
	 * <p>
	 * If the given storage contains applications, those applications will be installed.
	 *
	 * @param applicationManager
	 *            the applications manager it depends on.
	 * @param storage
	 *            the storage used to persist the application's stream and retrieve applications.
	 * @throws IOException
	 *             If an IO error occurs.
	 */
	public PersistentApplicationInstallerKF(ApplicationManagerKF applicationManager, Storage storage)
			throws IOException {
		super(applicationManager);
		this.applicationStorage = new ApplicationStorage(storage);
		this.mappingsApplicationsStorageToken = new PackedMap<>();

		// install all applications found in the given storage.
		for (String token : this.applicationStorage.list()) {
			try (InputStream stream = this.applicationStorage.load(token)) {
				Application application = installApplication(stream, new NullProgressMonitor());
				this.mappingsApplicationsStorageToken.put(application, token);
			} catch (IOException e) {
				// do nothing
			} catch (ApplicationOperationException e) {
				if (e.getCause() instanceof IncompatibleFeatureException) {
					// should we remove the token from the storage ?
					this.applicationStorage.remove(token);
				}
			}
		}
	}

	@Override
	public Application install(InputStream stream, IProgressMonitor monitor) throws ApplicationOperationException {
		checkPermission(null, ApplicationPermission.INSTALL_ACTION);
		Kernel.enter();
		try {
			monitor.beginTask(INSTALL, 3);
			String token = this.applicationStorage.store(stream);
			monitor.worked(1);
			try (InputStream tokenStream = this.applicationStorage.load(token)) {
				Application application = installApplication(tokenStream, monitor);
				this.mappingsApplicationsStorageToken.put(application, token);
				return application;
			} catch (IOException | ApplicationOperationException e) {
				this.applicationStorage.remove(token);
				throw e;
			}
		} catch (IOException e) {
			throw new ApplicationOperationException(e);
		}
		// Kernel.exit();
	}

	/**
	 * Installs an application from an input stream.
	 * <p>
	 * This method can be overridden in sub classes in order to perform some checks (like verifying a signature, etc).
	 *
	 * @param stream
	 *            the stream containing the application.
	 * @param monitor
	 *            the progress monitor to use for reporting progress to the user. It is the caller's responsibility to
	 *            call done() on the given monitor.
	 * @return the application instance.
	 * @throws ApplicationOperationException
	 *             if any unexpected errors occur while installing this application.
	 */
	protected Application installApplication(InputStream stream, IProgressMonitor monitor)
			throws ApplicationOperationException {
		return super.installKF(stream, monitor);
	}

	@Override
	public boolean uninstall(Application application, IProgressMonitor monitor) throws ApplicationOperationException {
		checkPermission(application, ApplicationPermission.UNINSTALL_ACTION);
		Kernel.enter();
		monitor.beginTask(UNINSTALL, 3);
		String token = this.mappingsApplicationsStorageToken.remove(application);
		boolean isUninstalled = super.uninstallKF(application, monitor);
		if (isUninstalled && token != null) {
			try {
				this.applicationStorage.remove(token);
				monitor.worked(1);
			} catch (IOException e) {
				// What TODO if an exception occurs here ?
				// do nothing
			}
		}
		return isUninstalled;
	}
}
