/*
 * Java
 *
 * Copyright 2015-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.widget.container;

import ej.annotation.Nullable;
import ej.giml.annotation.Element;
import ej.giml.annotation.ElementAttribute;
import ej.giml.annotation.ElementChild;
import ej.giml.annotation.ElementConstructor;
import ej.mwt.MWT;
import ej.mwt.Widget;
import ej.style.Style;
import ej.style.container.Rectangle;
import ej.widget.StyledComposite;

/**
 * Lays out its children to fit in 3 regions.
 * <p>
 * The first and the last widgets are packed (get just minimal needed space) against their side, the center gets the
 * remaining size between.
 * <ul>
 * <li>horizontal mode:<br>
 * <img src="doc-files/simpledock_horizontal.png" alt="Horizontal simple dock."></li>
 * <li>vertical mode:<br>
 * <img src="doc-files/simpledock_vertical.png" alt="Vertical simple dock."></li>
 * </ul>
 */
@Element
public class SimpleDock extends StyledComposite {

	private boolean horizontal;
	@Nullable
	private Widget first;
	@Nullable
	private Widget center;
	@Nullable
	private Widget last;

	/**
	 * Creates a horizontal simple dock composite.
	 */
	public SimpleDock() {
		this.horizontal = true;
	}

	/**
	 * Creates a simple dock composite specifying its orientation.
	 *
	 * @param horizontal
	 *            <code>true</code> to set the composite horizontal, <code>false</code> to set the composite vertical.
	 */
	@ElementConstructor
	public SimpleDock(@ElementAttribute(defaultValue = "true") boolean horizontal) {
		this.horizontal = horizontal;
	}

	/**
	 * Sets the composite orientation: horizontal or vertical.
	 *
	 * @param horizontal
	 *            <code>true</code> to set the composite horizontal, <code>false</code> to set the composite vertical.
	 */
	public void setHorizontal(boolean horizontal) {
		this.horizontal = horizontal;
		invalidate();
	}

	/**
	 * Sets the first widget of this composite. If the simple dock is horizontal, the first widget is on the left side,
	 * otherwise it is on the upper side.
	 *
	 * @param widget
	 *            the widget to add.
	 * @throws NullPointerException
	 *             if the given widget is <code>null</code>.
	 * @throws IllegalArgumentException
	 *             if the given widget or one of its children is already connected to a panel.
	 * @see #add(Widget)
	 */
	@ElementChild
	public void setFirst(Widget widget) {
		Widget firstLocal = this.first; // avoid synchronizing
		if (firstLocal != null) {
			remove(firstLocal);
		}
		this.first = widget;
		add(widget);
	}

	/**
	 * Sets the last widget of this composite. If the simple dock is horizontal, the last widget is on the right side,
	 * otherwise it is on the lower side.
	 *
	 * @param widget
	 *            the widget to add.
	 * @throws NullPointerException
	 *             if the given widget is <code>null</code>.
	 * @throws IllegalArgumentException
	 *             if the given widget or one of its children is already connected to a panel.
	 * @see #add(Widget)
	 */
	@ElementChild
	public void setLast(Widget widget) {
		Widget lastLocal = this.last; // avoid synchronizing
		if (lastLocal != null) {
			remove(lastLocal);
		}
		this.last = widget;
		add(widget);
	}

	/**
	 * Sets the center widget of this composite. The center widget is between the first and the last widgets.
	 *
	 * @param widget
	 *            the widget to add.
	 * @throws NullPointerException
	 *             if the given widget is <code>null</code>.
	 * @throws IllegalArgumentException
	 *             if the given widget or one of its children is already connected to a panel.
	 * @see #add(Widget)
	 */
	@ElementChild
	public void setCenter(Widget widget) {
		Widget centerLocal = this.center; // avoid synchronizing
		if (centerLocal != null) {
			remove(centerLocal);
		}
		this.center = widget;
		add(widget);
	}

	@Override
	public void remove(Widget widget) {
		if (widget == this.first) {
			this.first = null;
		} else if (widget == this.last) {
			this.last = null;
		} else if (widget == this.center) {
			this.center = null;
		}
		super.remove(widget);
	}

	/**
	 * Removes the first widget of this simple dock.
	 *
	 * @see #remove(Widget)
	 */
	public void removeFirst() {
		Widget first = this.first;
		if (first != null) {
			remove(first);
			this.first = null;
		}
	}

	/**
	 * Removes the last widget of this simple dock.
	 *
	 * @see #remove(Widget)
	 */
	public void removeLast() {
		Widget last = this.last;
		if (last != null) {
			remove(last);
			this.last = null;
		}
	}

	/**
	 * Removes the center widget of this simple dock.
	 *
	 * @see #remove(Widget)
	 */
	public void removeCenter() {
		Widget center = this.center;
		if (center != null) {
			remove(center);
			this.center = null;
		}
	}

	@Override
	public void removeAllWidgets() {
		super.removeAllWidgets();
		this.first = null;
		this.last = null;
		this.center = null;
	}

	@Override
	public Rectangle validateContent(Style style, Rectangle bounds) {
		int widthHint = bounds.getWidth();
		int heightHint = bounds.getHeight();

		boolean computeWidth = widthHint == MWT.NONE;
		boolean computeHeight = heightHint == MWT.NONE;

		// compute widgets expected width
		int firstWidth;
		int centerWidth;
		int lastWidth;
		if (computeWidth) {
			firstWidth = MWT.NONE;
			centerWidth = MWT.NONE;
			lastWidth = MWT.NONE;
		} else {
			if (this.horizontal) {
				firstWidth = MWT.NONE;
				lastWidth = MWT.NONE;
				centerWidth = MWT.NONE; // reset after other layouts later
			} else {
				firstWidth = widthHint;
				centerWidth = widthHint;
				lastWidth = widthHint;
			}
		}
		// compute widgets expected height
		int firstHeight;
		int centerHeight;
		int lastHeight;
		if (computeHeight) {
			firstHeight = MWT.NONE;
			centerHeight = MWT.NONE;
			lastHeight = MWT.NONE;
		} else {
			if (this.horizontal) {
				firstHeight = heightHint;
				centerHeight = heightHint;
				lastHeight = heightHint;
			} else {
				firstHeight = MWT.NONE;
				lastHeight = MWT.NONE;
				centerHeight = MWT.NONE; // reset after other layouts later
			}
		}

		Widget firstLocal = this.first; // avoid synchronizing
		Widget lastLocal = this.last; // avoid synchronizing
		Widget centerLocal = this.center; // avoid synchronizing

		// validate widgets and get their preferred widgets
		int firstPreferredWidth;
		int firstPreferredHeight;
		if (firstLocal != null) {
			firstLocal.validate(firstWidth, firstHeight);
			firstPreferredWidth = firstLocal.getPreferredWidth();
			firstPreferredHeight = firstLocal.getPreferredHeight();
		} else {
			firstPreferredWidth = 0;
			firstPreferredHeight = 0;
		}
		int lastPreferredWidth;
		int lastPreferredHeight;
		if (lastLocal != null) {
			lastLocal.validate(lastWidth, lastHeight);
			lastPreferredWidth = lastLocal.getPreferredWidth();
			lastPreferredHeight = lastLocal.getPreferredHeight();
		} else {
			lastPreferredWidth = 0;
			lastPreferredHeight = 0;
		}

		// center take remaining size
		if (!computeWidth && this.horizontal) {
			centerWidth = centerSize(widthHint, firstPreferredWidth, lastPreferredWidth);
		}
		if (!computeHeight && !this.horizontal) {
			centerHeight = centerSize(heightHint, firstPreferredHeight, lastPreferredHeight);
		}
		int centerPreferredWidth;
		int centerPreferredHeight;
		if (centerLocal != null) {
			centerLocal.validate(centerWidth, centerHeight);
			centerPreferredWidth = centerLocal.getPreferredWidth();
			centerPreferredHeight = centerLocal.getPreferredHeight();
		} else {
			centerPreferredWidth = 0;
			centerPreferredHeight = 0;
		}

		// compute composite preferred size if necessary
		widthHint = computePreferredSize(true, widthHint, computeWidth, firstPreferredWidth, lastPreferredWidth,
				centerPreferredWidth);
		heightHint = computePreferredSize(false, heightHint, computeHeight, firstPreferredHeight, lastPreferredHeight,
				centerPreferredHeight);

		return new Rectangle(0, 0, widthHint, heightHint);
	}

	private int computePreferredSize(boolean width, int sizeHint, boolean computeSize, int firstPreferredSize,
			int lastPreferredSize, int centerPreferredSize) {
		if (computeSize) {
			if (width == this.horizontal) {
				return firstPreferredSize + centerPreferredSize + lastPreferredSize;
			} else {
				return max(firstPreferredSize, lastPreferredSize, centerPreferredSize);
			}
		}
		return sizeHint;
	}

	private int centerSize(int sizeHint, int firstPreferredSize, int lastPreferredSize) {
		return Math.max(0, sizeHint - firstPreferredSize - lastPreferredSize);
	}

	private int max(int first, int second, int third) {
		return Math.max(Math.max(first, second), third);
	}

	@Override
	protected void setBoundsContent(Rectangle bounds) {
		Widget firstLocal = this.first; // avoid synchronizing
		Widget lastLocal = this.last; // avoid synchronizing

		int x = bounds.getX();
		int y = bounds.getY();
		int width = bounds.getWidth();
		int height = bounds.getHeight();

		// compute widgets bounds
		int firstX = x;
		int firstY = y;
		int firstWidth;
		int firstHeight;
		int lastX;
		int lastY;
		int lastWidth;
		int lastHeight;
		int centerX;
		int centerY;
		int centerWidth;
		int centerHeight;
		if (this.horizontal) {
			if (firstLocal != null) {
				firstWidth = firstLocal.getPreferredWidth();
			} else {
				firstWidth = 0;
			}
			firstHeight = height;
			if (lastLocal != null) {
				lastWidth = lastLocal.getPreferredWidth();
			} else {
				lastWidth = 0;
			}
			lastHeight = height;
			lastX = x + width - lastWidth;
			lastY = y;
			centerX = x + firstWidth;
			centerY = y;
			centerWidth = Math.max(0, width - firstWidth - lastWidth);
			centerHeight = height;
		} else {
			firstWidth = width;
			if (firstLocal != null) {
				firstHeight = firstLocal.getPreferredHeight();
			} else {
				firstHeight = 0;
			}
			lastWidth = width;
			if (lastLocal != null) {
				lastHeight = lastLocal.getPreferredHeight();
			} else {
				lastHeight = 0;
			}
			lastX = x;
			lastY = y + height - lastHeight;
			centerX = x;
			centerY = y + firstHeight;
			centerWidth = width;
			centerHeight = Math.max(0, height - firstHeight - lastHeight);
		}

		// set widgets bounds
		if (firstLocal != null) {
			firstLocal.setBounds(firstX, firstY, firstWidth, firstHeight);
		}
		if (lastLocal != null) {
			lastLocal.setBounds(lastX, lastY, lastWidth, lastHeight);
		}
		Widget centerLocal = this.center; // avoid synchronizing
		if (centerLocal != null) {
			centerLocal.setBounds(centerX, centerY, centerWidth, centerHeight);
		}
	}

}
