/*
 * Java
 *
 * Copyright 2017-2019 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.widget.container.transition;

import java.util.ArrayList;
import java.util.List;

import ej.animation.Animation;
import ej.animation.Animator;
import ej.annotation.Nullable;
import ej.bon.Util;
import ej.components.dependencyinjection.ServiceLoaderFactory;
import ej.microui.display.Display;
import ej.motion.Motion;
import ej.motion.MotionFactory;
import ej.motion.quart.QuartEaseOutMotion;
import ej.mwt.MWT;
import ej.mwt.Panel;
import ej.mwt.Widget;
import ej.style.Style;
import ej.style.container.Rectangle;
import ej.style.util.StyleHelper;
import ej.widget.StyledComposite;
import ej.widget.animation.AnimationListenerRegistry;
import ej.widget.animation.AnimationStepListener;

/**
 * A transition container holds only one widget at a time. When a new widget is shown (replacing the current one), an
 * animated transition is done.
 *
 * @since 2.3.0
 */
public abstract class TransitionContainer extends StyledComposite implements Animation {

	/**
	 * Default transition duration.
	 */
	public static final int DURATION = 300;

	private final List<AnimationStepListener> animationStepListeners;

	private MotionFactory motionFactory;
	private int duration;

	@Nullable
	private Motion motion;
	@Nullable
	private String animationId;
	private long startTime;

	/**
	 * Creates a transition container with a motion factory that creates {@link QuartEaseOutMotion} and a default
	 * duration of {@value #DURATION}.
	 */
	public TransitionContainer() {
		this.animationStepListeners = new ArrayList<>();
		this.motionFactory = new MotionFactory() {
			@Override
			public Motion createMotion(int start, int stop, long duration) {
				return new QuartEaseOutMotion(start, stop, duration);
			}
		};
		this.duration = DURATION;
	}

	/**
	 * Sets the motion factory to use for the animations.
	 *
	 * @param motionFactory
	 *            the motion factory to set.
	 * @throws NullPointerException
	 *             if the given motion factory is <code>null</code>.
	 */
	public void setMotionFactory(MotionFactory motionFactory) {
		if (motionFactory == null) {
			throw new NullPointerException();
		}
		this.motionFactory = motionFactory;
	}

	/**
	 * Gets the motion factory.
	 *
	 * @return the motion factory.
	 */
	public MotionFactory getMotionFactory() {
		return this.motionFactory;
	}

	/**
	 * Sets the animation duration.
	 *
	 * @param duration
	 *            the duration to set.
	 */
	public void setDuration(int duration) {
		this.duration = duration;
	}

	/**
	 * Gets the animation duration.
	 *
	 * @return the duration.
	 */
	public int getDuration() {
		return this.duration;
	}

	/**
	 * Creates a motion based on the registered motion factory and duration.
	 *
	 * @param start
	 *            the start value.
	 * @param stop
	 *            the stop value.
	 * @return the created motion.
	 *
	 * @see #setMotionFactory(MotionFactory)
	 * @see #setDuration(int)
	 */
	protected Motion createMotion(int start, int stop) {
		return this.motionFactory.createMotion(start, stop, this.duration);
	}

	/**
	 * Adds an animation step listener.
	 *
	 * @param animationStepListener
	 *            the listener to add.
	 */
	public void addAnimationStepListener(AnimationStepListener animationStepListener) {
		this.animationStepListeners.add(animationStepListener);
	}

	/**
	 * Removes an animation step listener.
	 *
	 * @param animationStepListener
	 *            the listener to remove.
	 */
	public void removeAnimationStepListener(AnimationStepListener animationStepListener) {
		this.animationStepListeners.remove(animationStepListener);
	}

	/**
	 * Notifies the registered animation step listeners with the given values.
	 *
	 * @param start
	 *            the animation start value.
	 * @param stop
	 *            the animation stop value.
	 * @param from
	 *            the widget before the animation.
	 * @param to
	 *            the widget after the animation.
	 * @param forward
	 *            <code>true</code> if going forward, <code>false</code> otherwise.
	 * @see AnimationStepListener#initialize(int, int, Widget, Widget, boolean)
	 */
	protected void notifyAnimationStepListeners(int start, int stop, Widget from, Widget to, boolean forward) {
		for (AnimationStepListener animationStepListener : this.animationStepListeners) {
			animationStepListener.initialize(start, stop, from, to, forward);
		}
	}

	/**
	 * Notifies the registered animation step listeners with the given values.
	 *
	 * @param value
	 *            the step value.
	 * @see AnimationStepListener#onStep(int)
	 */
	protected void notifyAnimationStepListeners(int value) {
		for (AnimationStepListener animationStepListener : this.animationStepListeners) {
			animationStepListener.onStep(value);
		}
	}

	/**
	 * Shows a new widget.
	 *
	 * @param widget
	 *            the new widget to show.
	 * @param forward
	 *            <code>true</code> if going forward, <code>false</code> otherwise.
	 * @throws IllegalArgumentException
	 *             if the specified widget is already in a hierarchy (already contained in a composite or panel).
	 */
	public abstract void show(Widget widget, boolean forward);

	@Override
	public void hideNotify() {
		super.hideNotify();
		stopAnimation();
	}

	/**
	 * Creates a motion and starts an animation.
	 * <p>
	 * The registered animation step listeners will be notified.
	 * <p>
	 * The animation listeners registered to the {@link AnimationListenerRegistry} will also be notified.
	 * <p>
	 * The callers of this method must be careful to start the animation if and only if the widget is actually visible.
	 *
	 * @param widget
	 *            the new widget.
	 * @param previousWidget
	 *            the previous widget.
	 * @param forward
	 *            <code>true</code> if going forward, <code>false</code> otherwise.
	 * @param start
	 *            the start value of the motion.
	 * @param stop
	 *            the stop value of the motion.
	 */
	protected void startAnimation(Widget widget, Widget previousWidget, boolean forward, int start, int stop) {
		// if this is invisible, the animation is not started and the context is reset
		Panel panel = getPanel();
		boolean shown = panel != null && panel.isShown();
		if (shown) {
			String animationId = String.valueOf(Math.random());
			this.animationId = animationId;
			AnimationListenerRegistry.notifyOnStart(animationId);
			this.startTime = Util.platformTimeMillis();
			notifyAnimationStepListeners(0, getDuration(), previousWidget, widget, forward);
			this.motion = createMotion(start, stop);
			Animator animator = ServiceLoaderFactory.getServiceLoader().getService(Animator.class, Animator.class);
			animator.startAnimation(this);
		} else {
			resetContext();
		}
	}

	/**
	 * Stops the animation.
	 * <p>
	 * If no animation is running, nothing changes.
	 */
	protected void stopAnimation() {
		Animator animator = ServiceLoaderFactory.getServiceLoader().getService(Animator.class, Animator.class);
		animator.stopAnimation(this);
		onAnimationEnd();
	}

	/**
	 * Must be called when the animation is stopped or ends normally. No need to synchronize this method as we keep the
	 * animationID value in a local and AnimationListenerRegistry.notifyOnStop can deal with multiple calls
	 */
	private void onAnimationEnd() {
		String id = this.animationId;
		if (id != null) {
			AnimationListenerRegistry.notifyOnStop(id);
			this.animationId = null;
		}
	}

	/**
	 * Cleans up the context after the animation. The animation may not have been started.
	 */
	protected void resetContext() {
		// nothing to be done at this level, subclasses may add their own context cleanup.
	}

	/**
	 * Cancels the animation (stops it and resets the context).
	 */
	protected void cancelAnimation() {
		stopAnimation();
		resetContext();
	}

	@Override
	public boolean tick(long currentTimeMillis) {
		Motion motion = this.motion;
		if (motion != null) {
			boolean finished = motion.isFinished();
			if (finished) {
				notifyAnimationStepListeners(getDuration());
				// Do the last step of the animation before finalizing the animation (the steps are supposed to be fast,
				// the
				// finalization may be long).
				updateStep(motion.getStopValue());
				Display.getDefaultDisplay().callSerially(new Runnable() {
					@Override
					public void run() {
						resetContext();
						onAnimationEnd();
					}
				});
				return false;
			} else {
				notifyAnimationStepListeners((int) (currentTimeMillis - this.startTime));
				updateStep(motion.getCurrentValue());
				return true;
			}
		} else {
			return false;
		}
	}

	/**
	 * Updates the current step.
	 * <p>
	 * Called during the animation using the current value of the motion.
	 *
	 * @param step
	 *            the current step.
	 */
	protected abstract void updateStep(int step);

	@Override
	public Rectangle validateContent(Style style, Rectangle bounds) {
		int widthHint = bounds.getWidth();
		int heightHint = bounds.getHeight();

		if (getWidgetsCount() == 1) {
			Widget widget = getWidget(0);
			widget.validate(widthHint, heightHint);
			if (widthHint == MWT.NONE) {
				widthHint = widget.getPreferredWidth();
			}
			if (heightHint == MWT.NONE) {
				heightHint = widget.getPreferredHeight();
			}
		}

		return new Rectangle(0, 0, widthHint, heightHint);
	}

	@Override
	protected void setBoundsContent(Rectangle bounds) {
		if (getWidgetsCount() == 1) {
			Widget widget = getWidget(0);
			setBounds(widget, bounds);
		}
	}

	/**
	 * Sets the bounds of its child if one and only one.
	 */
	protected void setChildBounds() {
		if (getWidgetsCount() == 1) {
			Widget widget = getWidget(0);
			Rectangle bounds = new Rectangle(0, 0, getWidth(), getHeight());
			Rectangle contentBounds = StyleHelper.computeContentBounds(bounds, getStyle());
			setBounds(widget, contentBounds);
		}
	}

	private void setBounds(Widget widget, Rectangle contentBounds) {
		int contentX = contentBounds.getX();
		int contentY = contentBounds.getY();
		int contentWidth = contentBounds.getWidth();
		int contentHeight = contentBounds.getHeight();
		widget.setBounds(contentX, contentY, contentWidth, contentHeight);
	}

}
