/*
 * Copyright (c) 2009, 2014 IBM Corp.
 * Copyright 2015-2021 MicroEJ Corp. This file has been modified by MicroEJ Corp.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * and Eclipse Distribution License v1.0 which accompany this distribution.
 *
 * The Eclipse Public License is available at
 *    http://www.eclipse.org/legal/epl-v10.html
 * and the Eclipse Distribution License is available at
 *   http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * Contributors:
 *   Ian Craggs - MQTT 3.1.1 support
 */
package org.eclipse.paho.client.mqttv3.internal;

import org.eclipse.paho.client.mqttv3.IMqttActionListener;
import org.eclipse.paho.client.mqttv3.IMqttToken;
import org.eclipse.paho.client.mqttv3.MqttAsyncClient;
import org.eclipse.paho.client.mqttv3.MqttClientPersistence;
import org.eclipse.paho.client.mqttv3.MqttConnectOptions;
import org.eclipse.paho.client.mqttv3.MqttException;
import org.eclipse.paho.client.mqttv3.MqttPersistenceException;
import org.eclipse.paho.client.mqttv3.MqttToken;

import ej.annotation.Nullable;

/**
 * <p>
 * This class handles the connection of the AsyncClient to one of the available URLs.
 * </p>
 * <p>
 * The URLs are supplied as either the singleton when the client is created, or as a list in the connect options.
 * </p>
 * <p>
 * This class uses its own onSuccess and onFailure callbacks in preference to the user supplied callbacks.
 * </p>
 * <p>
 * An attempt is made to connect to each URL in the list until either a connection attempt succeeds or all the URLs have
 * been tried
 * </p>
 * <p>
 * If a connection succeeds then the users token is notified and the users onSuccess callback is called.
 * </p>
 * <p>
 * If a connection fails then another URL in the list is attempted, otherwise the users token is notified and the users
 * onFailure callback is called
 * </p>
 */
public class ConnectActionListener implements IMqttActionListener {

	@Nullable
	private final MqttClientPersistence persistence;
	private final MqttAsyncClient client;
	private final ClientComms comms;
	private final MqttConnectOptions options;
	private final MqttToken userToken;
	@Nullable
	private final Object userContext;
	@Nullable
	private final IMqttActionListener userCallback;
	private final int originalMqttVersion;

	/**
	 * @param persistence
	 * @param client
	 * @param comms
	 * @param options
	 * @param userToken
	 * @param userContext
	 * @param userCallback
	 */
	public ConnectActionListener(MqttAsyncClient client, MqttClientPersistence persistence, ClientComms comms,
			MqttConnectOptions options, MqttToken userToken, @Nullable Object userContext,
			@Nullable IMqttActionListener userCallback) {
		this.persistence = persistence;
		this.client = client;
		this.comms = comms;
		this.options = options;
		this.userToken = userToken;
		this.userContext = userContext;
		this.userCallback = userCallback;
		this.originalMqttVersion = options.getMqttVersion();
	}

	/**
	 * If the connect succeeded then call the users onSuccess callback
	 *
	 * @param token
	 */
	@Override
	public void onSuccess(IMqttToken token) {
		if (this.originalMqttVersion == MqttConnectOptions.MQTT_VERSION_DEFAULT) {
			this.options.setMqttVersion(MqttConnectOptions.MQTT_VERSION_DEFAULT);
		}
		this.userToken.internalTok.markComplete(token.getResponse(), null);
		this.userToken.internalTok.notifyComplete();

		if (this.userCallback != null) {
			this.userToken.setUserContext(this.userContext);
			this.userCallback.onSuccess(this.userToken);
		}
	}

	/**
	 * The connect failed, so try the next URI on the list. If there are no more URIs, then fail the overall connect.
	 *
	 * @param token
	 * @param exception
	 */
	@Override
	public void onFailure(IMqttToken token, Throwable exception) {

		int numberOfURIs = this.comms.getNetworkModules().length;
		int index = this.comms.getNetworkModuleIndex();

		if ((index + 1) < numberOfURIs || (this.originalMqttVersion == MqttConnectOptions.MQTT_VERSION_DEFAULT
				&& this.options.getMqttVersion() == MqttConnectOptions.MQTT_VERSION_3_1_1)) {

			if (this.originalMqttVersion == MqttConnectOptions.MQTT_VERSION_DEFAULT) {
				if (this.options.getMqttVersion() == MqttConnectOptions.MQTT_VERSION_3_1_1) {
					this.options.setMqttVersion(MqttConnectOptions.MQTT_VERSION_3_1);
				} else {
					this.options.setMqttVersion(MqttConnectOptions.MQTT_VERSION_3_1_1);
					this.comms.setNetworkModuleIndex(index + 1);
				}
			} else {
				this.comms.setNetworkModuleIndex(index + 1);
			}
			try {
				connect();
			} catch (MqttPersistenceException e) {
				onFailure(token, e); // try the next URI in the list
			}
		} else {
			if (this.originalMqttVersion == MqttConnectOptions.MQTT_VERSION_DEFAULT) {
				this.options.setMqttVersion(MqttConnectOptions.MQTT_VERSION_DEFAULT);
			}
			MqttException ex;
			if (exception instanceof MqttException) {
				ex = (MqttException) exception;
			} else {
				ex = new MqttException(exception);
			}
			this.userToken.internalTok.markComplete(null, ex);
			this.userToken.internalTok.notifyComplete();

			if (this.userCallback != null) {
				this.userToken.setUserContext(this.userContext);
				this.userCallback.onFailure(this.userToken, exception);
			}
		}
	}

	/**
	 * Start the connect processing
	 *
	 * @throws MqttPersistenceException
	 */
	public void connect() throws MqttPersistenceException {
		MqttToken token = new MqttToken(this.client.getClientId());
		token.setActionCallback(this);
		token.setUserContext(this);

		this.persistence.open(this.client.getClientId(), this.client.getServerURI());

		if (this.options.isCleanSession()) {
			this.persistence.clear();
		}

		if (this.options.getMqttVersion() == MqttConnectOptions.MQTT_VERSION_DEFAULT) {
			this.options.setMqttVersion(MqttConnectOptions.MQTT_VERSION_3_1_1);
		}

		try {
			this.comms.connect(this.options, token);
		} catch (MqttException e) {
			onFailure(token, e);
		}
	}

}
