/*
 * Copyright (c) 2009, 2014 IBM Corp.
 * Copyright 2018-2021 MicroEJ Corp. This file has been modified by MicroEJ Corp.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * and Eclipse Distribution License v1.0 which accompany this distribution.
 *
 * The Eclipse Public License is available at
 *    http://www.eclipse.org/legal/epl-v10.html
 * and the Eclipse Distribution License is available at
 *   http://www.eclipse.org/org/documents/edl-v10.php.
 *
 * Contributors:
 *    Dave Locke - initial API and implementation and/or initial documentation
 *    Ian Craggs - MQTT 3.1.1 support
 */
package org.eclipse.paho.client.mqttv3.internal.wire;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

import org.eclipse.paho.client.mqttv3.MqttException;
import org.eclipse.paho.client.mqttv3.MqttMessage;

import ej.annotation.Nullable;

/**
 * An on-the-wire representation of an MQTT CONNECT message.
 */
public class MqttConnect extends MqttWireMessage {

	public static final String KEY = "Con";

	private final String clientId;
	private boolean cleanSession;
	@Nullable
	private MqttMessage willMessage;
	@Nullable
	private String userName;
	@Nullable
	private char[] password;
	private final int keepAliveInterval;
	@Nullable
	private String willDestination;
	private int MqttVersion;

	/**
	 * Constructor for an on the wire MQTT connect message
	 *
	 * @param info
	 * @param data
	 * @throws IOException
	 * @throws MqttException
	 */
	public MqttConnect(byte info, byte[] data) throws IOException, MqttException {
		super(MqttWireMessage.MESSAGE_TYPE_CONNECT);
		ByteArrayInputStream bais = new ByteArrayInputStream(data);
		DataInputStream dis = new DataInputStream(bais);

		String protocol_name = decodeUTF8(dis);
		int protocol_version = dis.readByte();
		byte connect_flags = dis.readByte();
		this.keepAliveInterval = dis.readUnsignedShort();
		this.clientId = decodeUTF8(dis);
		dis.close();
	}

	public MqttConnect(String clientId, int MqttVersion, boolean cleanSession, int keepAliveInterval, String userName,
			char[] password, MqttMessage willMessage, String willDestination) {
		super(MqttWireMessage.MESSAGE_TYPE_CONNECT);
		this.clientId = clientId;
		this.cleanSession = cleanSession;
		this.keepAliveInterval = keepAliveInterval;
		this.userName = userName;
		this.password = password;
		this.willMessage = willMessage;
		this.willDestination = willDestination;
		this.MqttVersion = MqttVersion;
	}

	@Override
	public String toString() {
		String rc = super.toString();
		rc += " clientId " + this.clientId + " keepAliveInterval " + this.keepAliveInterval;
		return rc;
	}

	@Override
	protected byte getMessageInfo() {
		return (byte) 0;
	}

	public boolean isCleanSession() {
		return this.cleanSession;
	}

	@Override
	protected byte[] getVariableHeader() throws MqttException {
		try {
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			DataOutputStream dos = new DataOutputStream(baos);

			if (this.MqttVersion == 3) {
				encodeUTF8(dos, "MQIsdp");
			} else if (this.MqttVersion == 4) {
				encodeUTF8(dos, "MQTT");
			}
			dos.write(this.MqttVersion);

			byte connectFlags = 0;

			if (this.cleanSession) {
				connectFlags |= 0x02;
			}

			if (this.willMessage != null) {
				connectFlags |= 0x04;
				connectFlags |= (this.willMessage.getQos() << 3);
				if (this.willMessage.isRetained()) {
					connectFlags |= 0x20;
				}
			}

			if (this.userName != null) {
				connectFlags |= 0x80;
				if (this.password != null) {
					connectFlags |= 0x40;
				}
			}
			dos.write(connectFlags);
			dos.writeShort(this.keepAliveInterval);
			dos.flush();
			return baos.toByteArray();
		} catch (IOException ioe) {
			throw new MqttException(ioe);
		}
	}

	@Override
	public byte[] getPayload() throws MqttException {
		try {
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			DataOutputStream dos = new DataOutputStream(baos);
			encodeUTF8(dos, this.clientId);

			if (this.willMessage != null) {
				encodeUTF8(dos, this.willDestination);
				dos.writeShort(this.willMessage.getPayload().length);
				dos.write(this.willMessage.getPayload());
			}

			if (this.userName != null) {
				encodeUTF8(dos, this.userName);
				if (this.password != null) {
					encodeUTF8(dos, new String(this.password));
				}
			}
			dos.flush();
			return baos.toByteArray();
		} catch (IOException ex) {
			throw new MqttException(ex);
		}
	}

	/**
	 * Returns whether or not this message needs to include a message ID.
	 */
	@Override
	public boolean isMessageIdRequired() {
		return false;
	}

	@Override
	public String getKey() {
		return KEY;
	}
}
