/*
 * Java
 *
 * Copyright 2015-2018 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 *
 */
package com.microej.wadapps.storage.kf;

import java.io.IOException;
import java.io.InputStream;
import java.util.logging.Level;
import java.util.logging.Logger;

import ej.components.dependencyinjection.ServiceLoaderFactory;
import ej.kf.Kernel;
import ej.util.filemanager.FilesManager;
import ej.util.filemanager.StorageID;
import ej.wadapps.admin.Application;
import ej.wadapps.admin.ApplicationsManager;
import ej.wadapps.storage.Storage;

/**
 * Implementation of storage API.
 */
@SuppressWarnings("nls")
public class StorageImpl implements Storage {
	/**
	 * The character used to separate directories with sub-directories or files.
	 */
	public static final char SEPARATOR_CHAR = '/';
	private static final boolean DEBUG = false;
	private static final String KERNEL_PARENT_DIRECTORY = "kernel";
	private static final String FEATURES_PARENT_DIRECTORY = "features";
	private FilesManager filesManager;
	private ApplicationsManager applicationsManager;
	private static final String[] EMPTY = new String[0];
	private final Logger logger = Logger.getLogger(Storage.class.getName());

	@Override
	public void store(String id, InputStream stream) throws IOException {
		StorageID.check(id);
		String dir = getParentDirectory();
		Kernel.enter();
		initialize();
		this.filesManager.write(dir, id, stream);
	}

	@Override
	public InputStream load(String id) throws IOException {
		StorageID.check(id);
		String dir = getParentDirectory();
		Kernel.enter();
		initialize();
		return this.filesManager.read(dir, id);
	}

	@Override
	public boolean exists(String id) throws IOException {
		StorageID.check(id);
		String dir = getParentDirectory();
		Kernel.enter();
		initialize();
		return this.filesManager.exists(dir, id);
	}

	@Override
	public void move(String src, String dst) throws IOException {
		StorageID.check(src);
		StorageID.check(dst);
		String dir = getParentDirectory();
		Kernel.enter();
		initialize();
		this.filesManager.move(dir, src, dst);
	}

	@Override
	public void remove(String id) throws IOException {
		StorageID.check(id);
		String dir = getParentDirectory();
		Kernel.enter();
		initialize();
		if (!this.filesManager.delete(dir, id)) {
			throw new IOException("Cannot delete " + id);
		}
	}

	@Override
	public String[] getIds() throws IOException {
		String dir = getParentDirectory();
		Kernel.enter();
		initialize();
		String[] list = this.filesManager.list(dir);
		if (list == null) {
			list = EMPTY;
		}

		return list;
	}

	private synchronized void initialize() throws IOException {
		if (this.filesManager == null) {
			this.filesManager = ServiceLoaderFactory.getServiceLoader().getService(FilesManager.class);
			if (this.filesManager == null) {
				throw new IOException("Cannot found an implementation of " + FilesManager.class.getName());
			}
			try {
				this.filesManager.initialize();
			} catch (IOException e) {
				if (DEBUG) {
					this.logger.log(Level.SEVERE, "Failed to initialize the storage", e);
				}
				this.filesManager = null;
				throw new IOException("Failed to initialize the storage", e);
			}
		}
	}

	private String getParentDirectory() {
		if (Kernel.isInKernelMode()) {
			return KERNEL_PARENT_DIRECTORY;
		} else {
			if (this.applicationsManager == null) {
				Kernel.enter();
				this.applicationsManager = ServiceLoaderFactory.getServiceLoader()
						.getService(ApplicationsManager.class);
				Kernel.exit();
			}
			Application application = this.applicationsManager.getCurrentApplication();
			return FEATURES_PARENT_DIRECTORY + SEPARATOR_CHAR + application.getIdentifier();
		}
	}
}
