/*
 * Java
 *
 * Copyright 2015-2018 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 *
 */
package ej.util.filemanager.ram;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

import ej.util.filemanager.FilesManager;

public class RAMFilesManager implements FilesManager {
	private final Map<String, byte[]> streams = new HashMap<>();

	private final Logger logger = Logger.getLogger(RAMFilesManager.class.getName());

	@Override
	public void initialize() throws IOException {
	}

	@Override
	public void shutdown() throws IOException {
	}

	@Override
	public void write(String dirPath, String name, InputStream contents) throws IOException {
		ByteArrayOutputStream os = new ByteArrayOutputStream();
		byte[] buffer = new byte[512];
		int r = contents.read(buffer);
		while (r != -1) {
			os.write(buffer, 0, r);
			r = contents.read(buffer);
		}

		String key = createKey(dirPath, name);
		this.logger.info("Write " + key);
		this.streams.put(key, os.toByteArray());
	}

	@Override
	public InputStream read(String dirPath, String name) {
		String key = createKey(dirPath, name);
		this.logger.info("Read " + key);
		byte[] bytes = this.streams.get(key);
		if (bytes == null) {
			return null;
		} else {
			return new ByteArrayInputStream(bytes);
		}
	}

	@Override
	public boolean exists(String dirPath, String name) throws IOException {
		String key = createKey(dirPath, name);
		this.logger.finer("Does " + key + " exist?");
		return this.streams.containsKey(key);
	}

	@Override
	public void move(String dirPath, String src, String dst) throws IOException {
		String srcKey = createKey(dirPath, src);
		if (!this.streams.containsKey(srcKey)) {
			throw new IllegalArgumentException(src + " does not exist.");
		}
		this.logger.info("Move " + src + " to " + dst);
		byte[] ret = this.streams.remove(srcKey);
		if (null == ret) {
			throw new IOException("Cannot move '" + src + "' to '" + dst + "'.");
		}
		String dstKey = createKey(dirPath, dst);
		this.streams.put(dstKey, ret);
	}

	@Override
	public boolean delete(String dirPath, String name) {
		String key = createKey(dirPath, name);
		if (!this.streams.containsKey(key)) {
			throw new IllegalArgumentException(name + " does not exist.");
		}
		this.logger.info("Delete " + key);
		return this.streams.remove(key) != null;
	}

	@Override
	public String[] list(String dirPath) throws IOException {
		String prefixKey = dirPath + SEPARATOR_CHAR;
		List<String> keys = new ArrayList<>();
		Iterator<String> it = this.streams.keySet().iterator();
		while (it.hasNext()) {
			String key = it.next();
			if (key.startsWith(prefixKey)) {
				keys.add(getId(key, dirPath));
			}
		}
		return keys.toArray(new String[keys.size()]);
	}

	private String createKey(String dirPath, String name) {
		return dirPath + SEPARATOR_CHAR + name;
	}

	private String getId(String key, String dirPath) {
		return key.substring(key.lastIndexOf(SEPARATOR_CHAR) + 1);
	}

	@Override
	public void clear() {
		this.streams.clear();
	}

	private static final char SEPARATOR_CHAR = '/';
}
